package net.sf.basedb.clients.web.extensions.fileviewer;

import java.io.IOException;
import java.io.StringWriter;
import net.sf.basedb.clients.web.extensions.ExtensionsControl;
import net.sf.basedb.clients.web.extensions.JspContext;
import net.sf.basedb.clients.web.extensions.toolbar.ButtonAction;
import net.sf.basedb.util.extensions.ExtensionsInvoker;

/**
	Helper class for handling the "file-viewer" extension point. 
	A {@link JspContext} and {@link FileViewerContext} is needed.
	This class will initialize the extension point in the constructor.
	To invoke extensions, first update the FileViewerContext with the 
	current file and other information and then call {@link #render()}.

	@author nicklas
	@since 3.16
*/
public class FileViewerUtil 
{

	/**
		Get an existing instance that is stored in the given JSP context.
		@return A FileViewerUtil instance or null
		@since 3.17
	*/
	public static FileViewerUtil get(JspContext jspContext)
	{
		return (FileViewerUtil)jspContext.getAttribute("net.sf.basedb.clients.web.file-viewer");
	}

	/**
		Get an existing instance or create a new instance. If a new instance is
		created it is stored in the given JSP context if {@link ExtensionsInvoker#getNumExtensions()}
		is more than 0.
		
		@return A FileViewerUtil instance
		@since 3.17
	*/
	public static FileViewerUtil getOrCreate(JspContext jspContext, FileViewerContext fileContext)
	{
		FileViewerUtil fileViewer = get(jspContext);
		if (fileViewer == null)
		{
			fileViewer = new FileViewerUtil(jspContext, fileContext);
			if (fileViewer.getInvoker().getNumExtensions() > 0)
			{
				jspContext.setAttribute("net.sf.basedb.clients.web.file-viewer", fileViewer);
			}
		}
		return fileViewer;
	}
	
	private final JspContext jspContext;
	private final FileViewerContext fileContext;
	private final ExtensionsInvoker<ButtonAction> fileActions;
	private String prefix = "[";
	private String suffix = "]";

	/**
	 
		@deprecated In 3.17, use {@link #getOrCreate(JspContext, FileViewerContext)} instead
	*/
	@Deprecated
	public FileViewerUtil(JspContext jspContext, FileViewerContext fileContext)
	{
		this.jspContext = jspContext;
		this.fileContext = fileContext;
		Object currentItem = jspContext.getCurrentItem();
		jspContext.setCurrentItem(fileContext);
		this.fileActions = ExtensionsControl.useExtensions(jspContext, "net.sf.basedb.clients.web.file-viewer");
		jspContext.setCurrentItem(currentItem);
	}
	
	/**
		Get the file context.
	*/
	public FileViewerContext getFileContext()
	{
		return fileContext;
	}

	/**
		Get the JSP context.
	*/
	public JspContext getJspContext()
	{
		return jspContext;
	}

	/**
		Get the invoker that is handling the extensions.
		@since 3.17
	*/
	public ExtensionsInvoker<ButtonAction> getInvoker()
	{
		return fileActions;
	}
	
	/**
		Sets a prefix and suffix string to use before and after the list of actions.
		The default is "[" and "]".
	*/
	public void setPrefixAndSuffix(String prefix, String suffix)
	{
		this.prefix = prefix;
		this.suffix = suffix;
	}
	
	/**
		Render the extensions with the current file context information
		and write it to the JSP output.
	*/
	public void render()
		throws IOException
	{
		String result = renderAsString();
		if (result.length() > 0)
		{
			jspContext.getOut().write(result);
		}
	}
	
	/**
		Render the extensions with the current file context information
		and return it as a string. Returns an empty string if there are no
		actions.
	*/
	public String renderAsString()
	{
		// Need to repeat this since to make sure the correct context is always used
		Object currentItem = jspContext.getCurrentItem();
		jspContext.setCurrentItem(fileContext);
		
		// Create a writer for collecting output from the extensions
		StringWriter actions = new StringWriter();
		jspContext.getPageContext().pushBody(actions);
		fileActions.renderDefault();
		jspContext.getPageContext().popBody();
		jspContext.setCurrentItem(currentItem);
		
		String result = "";
		if (actions.getBuffer().length() > 0)
		{
			result = "<span class=\"file-actions\">" + prefix + actions.toString() + suffix + "</span>";
		}
		return result;
	}
	
}
