/**
	$Id$

	Copyright (C) 2012 Nicklas Nordborg

	This file is part of BASE - BioArray Software Environment.
	Available at http://base.thep.lu.se/

	BASE is free software; you can redistribute it and/or
	modify it under the terms of the GNU General Public License
	as published by the Free Software Foundation; either version 3
	of the License, or (at your option) any later version.

	BASE is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with BASE. If not, see <http://www.gnu.org/licenses/>.
*/
package net.sf.basedb.clients.web.extensions.list;


import net.sf.basedb.clients.web.extensions.ExtensionsControl;
import net.sf.basedb.clients.web.extensions.JspContext;
import net.sf.basedb.clients.web.extensions.toolbar.ToolbarUtil;
import net.sf.basedb.core.Annotatable;
import net.sf.basedb.core.Item;
import net.sf.basedb.core.Ownable;
import net.sf.basedb.core.Removable;
import net.sf.basedb.core.Shareable;
import net.sf.basedb.core.data.AnnotatableData;
import net.sf.basedb.core.data.OwnableData;
import net.sf.basedb.core.data.RemovableData;
import net.sf.basedb.core.data.ShareableData;
import net.sf.basedb.core.plugin.GuiContext;
import net.sf.basedb.util.extensions.ExtensionPointBean;
import net.sf.basedb.util.extensions.ExtensionsInvoker;
import net.sf.basedb.util.extensions.Registry;

/**
	Utility class for list column extension points.
	
	@author Nicklas
	@since 3.2
	@base.modified $Date$
*/
public class ListColumnUtil
{

	/**
		An array containing all items that we are registering an
		extension point for. The extension point id for an item
		is constructed as: EP_PREFIX + Item.name().toLowerCase().
	*/
	public static final Item[] LIST_ITEMS = ToolbarUtil.TOOLBAR_ITEMS;

	
	/**
		ID prefixes for list column extension point. The final
		ID is the prefix + Item.name().toLowerCase().
	*/
	public static String EP_PREFIX = "net.sf.basedb.clients.web.listcolumn.";

	
	/**
		ID for the list column extension point that appear on all list pages.
	*/
	public static String EP_ALL = EP_PREFIX + "all";

	/**
		ID for the list column extension point that appear on all pages
		that list {@link Ownable} items.
	*/
	public static String EP_OWNABLE = EP_PREFIX + "ownable";
	
	/**
		ID for the list column extension point that appear on all pages
		that list {@link Shareable} items.
	*/
	public static String EP_SHAREABLE = EP_PREFIX + "shareable";
	
	/**
		ID for the list column extension point that appear on all pages
		that list {@link Removable} items.
	*/
	public static String EP_REMOVABLE = EP_PREFIX + "removeable";

	/**
		ID for the list column extension point that appear on all pages
		that list {@link Annotatable} items.
	*/
	public static String EP_ANNOTATABLE = EP_PREFIX + "annotatable";

	/**
		Use list column extensions for a given gui context. This method
		will assemble all specific and generic extension points and
		then call {@link ExtensionsControl#useExtensions(JspContext, String...)}.
		
		@param jspContext The current jsp context
		@return An invoker instance
	*/
	public static <I> ExtensionsInvoker<ListColumnAction<I, ?>> useExtensions(JspContext jspContext)
	{
		String[] ep = new String[6]; // 6 is the maximum number of extension points
		int index = 0;
		GuiContext guiContext = jspContext.getGuiContext();
		Item itemType = guiContext.getItem();
		Class<?> dataClass = itemType.getDataClass();
		if (dataClass != null)
		{
			ep[index++] = EP_ALL;
			if (OwnableData.class.isAssignableFrom(dataClass)) 
			{
				ep[index++] = EP_OWNABLE;
			}
			if (ShareableData.class.isAssignableFrom(dataClass)) 
			{
				ep[index++] = EP_SHAREABLE;
			}
			if (RemovableData.class.isAssignableFrom(dataClass)) 
			{
				ep[index++] = EP_REMOVABLE;
			}
			if (AnnotatableData.class.isAssignableFrom(dataClass)) 
			{
				ep[index++] = EP_ANNOTATABLE;
			}
		}
		ep[index++] = EP_PREFIX + itemType.name().toLowerCase();
		return ExtensionsControl.useExtensions(jspContext, ep);
	}
	

	public static void registerExtensionPoints(Registry registry)
	{
		register(registry, EP_ALL, "List column: all list pages");
		register(registry, EP_OWNABLE, "List column: list pages for items that are owned by a user");
		register(registry, EP_SHAREABLE, "List column: list pages for items that can be shared");
		register(registry, EP_REMOVABLE, "List column: list pages for items that can be put in the trashcan");
		register(registry, EP_ANNOTATABLE, "List column: list pages for annotatable items");
		
		// Special lists
		register(registry, EP_PREFIX + "trashcan", "List column: trashcan list");
		register(registry, EP_PREFIX + "allitems", "List column: all items");
		register(registry, EP_PREFIX + "projectitems", "List column: project items");
		
		for (Item item : LIST_ITEMS)
		{
			String suffix = item.name().toLowerCase();
			register(registry, EP_PREFIX + suffix, item + ": List column");
		}
	}
	
	@SuppressWarnings("rawtypes")
	private static void register(Registry registry, String id, String name)
	{
		ExtensionPointBean<ListColumnAction> column = new ExtensionPointBean<ListColumnAction>();
		column.setActionClass(ListColumnAction.class);
		column.setRendererFactory(null);
		column.setAllowRendererOverrider(false);
		column.setId(id);
		column.setName(name);
		column.setDescription(
			"Extension point for adding a column to the list on the specified page. " +
			"Extensions should provide 'ListColumnAction' instances with " +
			"information about the column. Actions are always rendered by the " + 
			"Table taglib. This extension point supports " +
			"custom stylesheets and java-scripts."
			);
		registry.registerExtensionPoint(column, null);
	}
}
