package net.sf.basedb.clients.web.extensions.list;

import java.util.Collections;
import java.util.HashMap;
import java.util.Map;
import java.util.Set;
import java.util.TreeSet;

import net.sf.basedb.clients.web.Base;
import net.sf.basedb.clients.web.formatter.FormatterFactory;
import net.sf.basedb.clients.web.formatter.LinkedItemFormatter;
import net.sf.basedb.core.Annotatable;
import net.sf.basedb.core.AnnotationType;
import net.sf.basedb.core.DbControl;
import net.sf.basedb.core.Item;
import net.sf.basedb.core.ItemList;
import net.sf.basedb.core.ItemQuery;
import net.sf.basedb.core.ItemSubtype;
import net.sf.basedb.core.Metadata;
import net.sf.basedb.core.RawDataTypes;
import net.sf.basedb.core.Type;
import net.sf.basedb.core.query.Expressions;
import net.sf.basedb.core.query.Hql;
import net.sf.basedb.core.query.Restriction;
import net.sf.basedb.core.query.Restrictions;
import net.sf.basedb.core.Metadata.PropertyPath;
import net.sf.basedb.core.RawDataType;
import net.sf.basedb.core.SyncFilter.SourceItemTransform;
import net.sf.basedb.util.Enumeration;
import net.sf.basedb.util.NameableComparator;
import net.sf.basedb.util.Values;
import net.sf.basedb.util.filter.Filter;
import net.sf.basedb.util.formatter.CollectionFormatter;
import net.sf.basedb.util.formatter.NameableFormatter;
import net.sf.basedb.util.formatter.ToStringFormatter;
import net.sf.basedb.util.listable.ListableUtil;
import net.sf.basedb.util.listable.RestrictionTransformer;
import net.sf.basedb.util.listable.SourceItemTransformer;
import net.sf.basedb.util.listable.SourceItemTransformerFactory;

/**
	A column action implementation that finds a parent or child item of a given 
	subtype and	then get an annotation or other property value from it.
	
	@author nicklas
	@since 3.16, 3.17
*/
public abstract class RelatedItemColumn
	extends AbstractListColumnBean<Annotatable, Object>
{
	
	/**
		For now, we assume that 'expr' is /DIRECTION/ITEMTYPE/subtype-id/#annotationtype-id or
		/DIRECTION/ITEMTYPE/subtype-id/property
		
		DIRECTION=CHILD|PARENT (PARENT is assumed if not specified)
		'subtype-id' can also be the rawdatatype-id if the ITEMTYPE=Item.RAWBIOASSAY
	*/
	@SuppressWarnings({ "unchecked", "rawtypes" })
	public static RelatedItemColumn create(DbControl dc, int index, Item sourceType, String expr, RelatedItemHelper helper)
	{
		RelatedItemColumn col = null;
		try
		{
			String[] tmp = expr.split("/", 5);
			SourceItemTransform direction = SourceItemTransform.CHILD_TO_PARENT;
			int baseIndex = 1;
			if (tmp.length == 5)
			{
				baseIndex = 2;
				direction = "CHILD".equals(tmp[1]) ? SourceItemTransform.PARENT_TO_CHILD : SourceItemTransform.CHILD_TO_PARENT;
			}
			Item targetType = Item.valueOf(tmp[baseIndex]);
			Filter<? super ItemList> itemListFilter = null;
			Restriction targetRestriction = null;
			String subtypeName = null;
			if (targetType == Item.RAWBIOASSAY)
			{
				RawDataType rdt = RawDataTypes.getSafeRawDataType(tmp[baseIndex+1]);
				subtypeName = rdt.getName();
				targetRestriction = Restrictions.eq(Hql.property("rawDataType"), Expressions.string(rdt.getId()));
				itemListFilter = new RawDataTypeFilter(rdt);
			}
			else
			{
				ItemSubtype subtype = ItemSubtype.getById(dc, Values.getInt(tmp[baseIndex+1]));
				subtypeName = subtype.getName();
				targetRestriction = Restrictions.eq(Hql.property("itemSubtype"), Hql.entity(subtype));
				itemListFilter = new SubtypeFilter(subtype);
			}
			String propSpec = tmp[baseIndex+2];
			if (propSpec.startsWith("#"))
			{
				AnnotationType at = AnnotationType.getById(dc, Values.getInt(propSpec.substring(1)));
				col = new RelatedItemAnnotationColumn(index, expr, sourceType, targetType, direction, targetRestriction, helper, at);
				col.setTitle(subtypeName+"."+at.getName() + " [A]");
				col.setTooltip(subtypeName + " › "+at.getName());
				col.setValueType(at.getValueType());
				col.setClazz("relateditemcol");
				col.setExportFormatter(FormatterFactory.getTypeFormatter(dc.getSessionControl(), at.getValueType()));
				if (!helper.lazy) col.setFormatter(new CollectionFormatter(col.getExportFormatter()));
			}
			else if (propSpec.equals("§itemLists"))
			{
				col = new RelatedItemListColumn(dc, index, expr, sourceType, targetType, direction, targetRestriction, itemListFilter, helper);
				col.setTitle(subtypeName+".Item list");
				col.setValueType(Type.INT);
				col.setClazz("relateditemcol");
				col.setExportFormatter(new NameableFormatter());
				if (!helper.lazy) col.setFormatter(new CollectionFormatter(new LinkedItemFormatter()));
			}
			else
			{
				PropertyPath property = null;
				if (".".equals(propSpec)) // special case '.' is equal to the "parent" item
				{
					col = new RelatedItemPropertyColumn(index, expr, sourceType, targetType, direction, targetRestriction, helper, null);
					col.setTitle(subtypeName);
					col.setTooltip(subtypeName);
				}
				else
				{
					Metadata m = Metadata.getInstance(targetType.getItemClass());
					property = m.getPropertyPath(propSpec, false);
					col = new RelatedItemPropertyColumn(index, expr, sourceType, targetType, direction, targetRestriction, helper, property);
					col.setTitle(subtypeName+"."+pathToTitle(propSpec));
					col.setTooltip(subtypeName + " › "+propSpec);
				}
				col.setClazz("relateditemcol");
				if (property == null || property.getHibernateType().isEntityType())
				{
					col.setExportFormatter(new NameableFormatter());
					if (!helper.lazy) col.setFormatter(new CollectionFormatter(new LinkedItemFormatter()));
					if ("job.itemSubtype".equals(propSpec))
					{
						ItemQuery<ItemSubtype> jobSubtypesQuery = Base.getSubtypesQuery(Item.JOB);
						col.setEnumeration(Enumeration.fromItems(jobSubtypesQuery.list(dc), null));
						col.setValueType(Type.INT);
					}
					else
					{
						if (col.getFilterProperty()==null) 
						{
							// Add '.name' to make filtering work
							String nameProperty = col.getProperty() + ".name";
							if (nameProperty.contains("/..")) nameProperty = nameProperty.replace("/..", "/");
							col.setFilterProperty(nameProperty);
						}
						col.setValueType(Type.STRING);
					}
				}
				else
				{
					Type colType = Type.fromHibernateType(property.getHibernateType());
					if (colType != null)
					{
						col.setValueType(colType);
						col.setExportFormatter(FormatterFactory.getTypeFormatter(dc.getSessionControl(), colType));
					}
					else
					{
						// Last fallback -- if we get here the Type.fromHibernateType() should be update with special cases
						col.setValueType(Type.STRING);
						col.setExportFormatter(new ToStringFormatter<>());
					}
					if (!helper.lazy) col.setFormatter(new CollectionFormatter(col.getExportFormatter()));
				}
			}
			col.setClazz(col.getClazz() + (direction == SourceItemTransform.CHILD_TO_PARENT ? " parentitem" : " childitem"));
			col.setTooltip((direction == SourceItemTransform.CHILD_TO_PARENT ? "Parent › " : "Child › ")+col.getTooltip());
		}
		catch (RuntimeException ex)
		{
			ex.printStackTrace(System.out);
		}
		
		
		return col;
	}

	private static Map<String, String> pathTitles = new HashMap<>();
	
	static
	{
		// See also add_relateditem_columns.js
		pathTitles.put("externalId", "External ID");
		pathTitles.put("bioWell.bioPlate", "Bioplate");
		pathTitles.put("entryDate", "Registered");
		pathTitles.put("remainingQuantity", "Remaining quantity");
		pathTitles.put("originalQuantity", "Original quantity");
		pathTitles.put("creationEvent.entryDate", "Registration date");
		pathTitles.put("creationEvent.eventDate", "Creation date");
		pathTitles.put("creationEvent.protocol", "Protocol");
		pathTitles.put("creationEvent.hardware", "Hardware");
		pathTitles.put("protocol", "Protocol");
		pathTitles.put("hardware", "Hardware");
		pathTitles.put("software", "Software");
		pathTitles.put("tag", "Tag");
		pathTitles.put("job.itemSubtype", "Job type");
		pathTitles.put("$itemLists", "Item lists");
	}
	
	/**
		Convert some paths to a more user-friendly title instead.
	*/
	private static String pathToTitle(String path)
	{
		String title = pathTitles.get(path);
		return title == null ? path.substring(0, 1).toUpperCase() + path.substring(1) : title;
	}
	
	private final int index;
	private final Item sourceType;
	private final Item targetType;
	private final RelatedItemHelper helper;
	private final SourceItemTransformerFactory transformerFactory;
	private final SourceItemTransformer transformer;
	
	RelatedItemColumn(int index, String id, Item sourceType, Item targetType, SourceItemTransform direction, Restriction targetRestriction, RelatedItemHelper helper)
	{
		this.index = index;
		this.helper = helper;
		this.sourceType = sourceType;
		this.targetType = targetType;
		this.transformerFactory = ListableUtil.getTransformerFactory(targetType);
		SourceItemTransformer tmp = transformerFactory.create(sourceType, direction);
		this.transformer = targetRestriction == null ? tmp : new RestrictionTransformer(tmp, targetRestriction);
		setId(id);
		setProperty(id);
		setExportable(true);
		setFilterable(true);
	}

	/**
		We finalize this implementation to make sure that the helper 
		implementation always get a chance to re-cycle transactions.
		Subclasses should implement {@link #getValue(RelatedItemHelper, Annotatable)}
		to return the requested value from the item and use the
		DbControl from {@link RelatedItemHelper#dc}.
	*/
	@Override
	public final Object getValue(DbControl dc, Annotatable item) 
	{
		if (helper.lazy) return getProxyTag(item);
		item = helper.recycle(dc, item);
		return getValue(helper, item);
	}

	/**
		Generate a proxy tag that is returned instead of actual data. A
		JavaScript in 'lazy-parent-items.js' will scan the table and
		inject the actual data in batches.
	*/
	private String getProxyTag(Annotatable item)
	{
		StringBuilder sb = new StringBuilder();
		sb.append("<div");
		sb.append(" class=\"lazy-item not-loaded\"");
		sb.append(" id=\"lazy-item-").append(index).append("-").append(item.getId()).append("\"");
		sb.append(" data-item-id=\"").append(item.getId()).append("\"");
		sb.append(" data-index=\"").append(index).append("\"");
		sb.append("></div>");
		return sb.toString();
	}
	
	/**
		We finalize this implementation to make sure that the helper 
		implementation always get a chance to re-cycle transactions.
		Subclasses should implement {@link #getValue(RelatedItemHelper, Annotatable)}
		to return the requested value from the item and use the
		DbControl from {@link RelatedItemHelper#dc}.
	*/
	@Override
	public final Object getExportValue(DbControl dc, Annotatable item) 
	{
		item = helper.recycle(dc, item);
		return getValue(helper, item);
	}

	
	/**
		Alternate implementation for loading data.
	*/
	protected abstract Object getValue(RelatedItemHelper helper, Annotatable item);
	
	/**
		Get all parent or child items to the requested item that matches the specified filter.
	*/
	@SuppressWarnings({ "unchecked", "rawtypes" })
	protected Set<Annotatable> getRelatedItems(Annotatable item)
	{
		Set<Integer> relatedIds = transformer.transform(helper.transformContext, Collections.singleton(item.getId()));
		
		Set<Annotatable> relatedItems = new TreeSet<Annotatable>(new NameableComparator(false));
		for (Integer id : relatedIds)
		{
			Annotatable a = targetType.getById(helper.dc, id);
			relatedItems.add(a);
		}
		return relatedItems;
	}
	
}