/*
	$Id: Icon.java 7703 2019-04-11 07:31:26Z nicklas $

	Copyright (C) 2005 Nicklas Nordborg
	Copyright (C) 2006 Jari Häkkinen, Nicklas Nordborg

	This file is part of BASE - BioArray Software Environment.
	Available at http://base.thep.lu.se/

	BASE is free software; you can redistribute it and/or
	modify it under the terms of the GNU General Public License
	as published by the Free Software Foundation; either version 3
	of the License, or (at your option) any later version.

	BASE is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with BASE. If not, see <http://www.gnu.org/licenses/>.
*/
package net.sf.basedb.clients.web.taglib;

import javax.servlet.jsp.JspException;
import javax.servlet.jsp.JspTagException;

/**
	This tag is used to make it easier to display icons
	on a BASE web page.
	<p>
	<b>Example:</b>
	<pre>
   &lt;%@ taglib prefix="base" uri="/WEB-INF/base.tld" %&gt;
   &lt;base:icon image="required.png" /&gt;
</pre>

	<p>
	<b>Syntax:</b>
	<pre>
   &lt;base:icon
      id=...
      clazz=...
      style=...
      image=...
      tooltip=...
      visible=...
      tabindex=...
   &gt;
</pre>

	<table class="as-grid">
	<caption>Tag attributes</caption>
	<tr>
		<td><b>Attribute</b></td>
		<td><b>Default value</b></td>
		<td><b>Required</b></td>
		<td><b>Description</b></td>
	</tr>
	<tr>
		<td>id</td>
		<td>-</td>
		<td>no</td>
		<td>The ID of the icon. The value of this attribute
		goes directly into the standard <code>id</code> attribute.
		This means that if you need a reference to the note object in your own
		JavaScript code, you may for example use the following code:
		<p>
		<code>
		icon = document.getElementById('&lt;id&gt;');
		</code>
		</td>
	</tr>
	<tr>
		<td>clazz</td>
		<td>icon</td>
		<td>no</td>
		<td>
			The value if this attribute goes directly into the standard HTML 
			<code>class</code> attribute. This allows you to apply different styles
			to different icons.
		</td>
	</tr>
	<tr>
		<td>style</td>
		<td>-</td>
		<td>no</td>
		<td>
			The value if this attribute goes directly into the standard HTML 
			<code>style</code> attribute. This allows you to apply different styles
			to different icons.
		</td>
	</tr>
	<tr>
		<td>image</td>
		<td></td>
		<td>yes</td>
		<td>
			The name of the image to display. The image must be in the <code>/images</code>
			directory. Only include the filename, not the path.
			This taglib will find the correct path from the current jsp page.
		</td>
	</tr>
	<tr>
		<td>tooltip</td>
		<td></td>
		<td>no</td>
		<td>
			A short explanation that will show up as a tooltip.
		</td>
	</tr>
	<tr>
		<td>visible</td>
		<td>true</td>
		<td>no</td>
		<td>
			If the note should be visible or not.
		</td>
	</tr>
	<tr>
		<td>tabindex</td>
		<td>0</td>
		<td>no</td>
		<td>
			Tab index value to control the focus order. Use -1 to disable tabbing to the icon.
		</td>
	</tr>
	</table>

	@author Nicklas
	@version 2.0
*/
public class Icon 
	extends StylableTag
{

	/**
		
	*/
	private static final long serialVersionUID = 6886281373891822996L;

	/**
		The parent &lt;base:page&gt; tag.
	*/
	private Page page;

	/**
		The filename of the image.
	*/
	private String image = null;

	/**
		An optional tooltip.
	*/
	private String tooltip = null;
	
	/**
		If the icon should be visible or not.
	*/
	private boolean visible = true;

	/**
		If present, the icon is enclosed with an &lt;a&gt; tag 
		that opens the given url in a new window.
	 */
	private String href = null;
	
	/**
		If the icon should be enabled or not (if an onclick has been set).
	*/
	private boolean enabled = true;
	
	private int tabIndex;
	
	public Icon()
	{
		super("icon");
	}
	
	/*
		Taglib initialization methods
	*/
	public void setImage(String image)
	{
		this.image = image;
	}
	public String getImage()
	{
		return image;
	}

	public void setVisible(boolean visible)
	{
		this.visible = visible;
	}
	public boolean isVisible()
	{
		return visible;
	}
	
	public void setEnabled(boolean enabled)
	{
		this.enabled = enabled;
	}
	public boolean isEnabled()
	{
		return enabled;
	}

	public void setTooltip(String tooltip)
	{
		this.tooltip = tooltip;
	}
	public String getTooltip()
	{
		return tooltip;
	}

	/**
		@since 3.12
	*/
	public void setHref(String href)
	{
		this.href = href;
	}
	
	/**
		@since 3.12
	*/
	public String getHref()
	{
		return href;
	}
	
	/**
		@since 3.2
	*/
	public void setTabindex(int tabIndex)
	{
		this.tabIndex = tabIndex;
	}
	
	/**
		@since 3.2
	*/
	public int getTabindex()
	{
		return tabIndex;
	}

	@Override
	public int doStartTag()
		throws JspException
	{
		page = (Page)findAncestorWithClass(this, Page.class);
		if (page == null) throw new JspTagException("Tag <base:icon> must be inside a <base:page> tag");
		if (!isVisible()) return SKIP_BODY;

		String theImage = getImage();
		if (!theImage.startsWith("/"))
		{
			theImage = page.getRoot() + "images/" + theImage;
		}

		StringBuilder sb = new StringBuilder();
		if (getHref() != null)
		{
			sb.append("<a href=\"").append(getHref()).append("\" target=\"_blank\" rel=\"noopener noreferrer\">");
		}
		sb.append("<span ");
		addIdAndStyles(sb, getId() != null ? "link" : null, isEnabled() ? null : "disabled");
		if (getTooltip() != null) sb.append(" title=\"").append(getTooltip()).append("\"");
		if (getId() != null)
		{
			// Assume onclick event is attached by javascript
			if (isEnabled() && getTabindex() >= 0) 
			{
				sb.append(" tabindex=\"").append(getTabindex()).append("\"");
			}
		}
		addDynamicAttributes(sb);
		sb.append(">");
		sb.append("<img src=\"").append(theImage).append("\" alt=\"*\">");
		try
		{
			pageContext.getOut().print(sb.toString());
		}
		catch (Exception ex)
		{
			throw new JspTagException(ex.getMessage());
		}
		return EVAL_BODY_INCLUDE;
	}

	@Override
	public int doEndTag() 
		throws JspException
	{
		if (!isVisible()) return EVAL_PAGE;
		StringBuilder sb = new StringBuilder();
		sb.append("</span>");
		if (getHref() != null)
		{
			sb.append("</a>");
		}
		try
		{
			pageContext.getOut().print(sb.toString());
		}
		catch (Exception ex)
		{
			throw new JspTagException(ex.getMessage());
		}
		return EVAL_PAGE;
	}
	
	
	
}

