/*
	$Id: ColumnDef.java 7907 2021-02-12 07:16:27Z nicklas $

	Copyright (C) 2005 Nicklas Nordborg
	Copyright (C) 2006 Jari Häkkinen, Nicklas Nordborg
	Copyright (C) 2007 Nicklas Nordborg

	This file is part of BASE - BioArray Software Environment.
	Available at http://base.thep.lu.se/

	BASE is free software; you can redistribute it and/or
	modify it under the terms of the GNU General Public License
	as published by the Free Software Foundation; either version 3
	of the License, or (at your option) any later version.

	BASE is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with BASE. If not, see <http://www.gnu.org/licenses/>.
*/
package net.sf.basedb.clients.web.taglib.table;

import java.util.Arrays;
import java.util.Collections;
import java.util.Date;
import java.util.HashSet;
import java.util.Iterator;
import java.util.Set;

import net.sf.basedb.util.Enumeration;
import net.sf.basedb.core.ItemContext;
import net.sf.basedb.core.Permission;
import net.sf.basedb.core.PropertyFilter;
import net.sf.basedb.core.Type;
import net.sf.basedb.core.Unit;
import net.sf.basedb.clients.web.Base;
import net.sf.basedb.util.Values;
import net.sf.basedb.util.extensions.ExtensionsInvoker;
import net.sf.basedb.util.formatter.Formatter;
import net.sf.basedb.clients.web.extensions.JspContext;
import net.sf.basedb.clients.web.extensions.list.ListColumnAction;
import net.sf.basedb.clients.web.formatter.FormatterFactory;
import net.sf.basedb.clients.web.taglib.StylableTag;
import net.sf.basedb.clients.web.util.HTML;

import javax.servlet.jsp.JspException;
import javax.servlet.jsp.JspTagException;

import org.json.simple.JSONArray;
import org.json.simple.JSONObject;


/**
	This tag defines a column in the table. It should be placed directly after the
	{@link Table &lt;tbl:table&gt;} tag.

	<p>
	<b>Syntax:</b>
	<pre>
   &lt;tbl:columndef
      id=...
      clazz=...
      style=...
      property=...
      sortproperty=...
      filterproperty=...
      exportproperty=...
      datatype=int|long|float|double|string|date|boolean
      annotation=true|false
      enumeration=...
      smartenum=true|false
      title=...
      sortable=true|false
      filterable=true|false
      exportable=true|false
      show=always|never|auto
      tooltip=...
      formatter=...
      unit=...
      subtitle=...
   &gt;
</pre>

	<table class="as-grid">
	<caption>Tag attributes</caption>
	<tr>
		<td><b>Attribute</b></td>
		<td><b>Default value</b></td>
		<td><b>Required</b></td>
		<td><b>Description</b></td>
	</tr>
	<tr>
		<td>id</td>
		<td>-</td>
		<td>yes</td>
		<td>
			The ID of the column. Should be unique among the columns in the
			table. Do not use <code>all</code> as an ID since it is has a special
			meaning. The value of this attribute
			goes directly into the standard HTML <code>id</code> attribute.
			This means that if you need a reference to the table object in your own
			JavaScript code, you may for example use the following code:
			<pre class="code">
column = document.getElementById('&lt;id&gt;');
</pre>
			The ID is also used for cross-referencing between the different tags.
			For example, <code>&lt;tbl:data columns="..."&gt;</code> and
			<code>&lt;tbl:cell column="..."&gt;</code>.
		</td>
	</tr>
	<tr>
		<td>clazz</td>
		<td>columnheader</td>
		<td>no</td>
		<td>
			The value if this attribute goes directly into the standard HTML 
			<code>class</code> attribute. This allows you to apply different styles
			to different tables.
		</td>
	</tr>
	<tr>
		<td>style</td>
		<td>-</td>
		<td>no</td>
		<td>
			The value if this attribute goes directly into the standard HTML 
			<code>style</code> attribute. This allows you to apply different styles
			to different tables.
		</td>
	</tr>
	<tr>
		<td>property</td>
		<td>-</td>
		<td>no</td>
		<td>
			The name of the property on the item objects of this column. If the property
			isn't a property of the root item, prepend the value with a dollar sign ($)
			to indicate that the root alias should not be used when using this property
			in a query.
		</td>
	</tr>
	<tr>
		<td>sortproperty</td>
		<td>-</td>
		<td>no</td>
		<td>
			The name of the property to sort on when sorting on this column. If
			not specified, it takes the same value as the <code>property</code>
			value.
		</td>
	</tr>
	<tr>
		<td>filterproperty</td>
		<td>-</td>
		<td>no</td>
		<td>
			The name of the property to filter on when filtering on this column. If
			not specified, it takes the same value as the <code>property</code>
			value.
		</td>
	</tr>
	<tr>
		<td>exportproperty</td>
		<td>-</td>
		<td>no</td>
		<td>
			The name of the property to export on when exporting on this column. If
			not specified, it takes the same value as the <code>property</code>
			value.
		</td>
	</tr>
	<tr>
		<td>title</td>
		<td>-</td>
		<td>no</td>
		<td>
			The title of the column. The value is displayed in the header row.
		</td>
	</tr>
	<tr>
		<td>title</td>
		<td>-</td>
		<td>no</td>
		<td>
			As subtitle of the column. The value is displayed in a header row below
			the filter row.
		</td>
	</tr>
	<tr>
		<td>sortable</td>
		<td>false</td>
		<td>no</td>
		<td>
			If the table can be sorted by this column or not. If a true value
			is specified the title will be linked to a function that resubmits the
			page with the new sort order information.
		</td>
	</tr>
	
	<tr>
		<td>filterable</td>
		<td>false</td>
		<td>no</td>
		<td>
			If the table can be filtered by this column or not.
		</td>
	</tr>
	
	<tr>
		<td>exportable</td>
		<td>no</td>
		<td>no</td>
		<td>
			If values in this column can be exported or not.
		</td>
	</tr>

	<tr>
		<td>show</td>
		<td>auto</td>
		<td>no</td>
		<td>
			Specify if this column should be visible or hidden. If the
			value is <code>always</code> the column is always shown, if the
			value is <code>never</code> the column is never shown, if the value
			is <code>auto</code> it is only shown if it appears in the
			<code>columns</code> attribute of the <code>&lt;tbl:data&gt;</code>
			tag.
		</td>
	</tr>
	
	<tr>
		<td>enumeration</td>
		<td>-</td>
		<td>no</td>
		<td>A Enumeration&lt;String, String&gt; object containing the possible
			values of this property. This is used to generate a selection list for
			the filter.
		</td>
	</tr>
	
	<tr>
		<td>smartenum</td>
		<td>false</td>
		<td>no</td>
		<td>
			If the filter for an enumeration column should be displayed as
			a free text field or a selection list. Default is to use a selection
			list. Use this option only if both the hidden key and the displayed
			value of the enumeration is the same. The "smartness" of this options
			is that as the user types text, the browser will dynamically display a list
			of options matching the entered text.
		</td>
	</tr>
	
	<tr>
		<td>multiple</td>
		<td>true</td>
		<td>no</td>
		<td>
			If it should be possible to select multiple options for an 'enumeration' or
			only a single option. The default value is true.
		</td>
	</tr>
	
	<tr>
		<td>formatter</td>
		<td>-</td>
		<td>no</td>
		<td>A {@link Formatter} that formats the values in this column. If no formatter
			is specified the value is ouput as is. Values that should be formated must
			be specified in the <code>&lt;cell value="..."&gt;</code> attribute. Any tag content
			is outputted as entered.
		</td>
	</tr>
	<tr>
		<td>unit</td>
		<td>-</td>
		<td>no</td>
		<td>
			An options {@link Unit} object that should be unit the values are
			stored with in the database. Setting this property enables
			filter expressions with units.
		</td>
	</tr>
	</table>
	<p>
	This tag has no subtags.

	@author Nicklas
	@version 2.0
*/
public class ColumnDef
	extends StylableTag
{

	/**
		
	*/
	private static final long serialVersionUID = -272408513116029365L;

	public ColumnDef()
	{
		super(null);
	}
	
	
	
	/**
		The parent &lt;tbl:table&gt; tag.
	*/
	private Table table;
	
	
	/**
		The name of the property on the item class.
	*/
	private String property = null;
	
	/**
		If the property is an annotation or not.
	*/
	private boolean isAnnotation = false;

	/**
		Optional property used for sorting.
	*/
	private String sortProperty = null;

	/**
		Optional property used for filtering.
	*/
	private String filterProperty = null;
	
	/**
		Optional property used for exporting.
	*/
	private String exportProperty = null;
	
	/**
		The type of data in the property.
	*/
	private Type valueType = null;
	
	/**
		The title of the column.
	*/
	private String title = null;

	/**
		The subtitle of the column.
		@since 3.18
	*/
	private String subtitle = null;

	/**
		If the column should be visible or not. Valid values are: always, never, auto
	*/
	private String show = "auto";
	
	/**
		If the column is sortable or not
	*/
	private boolean sortable = false;
	
	/**
		If the column is filterable or not
	*/
	private boolean filterable = false;
	
	/**
		If the column is exportable or not
	*/
	private boolean exportable = false;
	
	/**
		A tooltip with explaining text.
	*/
	private String tooltip = null;
	
	/**
		Options for enum.
	*/	
	private Enumeration<String, String> enumeration;
	
	private boolean smartEnum;
	
	private boolean multiple = true;
	
	private transient Formatter<?> formatter;
	
	private transient Unit unit;
	
	private JspContext jspContext;
	
	private ExtensionsInvoker<ListColumnAction<?,?>> extensions;
	
	/*
		Taglib initialization methods
	*/
	public void setProperty(String property)
	{
		this.property = property;
	}
	public String getProperty()
	{
		return property;
	}
	
	public void setAnnotation(boolean isAnnotation)
	{
		this.isAnnotation = isAnnotation;
	}
	public boolean isAnnotation()
	{
		return isAnnotation;
	}

	public void setSortproperty(String sortProperty)
	{
		this.sortProperty = sortProperty;
	}
	public String getSortproperty()
	{
		return sortProperty == null ? property : sortProperty;
	}

	public void setFilterproperty(String filterProperty)
	{
		this.filterProperty = filterProperty;
	}
	public String getFilterproperty()
	{
		return filterProperty == null ? property : filterProperty;
	}
	
	public void setExportproperty(String exportProperty)
	{
		this.exportProperty = exportProperty;
	}
	public String getExportproperty()
	{
		return exportProperty == null ? property : exportProperty;
	}

	public void setDatatype(String dataType)
	{
		this.valueType = Type.fromValue(dataType);
	}
	public String getDatatype()
	{
		return valueType == null ? null : valueType.getStringValue();
	}
	
	public void setValueType(Type valueType)
	{
		this.valueType = valueType;
	}
	
	public Type getValueType()
	{
		return valueType;
	}
	
	public void setTitle(String title)
	{
		this.title = title;
	}
	public String getTitle()
	{
		return title;
	}

	/**
		@since 3.18
	*/
	public void setSubtitle(String subtitle)
	{
		this.subtitle = subtitle;
	}
	/**
		@since 3.18
	*/
	public String getSubtitle()
	{
		return subtitle;
	}

	public void setShow(String show)
	{
		this.show = show;
	}
	public String getShow()
	{
		return show;
	}

	public void setSortable(boolean sortable)
	{
		this.sortable = sortable;
	}
	public boolean isSortable()
	{
		return sortable;
	}

	public void setFilterable(boolean filterable)
	{
		this.filterable = filterable;
	}
	public boolean isFilterable()
	{
		return filterable;
	}

	public void setExportable(boolean exportable)
	{
		this.exportable = exportable;
	}
	public boolean isExportable()
	{
		return exportable;
	}

	public void setTooltip(String tooltip)
	{
		this.tooltip = tooltip;
	}
	public String getTooltip()
	{
		return tooltip;
	}

	public void setEnumeration(Enumeration<String, String> enumeration)
	{
		this.enumeration = enumeration;
	}
	public Enumeration<String, String> getEnumeration()
	{
		return enumeration;
	}
	
	/**
		@since 2.8
	*/
	public void setMultiple(boolean multiple)
	{
		this.multiple = multiple;
	}
	
	/**
		@since 2.7
	*/
	public void setSmartenum(boolean smartEnum)
	{
		this.smartEnum = smartEnum;
	}
	
	public void setFormatter(Formatter<?> formatter)
	{
		this.formatter = formatter;
	}
	
	public Formatter<?> getFormatter()
	{
		return formatter;
	}
	
	/**
		@since 2.9
	*/
	public void setUnit(Unit unit)
	{
		this.unit = unit;
	}
	
	/**
		@since 2.9
	*/
	public Unit getUnit()
	{
		return unit;
	}
	
	/**
		@since 3.2
	*/
	public void setJspcontext(JspContext jspContext)
	{
		this.jspContext = jspContext;
	}
	
	/**
		@since 3.2
	*/
	@SuppressWarnings({ "unchecked", "rawtypes" })
	public <I> void setExtensions(ExtensionsInvoker<ListColumnAction<I, ?>> extensions)
	{
		this.extensions = (ExtensionsInvoker)extensions;
	}
	
	@Override
	protected String getIdPrefix()
	{
		return table == null ? "" : table.getId() + ".";
	}
	
	
	/**
		Get a JSON representation of this column definition.
		@since 3.3
	*/
	public JSONObject asJsonObject()
	{
		JSONObject json = new JSONObject();
		json.put("id", getId());
		json.put("title", getTitle());
		json.put("subtitle", getSubtitle());
		json.put("property", getProperty());
		json.put("sortProperty", getSortproperty());
		json.put("filterProperty", getFilterproperty());
		json.put("exportProperty", getExportproperty());
		json.put("dataType", getDatatype());
		json.put("isAnnotation", isAnnotation());
		json.put("alwaysShow", "always".equals(getShow()));
		json.put("alwaysHide", "never".equals(getShow()));
		json.put("sortable", isSortable());
		json.put("filterable", isFilterable());
		json.put("exportable", isExportable());		
		return json;
	}
	
	/*
		Package methods
	*/

	/*
		Display methods
	*/
	@SuppressWarnings({ "unchecked", "rawtypes" })
	@Override
	public int doStartTag()
		throws JspException
	{
		if (!(getParent() instanceof Table)) throw new JspTagException("Tag <tbl:columndef> must be inside a <tbl:table> tag");
		table = (Table)getParent();
		int numFilterRows = table.getFilterrows();
		
		Iterator<ListColumnAction<?,?>> it = null;
		if (extensions != null)
		{
			it = extensions.iterator();
		}
		else
		{
			it = new NullIterator();
		}
		
		while (it.hasNext())
		{
			ListColumnAction<?, ?> column = it.next();
			
			if (column != null)
			{
				setId(column.getId());
				setClazz(column.getClazz());
				setStyle(column.getStyle());
				setTitle(column.getTitle());
				setSubtitle(column.getSubtitle());
				setTooltip(column.getTooltip());
				setProperty(column.getProperty());
				setValueType(column.getValueType());
				setSortable(column.isSortable());
				setSortproperty(column.getSortProperty());
				setFilterable(column.isFilterable());
				setFilterproperty(column.getFilterProperty());
				setExportable(column.isExportable());
				setExportproperty(column.getExportProperty());
				setEnumeration(column.getEnumeration());
				setFormatter(column.getFormatter());
			}
			
			table.addColumnDef(this, column);
	
			ItemContext context = null;
			boolean isPermission = "permission".equals(getId()); // requires special handling
			boolean isFilterable = (isFilterable() || isPermission) && table.getSc() != null &&  table.getItem() != null;
			boolean isSortable = isSortable() && table.getSc() != null &&  table.getItem() != null;
			boolean isVisible = table.isColumnVisible(getId());
			boolean hasFilter = false;
			if (isFilterable)
			{
				context = table.getSc().getCurrentContext(table.getItem(), table.getSubcontext());
				// Check if hidden columns have a filter
				if (!isVisible && context != null)
				{
					if (isPermission)
					{
						hasFilter = context.getItemPermission() != null;
					}
					else if (getFilterproperty() != null)
					{
						hasFilter = context.hasPropertyFilter(getFilterproperty());
					}
				}
			}
			String idPrefix = table.getId()+"."+getId();
			if (isVisible)
			{
				boolean allowColumnDrag = table.allowColumnDrag();
				StringBuilder sb = new StringBuilder();
				sb.append("<th");
				addIdAndStyles(sb, allowColumnDrag ? "table-col-draggable" : null);
				addDynamicAttributes(sb);
				if (allowColumnDrag) 
				{
					sb.append(" draggable=\"true\"");
				}
				sb.append(" data-column-id=\"").append(getId()).append("\"");
				sb.append(">");
				if (allowColumnDrag && !"always".equals(getShow()))
				{
					sb.append("<span class=\"table-col-hide\" title=\"Hide this column\"");
					sb.append(" data-column-id=\"").append(getId()).append("\"");
					sb.append("></span>");
				}
				if (!isSortable)
				{
					String theTooltip = getTooltip() != null ? getTooltip() : getTitle();
					sb.append("<span title=\"").append(theTooltip).append("\">");
					sb.append(getTitle());
					sb.append("</span>");
				}
				else
				{
					// Current sort direction
					ItemContext.SortDirection sortDirection = table.getSortDirection(getSortproperty());
					String image = table.getDirectionImage(sortDirection);
					
					// Next sort direction if clicking on this column
					ItemContext.SortDirection nextDirection = 
							sortDirection == null ? table.getDirection() : sortDirection.getReverseDirection();

					String theTooltip = getTooltip() != null ? 
						getTooltip() : "Sort by "+getTitle() + "; hold CTRL, ALT or SHIFT to sort by multiple columns";
					sb.append("<span class=\"link table-sort\"");
					sb.append(" id=\"").append(idPrefix).append(".sort\"");
					sb.append(" data-sort-property=\"").append(HTML.encodeTags(getSortproperty())).append("\"");
					sb.append(" data-sort-direction=\"").append(nextDirection.name()).append("\"");
					sb.append(" title=\"").append(theTooltip).append("\">");
					sb.append(getTitle()).append("</span>").append(image);
				}
				sb.append("</th>\n");
				table.setColumnContent(getId(), sb.toString());
				
				for (int filterNo = 0; filterNo < numFilterRows; filterNo++)
				{
					PropertyFilter filter = null;
					if (isFilterable && context != null && getFilterproperty() != null)
					{
						filter = context.getPropertyFilter(getFilterproperty(), filterNo);
						if (filter != null) hasFilter = true;
					}
					
					sb = new StringBuilder();
					sb.append("<th class=\"propertyfilter\">");
					if (isPermission)
					{
						if (filterNo == 0)
						{
							// Special case for permission filtering
							sb.append("<select name=\"permission\" class=\"table-submit\">\n");
							sb.append("<option value=\"\">");
							Permission permission = context.getItemPermission();
							Set<Permission> definedPermission = context.getItemType().getDefinedPermissions();
							hasFilter = permission != null;
							appendPermissionOption(sb, permission, Permission.READ, "Read [R]");
							if (definedPermission != null)
							{
								if (definedPermission.contains(Permission.USE))
								{
									appendPermissionOption(sb, permission, Permission.USE, "Use [U]");
								}
								if (definedPermission.contains(Permission.RESTRICTED_WRITE))
								{
									appendPermissionOption(sb, permission, Permission.RESTRICTED_WRITE, "Annotate [A]");
								}
								if (definedPermission.contains(Permission.WRITE))
								{
									appendPermissionOption(sb, permission, Permission.WRITE, "Write [W]");
								}
								if (definedPermission.contains(Permission.DELETE))
								{
									appendPermissionOption(sb, permission, Permission.DELETE, "Delete [D]");
								}
								if (definedPermission.contains(Permission.SET_OWNER))
								{
									appendPermissionOption(sb, permission, Permission.SET_OWNER, "Change owner [O]");
								}
								if (definedPermission.contains(Permission.SET_PERMISSION))
								{
									appendPermissionOption(sb, permission, Permission.SET_PERMISSION, "Change permissions [P]");
								}
							}
							sb.append("</select>");
						}
					}
					else if (isFilterable)
					{
						Type valueType = getValueType();
						Formatter<Date> dateFormatter = FormatterFactory.getDateFormatter(table.getSc());
		
						String inputName = "filter:" + valueType.name() + ":";
						if (filter != null && filter.isTemporary()) inputName = "tmp" + inputName;
						if (unit != null) inputName += unit.getId() + ":";
						inputName += HTML.encodeTags(getFilterproperty());
						if (filterNo > 0) inputName += "["+filterNo+"]";
						if (getEnumeration() != null)
						{
							String filterValue = filter == null ? "" : HTML.encodeTags(Base.getPropertyFilterString(filter, dateFormatter));
							if (smartEnum)
							{
								JSONArray options = new JSONArray();
								for (int i = 0; i < enumeration.size(); ++i)
								{
									options.add(enumeration.getKey(i));
								}
								
								sb.append("<div class=\"smartinput auto-init\" data-auto-init=\"smartenum\"");
								sb.append(" data-options=\"").append(HTML.encodeTags(options.toJSONString())).append("\"");
								sb.append(">");
								sb.append("<table><tr><td>");
								sb.append("<input class=\"text smartenum-display\" type=\"text\"");
								sb.append(" name=\"").append(inputName).append("\"");
								sb.append(" maxlength=\"").append(PropertyFilter.MAX_VALUE_LENGTH).append("\"");
								sb.append(" value=\"").append(filterValue).append("\"");
								sb.append(">");
								sb.append("</td><td class=\"dropdown\">");
								sb.append("<img src=\"").append(table.getPage().getRoot()).append("images/mini_scroll_down.png\" alt=\"▼\">");
								sb.append("</td></tr></table></div>");
							}
							else
							{
								Set<String> selected = Collections.emptySet();
								String displayName = "display:" + getFilterproperty();
								JSONArray options = new JSONArray();
								if (filter != null)
								{
									if (filter.getValue() != null)
									{
										selected = new HashSet<String>(Arrays.asList(filter.getValue().split("\\|")));
									}
									else
									{
										selected = Collections.singleton("");
									}
								}
								for (int i = 0; i < enumeration.size(); ++i)
								{
									String key = enumeration.getKey(i);
									JSONObject option = new JSONObject();
									option.put("key", key);
									option.put("value", enumeration.getValue(i));
									option.put("checked", selected.contains(key));
									options.add(option);
								}
								
								sb.append("<div class=\"multiselect auto-init\" data-auto-init=\"multiselect\"");
								sb.append(" data-multiple=\"").append(multiple ? "1" : "0").append("\"");
								sb.append(" data-options=\"").append(HTML.encodeTags(options.toJSONString())).append("\"");
								sb.append(">");
								sb.append("<table><tr><td>");
								sb.append("<input type=\"hidden\" class=\"multiselect-hidden\"");
								sb.append(" name=\"").append(inputName).append("\"");
								sb.append(" value=\"").append(filterValue).append("\"");
								sb.append(">");
								sb.append("<input type=\"text\" class=\"text multiselect-display\"");
								sb.append(" name=\"").append(displayName).append("\"");
								sb.append(">");
								sb.append("</td><td class=\"dropdown\">");
								sb.append("<img src=\"").append(table.getPage().getRoot()).append("images/mini_scroll_down.png\" alt=\"▼\">");
								sb.append("</td></tr></table></div>");
							}
						}
						else if (valueType == Type.BOOLEAN)
						{
							boolean filterValue = filter == null ? false : Values.getBoolean(filter.getValue());
							sb.append("<input type=\"radio\" name=\"").append(inputName).append("\" id=\"").append(inputName).append(".any\"");
							sb.append(filter == null ? " checked" : "");
							sb.append(" value=\"\" class=\"table-submit\">").append("<label for=\"").append(inputName).append(".any\">any</label>");
							
							sb.append("<input type=\"radio\" name=\"").append(inputName).append("\" id=\"").append(inputName).append(".true\"");
							sb.append(filter != null && filterValue ? " checked" : "");
							sb.append(" value=\"true\" class=\"table-submit\">").append("<label for=\"").append(inputName).append(".true\">true</label>");
							
							sb.append("<input type=\"radio\" name=\"").append(inputName).append("\" id=\"").append(inputName).append(".false\"");
							sb.append(filter != null && !filterValue ? " checked" : "");
							sb.append(" value=\"false\" class=\"table-submit\">").append("<label for=\"").append(inputName).append(".false\">false</label>");
						}
						else
						{
							String filterValue = HTML.encodeTags(Base.getPropertyFilterString(filter, dateFormatter));
							sb.append("<input class=\"text table-submit\" type=\"text\"");
							sb.append(" name=\"").append(inputName).append("\"");
							sb.append(" maxlength=\"").append(PropertyFilter.MAX_VALUE_LENGTH).append("\"");
							sb.append(" value=\"").append(filterValue).append("\"");
							sb.append(">");
						}
					}
					else
					{
						sb.append("&nbsp;");	
					}
					sb.append("</th>");
					table.setColumnFilter(getId(), filterNo, sb.toString());
				}
			}
			if (isFilterable && hasFilter) 
			{
				table.addFilteredColumn(getId(), isVisible);
			}
		}
		return SKIP_BODY;
	}

	private void appendPermissionOption(StringBuilder sb, Permission selected, Permission toAdd, String display)
	{
		sb.append("<option value=\"").append(toAdd.name()).append("\"");
		sb.append(selected == toAdd ? " selected" : "").append(">").append(display).append("\n");
	}
	
	/**
		Iterator implementation that returns a single null value and nothing more.
	*/
	static class NullIterator<A, B>
		implements Iterator<ListColumnAction<A, B>>
	{

		private boolean hasNext = true;
		@Override
		public boolean hasNext() 
		{
			return hasNext;
		}
		@Override
		public ListColumnAction<A, B> next() 
		{
			hasNext = false;
			return null;
		}
	}
	
}


