/*
	$Id: Row.java 7884 2020-12-03 12:43:31Z nicklas $

	Copyright (C) 2005 Nicklas Nordborg
	Copyright (C) 2006 Jari Häkkinen, Nicklas Nordborg

	This file is part of BASE - BioArray Software Environment.
	Available at http://base.thep.lu.se/

	BASE is free software; you can redistribute it and/or
	modify it under the terms of the GNU General Public License
	as published by the Free Software Foundation; either version 3
	of the License, or (at your option) any later version.

	BASE is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with BASE. If not, see <http://www.gnu.org/licenses/>.
*/
package net.sf.basedb.clients.web.taglib.table;

import javax.servlet.jsp.JspException;
import javax.servlet.jsp.JspTagException;

import net.sf.basedb.clients.web.taglib.StylableTag;

/**
	This tag defines a new row in the table. It should be placed inside a
	{@link Rows &lt;tbl:rows&gt;} tag.

	<p>
	<b>Syntax:</b>
	<pre>
   &lt;tbl:row
      id=...
      clazz=...
      style=...
      highlight=...
      visible=true|false
   &gt;
</pre>

	<table class="as-grid">
	<caption>Tag attributes</caption>
	<tr>
		<td><b>Attribute</b></td>
		<td><b>Default value</b></td>
		<td><b>Required</b></td>
		<td><b>Description</b></td>
	</tr>
	<tr>
		<td>id</td>
		<td>-</td>
		<td>no</td>
		<td>
			The ID of the row. The value of this attribute
			goes directly into the standard HTML <code>id</code> attribute.
			This means that if you need a reference to the table object in your own
			JavaScript code, you may for example use the following code:
			<pre class="code">
row = document.getElementById('&lt;id&gt;');
</pre>
		</td>
	</tr>
	<tr>
		<td>clazz</td>
		<td>-</td>
		<td>no</td>
		<td>
			The value if this attribute goes directly into the standard HTML 
			<code>class</code> attribute. This allows you to apply different styles
			to different tables.
		</td>
	</tr>
	<tr>
		<td>style</td>
		<td>-</td>
		<td>no</td>
		<td>
			The value if this attribute goes directly into the standard HTML 
			<code>style</code> attribute. This allows you to apply different styles
			to different tables.
		</td>
	</tr>
	<tr>
		<td>highlight</td>
		<td>highlight</td>
		<td>no</td>
		<td>
			Optionall class name that is applied when the mouse is over the current row.
		</td>
	</tr>
	<tr>
		<td>visible</td>
		<td>true</td>
		<td>no</td>
		<td>
			If the row should be visible or not.
		</td>
	</tr>
	</table>
	<p>
	The <code>&lt;tbl:row&gt;</code> may contain the following subtags:
	<ul>
	<li>{@link Cell &lt;tbl:cell&gt;}: Creates a new cell in the row
	</ul>

	@author Nicklas
	@version 2.0
*/
public class Row
	extends StylableTag
{

	/**
		
	*/
	private static final long serialVersionUID = 1975591924471893645L;

	public Row()
	{
		super(null);
	}
	
	/**
		The parent &lt;tbl:table&gt; tag.
	*/
	private Table table;
	
	/**
		The parent &lt;tbl:data&gt; tag.
	*/
	private Data data;

	/**
		Optional <code>class</code> attribute that is applied when the mouse is over the current row.
	*/
	private String highlight = "highlight";
	
	/**
		If the row should be visible or not.
	*/
	private boolean visible = true;

	/*
		Taglib initialization methods
	*/
	/**
		@deprecated In 3.18 due to type in the method name
	*/
	@Deprecated
	public void setHightligh(String highlight)
	{
		this.highlight = highlight;
	}
	/**
		@since 3.18 due to a typo in the method name
	*/
	public void setHighlight(String highlight)
	{
		this.highlight = highlight;
	}
	public String getHighlight()
	{
		return highlight;
	}
	
	public void setVisible(boolean visible)
	{
		this.visible = visible;
	}
	public boolean isVisible()
	{
		return visible;
	}
	/*
		Package methods
	*/
	Table getTable()
	{
		return table;
	}

	/*
		Display methods
	*/
	@Override
	public int doStartTag()
		throws JspException
	{
		if (!(getParent() instanceof Rows)) throw new JspTagException("Tag <tbl:row> must be inside a <tbl:rows> tag");
		table = ((Rows)getParent()).getTable();
		if (!isVisible()) return SKIP_BODY;

		StringBuilder sb = new StringBuilder();
		sb.append("<tr");
		addIdAndStyles(sb, table.getNextRowClass(), getHighlight());
		addDynamicAttributes(sb);
		sb.append(">\n");
		try
		{
			pageContext.getOut().print(sb.toString());
		}
		catch (Exception ex)
		{
			throw new JspTagException(ex.getMessage());
		}
		return EVAL_BODY_INCLUDE;
	}

	@Override
	public int doEndTag()
		throws JspException
	{
		if (!isVisible()) return EVAL_PAGE;
		StringBuilder sb = new StringBuilder();
		sb.append("</tr>\n");
		try
		{
			table.writeColumns();
			pageContext.getOut().print(sb.toString());
		}
		catch (Exception ex)
		{
			throw new JspTagException(ex.getMessage());
		}
		return EVAL_PAGE;
	}


}

