/*
	$Id: FileSet.java 7605 2019-02-26 10:10:15Z nicklas $

	Copyright (C) 2007 Nicklas Nordborg

	This file is part of BASE - BioArray Software Environment.
	Available at http://base.thep.lu.se/

	BASE is free software; you can redistribute it and/or
	modify it under the terms of the GNU General Public License
	as published by the Free Software Foundation; either version 3
	of the License, or (at your option) any later version.

	BASE is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with BASE. If not, see <http://www.gnu.org/licenses/>.
*/
package net.sf.basedb.core;

import java.util.ArrayList;
import java.util.Collections;
import java.util.HashSet;
import java.util.Iterator;
import java.util.LinkedList;
import java.util.List;
import java.util.Set;

import net.sf.basedb.core.query.EntityQuery;
import net.sf.basedb.core.query.Expressions;
import net.sf.basedb.core.query.Hql;
import net.sf.basedb.core.query.Restrictions;
import net.sf.basedb.core.data.BasicData;
import net.sf.basedb.core.data.FileSetData;
import net.sf.basedb.core.data.FileSetMemberData;
import net.sf.basedb.core.hibernate.TypeWrapper;
import net.sf.basedb.util.extensions.ClientContext;
import net.sf.basedb.util.extensions.ExtensionsInvoker;
import net.sf.basedb.util.extensions.Registry;
import net.sf.basedb.util.extensions.manager.ExtensionsManager;
import net.sf.basedb.util.extensions.manager.Settings;
import net.sf.basedb.util.fileset.ResetMetadataRenderer;
import net.sf.basedb.util.fileset.ValidatingFileSetMember;
import net.sf.basedb.util.fileset.ValidationAction;
import net.sf.basedb.util.fileset.ValidationRenderer;

/**
	A file set is the container for files that have been attach to an
	item to hold data instead of importing it to the database.

	@author Nicklas
	@version 2.5
	@base.modified $Date: 2019-02-26 11:10:15 +0100 (Tue, 26 Feb 2019) $
*/
public class FileSet
	extends BasicItem
{
	/**
		The type of item represented by this class.
		@see Item#FILESET
		@see #getType()
	*/
	public static final Item TYPE = Item.FILESET;
	
	/**
		This filter gives everybody read permission to file sets.
	*/
	private static final QueryRuntimeFilter RUNTIME_FILTER = new QueryRuntimeFilterImpl();

	/**
		Create a new file set for the specified item.
	*/
	static FileSet getNew(DbControl dc, FileStoreEnabled item)
		throws BaseException
	{	
		FileSet fs = dc.newItem(FileSet.class, item);		
		fs.getData().setItemType(item.getType().getValue());		
		return fs;
	}
	
	/**
		Get a <code>FileSet</code> item when you know the id.
	
		@param dc The <code>DbControl</code> which will be used for
			permission checking and database access.
		@param id The id of the item to load
		@return The <code>FileSet</code> item
		@throws ItemNotFoundException If an item with the specified 
			id is not found
		@throws PermissionDeniedException If the logged in user doesn't 
			have read permission to the item
		@throws BaseException If there is another error
	*/
	public static FileSet getById(DbControl dc, int id)
		throws ItemNotFoundException, PermissionDeniedException, BaseException
	{
		FileSet fs = dc.loadItem(FileSet.class, id);
		if (fs == null) throw new ItemNotFoundException("FileSet[id="+id+"]");
		return fs;
	}
	
	/**
		Get a {@link ItemQuery} object configured to retrieve <code>FileSet</code> 
		items. This query may return items which the logged in user doesn't have 
		read permission to. At the moment there is no way to solve this problem.
		@return An {@link ItemQuery} object
	*/
	static ItemQuery<FileSet> getQuery()
	{
		return new ItemQuery<FileSet>(FileSet.class, RUNTIME_FILTER);
	}

	/**
		The item this file set belongs to.
	*/
	private FileStoreEnabled item;
	
	/**
		Regular constructor if the item is not known beforehand.
	*/
	FileSet(FileSetData data)
	{
		super(data);		
	}

	/**
		Special constructor to avoid {@link #getItem()} to hit the database
		when the item is known beforehand.
	*/
	FileSet(FileSetData data, FileStoreEnabled item)
	{
		super(data);
		this.item = item;		
	}

	@Override
	FileSetData getData()
	{
		return (FileSetData)getBasicData();
	}
	/*
		From the Identifiable interface
		-------------------------------------------
	*/
	@Override
	public Item getType()
	{
		return TYPE;
	}
	// -------------------------------------------

	/*
		From the BasicItem class
		-------------------------------------------
	*/
	/**
		Always return TRUE. A file set always has a one-to-one relation
		to some item. The file set is automatically deleted if that
		item is deleted.
	*/
	@Override
	public boolean isUsed()
		throws BaseException
	{
		return true;
	}
	
	/**
		Get the item this file set belongs to.
	*/
	@Override
	public Set<ItemProxy> getUsingItems()
		throws BaseException
	{
		Set<ItemProxy> using = super.getUsingItems();
		FileStoreEnabled item = getItem();
		using.add(new ItemProxy(item.getId(), item.getType()));
		return using;
	}
	
	/**
		READ permission is granted to all users. CREATE, WRITE and DELETE permissions are granted
		if the logged in user has WRITE permission on the associated item.
	*/
	@Override
	void initPermissions(int granted, int denied)
		throws BaseException
	{	
		FileStoreEnabled item = null;
		try
		{
			item = getItem();
		}
		catch (PermissionDeniedException ex)
		{}
		if (item != null && item.hasPermission(Permission.WRITE))
		{
			granted |= Permission.grant(Permission.CREATE, Permission.READ, Permission.WRITE, Permission.DELETE);				
		}
		else if (item != null && item.hasPermission(Permission.USE))
		{
			granted |= Permission.grant(Permission.USE);
		}
		else if (item != null && item.hasPermission(Permission.READ))
		{
			granted |= Permission.grant(Permission.READ);				
		}	
		super.initPermissions(granted, denied);
	}
	@Override
	PluginPermission getPluginPermissions()
	{
		return PluginPermission.NO_PLUGIN;
	}
	// -------------------------------------------

	/**
		Get the item this file set belongs to.
		@return The <code>FileStoreEnabled</code> item
		@throws PermissionDeniedException If the logged in user doesn't have 
			read permission to the item
		@throws BaseException If there is another error
	*/
	public FileStoreEnabled getItem()
		throws PermissionDeniedException, BaseException
	{
		if (item == null)
		{	
			Item type = getItemType();
			DbControl dc = getDbControl();
			String hql = "SELECT item"+
				" FROM "+type.getDataClass().getName()+" item"+
				" WHERE item.fileSet = :fileSet";
			org.hibernate.query.Query<? extends BasicData> query = 
				HibernateUtil.createQuery(dc.getHibernateSession(), hql, type.getDataClass());
			query.setParameter("fileSet", getId(), TypeWrapper.H_INTEGER);
			item = (FileStoreEnabled)dc.getItem(type.getItemClass(), HibernateUtil.loadData(query));
		}
		return item;
	}
	
	/**
		Set the item this file set belongs to.
	*/
	void setItem(FileStoreEnabled item)
	{
		this.item = item;
	}
	
	/**
		Get the {@link Item} type of the item this file set belongs to.
	*/
	public Item getItemType()
	{
		return Item.fromValue(getData().getItemType());
	}
	
	/**
		Get the members of this file set.
		@return A query returning <code>FileSetMember</code> items
	*/
	public ItemQuery<FileSetMember> getMembers()
	{
		return FileSetMember.getQuery(this);
	}
	
	/**
		Get a query that returns all members in a file set with a specific 
		generic {@link ItemSubtype}. Note! The query checks the generic type from the 
		{@link DataFileType#getGenericType()}, not the {@link File#getItemSubtype()}.
		
		@param genericType The system ID of a generic type as defined
			by {@link ItemSubtype#getSystemId()}
		@return A query returning <code>FileSetMember</code> items
	*/
	public ItemQuery<FileSetMember> getMembers(String genericType)
	{
		ItemQuery<FileSetMember> query = getMembers();
		query.restrictPermanent(
			Restrictions.eq(
				Hql.property("dataFileType.genericType.systemId"), 
				Expressions.string(genericType)
			)
		);
		return query;
	}
	
	/**
		Get a query that returns all members in a file set with a specific 
		{@link DataFileType}.
		
		@param type The type of the file
		@return A query returning <code>FileSetMember</code> items
		@since 3.0
	*/
	public ItemQuery<FileSetMember> getMembers(DataFileType type)
	{
		ItemQuery<FileSetMember> query = getMembers();
		if (type != null)
		{
			query.restrictPermanent(
				Restrictions.eq(
					Hql.property("dataFileType"), 
					Hql.entity(type)
				)
			);
		}
		return query;
	}

	
	/**
		Set a file of a given type as a member of this file set.
		If one ore more files of the given type already exists, they are
		replaced with this file. If no member exists, a new entry is created 
		and automatically saved to the database when {@link DbControl#commit()}
		is called. Calling this method is equivalent to first call
		{@link #removeAllOfType(DataFileType)} and then {@link 
		#addMember(File, DataFileType)} except that if the given file
		already is a member that member entry is kept.
		
		@param file The file to set as a member (or null to remove it)
		@param type The type of the file
		@return The new FileSetMember item
		@throws InvalidUseOfNullException If file or type is null
		@throws PermissionDeniedException If the logged in user doesn't
			have write permission for this file set or use permission for the
			file
	*/
	public FileSetMember setMember(File file, DataFileType type)
	{
		// Permission checks must be done first so we don't disrupt internal state
		checkPermission(Permission.WRITE);
		if (type == null) throw new InvalidUseOfNullException("type");
		if (type.getItemType() != getItemType())
		{
			throw new InvalidDataException("File type '" + type.getName() + 
				"' can't be used on " + getItemType());
		}
		if (file != null) file.checkPermission(Permission.USE);

		FileSetMember member = removeAllOfTypeExcept(type, file);
		if (member == null && file != null)
		{
			member = addMember(file, type);
		}
		return member;
	}
	
	/**
		Add a file as a member to the file set.
		
		@param file The file to add
		@param type The type of the file
		@return The new FileSetMember item
		@throws InvalidUseOfNullException If file or type is null
		@throws PermissionDeniedException If the logged in user doesn't
			have write permission for this file set or use permission for the
			file
		@since 3.0
	*/
	public FileSetMember addMember(File file, DataFileType type)
	{
		checkPermission(Permission.WRITE);
		if (file == null) throw new InvalidUseOfNullException("file");
		if (type == null) throw new InvalidUseOfNullException("type");
		if (type.getItemType() != getItemType())
		{
			throw new InvalidDataException("File type '" + type.getName() + 
				"' can't be used on " + getItemType());
		}
		file.checkPermission(Permission.USE);

		// Create new member entry
		FileSetMember member = FileSetMember.getNew(getDbControl(), this, file, type);
		getData().getMembers().add(member.getData());
		getDbControl().saveItemIf(this, member, false);
		return member;
	}
	
	/**
		Check if the file set contains at least one file of the specified type.
		@param type A <code>DataFileType</code> object
		@return TRUE if the file set contains a file of 
			the specified type, FALSE otherwise
	*/
	public boolean hasMember(DataFileType type)
	{
		if (type == null) return false;
		int typeId = type.getId();
		for (FileSetMemberData member : getData().getMembers())
		{
			if (member.getDataFileType().getId() == typeId) return true;
		}
		return false;
	}
	
	/**
		Get the number of members in the file set of the specified type.
		@param type A <code>DataFileType</code> object
		@return The number of files in the file set with the given type
	 */
	public int getNumMembers(DataFileType type)
	{
		if (type == null) return 0;
		int typeId = type.getId();
		int numMembers = 0;
		for (FileSetMemberData member : getData().getMembers())
		{
			if (member.getDataFileType().getId() == typeId) numMembers++;
		}
		return numMembers;	
	}

	/**
		Get the member of the given type assuming that there
		is only a single member of that type. If there is
		more than one member of the given type, it is undefined
		which file that is returned.
		
		@param type A <code>DataFileType</code> object 
		@return A FileSetMember object or null if no member
			of the specified type exists
	*/
	public FileSetMember getMember(DataFileType type)
	{
		if (type == null) return null;
		int typeId = type.getId();
		for (FileSetMemberData member : getData().getMembers())
		{
			if (member.getDataFileType().getId() == typeId)
			{
				return getDbControl().getItem(FileSetMember.class, member);
			}
		}
		return null;
	}
	
	/**
		Remove the given member from this file set.
		@param member The member to remove
		@since 3.0
	*/
	public void removeMember(FileSetMember member)
	{
		checkPermission(Permission.WRITE);
		if (member == null) throw new InvalidUseOfNullException("member");
		if (getData().getMembers().remove(member.getData()))
		{
			DbControl dc = getDbControl();
			dc.deleteItem(member);
			// Create context and invoke the extensions system
			ExtensionsInvoker<ValidationAction> invoker = getInvoker(dc);
			
			// Reset the metadata
			invoker.render(new ResetMetadataRenderer(Collections.singletonList(new ValidatingFileSetMember(member))));
		}
	}
	
	
	/**
		Remove a file as a member from the file set.
		
		@param file The file to remove
		@param type The type of the file
		@return TRUE if a member was removed, FALSE if not
		@throws InvalidUseOfNullException If file or type is null
		@throws PermissionDeniedException If the logged in user doesn't
			have write permission for this file set 
		@since 3.0
	*/
	public boolean removeMember(File file, DataFileType type)
	{
		checkPermission(Permission.WRITE);
		if (file == null) throw new InvalidUseOfNullException("file");
		if (type == null) throw new InvalidUseOfNullException("type");
			
		DbControl dc = getDbControl();
		int typeId = type.getId();
		int fileId = file.getId();
		Iterator<FileSetMemberData> it = getData().getMembers().iterator();
		// Keep track of removed members so that we can invoke metadata extensions
		List<ValidatingFileSetMember> removed = new ArrayList<ValidatingFileSetMember>();
		while (it.hasNext())
		{
			FileSetMemberData member = it.next();
			if (member.getDataFileType().getId() == typeId && member.getFile().getId() == fileId)
			{
				it.remove();
				if (member.getId() > 0)
				{
					FileSetMember m = dc.getItem(FileSetMember.class, member);
					dc.deleteItem(m);
					removed.add(new ValidatingFileSetMember(m));
				}
			}
		}
		if (removed.size() > 0)
		{
			// Create context and invoke the extensions system
			ExtensionsInvoker<ValidationAction> invoker = getInvoker(dc);
			
			// Reset the metadata
			invoker.render(new ResetMetadataRenderer(removed));
		}
		return removed.size() > 0;
	}

	
	/**
		Remove all member files of the given type. 

		@param type A <code>DataFileType</code> object 
		@throws InvalidUseOfNullException If type is null
		@throws PermissionDeniedException If the logged in user
			doesn't have write permission for this file set
		@return The number of removed files
		@since 3.0
	*/
	public int removeAllOfType(DataFileType type)
	{
		checkPermission(Permission.WRITE);
		if (type == null) throw new InvalidUseOfNullException("type");
		DbControl dc = getDbControl();
		int typeId = type.getId();
		Iterator<FileSetMemberData> it = getData().getMembers().iterator();
		// Keep track of removed members so that we can invoke metadata extensions
		List<ValidatingFileSetMember> removed = new ArrayList<ValidatingFileSetMember>();
		while (it.hasNext())
		{
			FileSetMemberData member = it.next();
			if (member.getDataFileType().getId() == typeId)
			{
				it.remove();
				if (member.getId() > 0)
				{
					FileSetMember m = dc.getItem(FileSetMember.class, member);
					dc.deleteItem(m);
					removed.add(new ValidatingFileSetMember(m));
				}
			}
		}
		if (removed.size() > 0)
		{
			// Create context and invoke the extensions system
			ExtensionsInvoker<ValidationAction> invoker = getInvoker(dc);
			
			// Reset the metadata
			invoker.render(new ResetMetadataRenderer(removed));
		}
		return removed.size();
	}
	
	/**
		Remove all member files of the given type. 
	
		@param type A <code>DataFileType</code> object 
		@throws InvalidUseOfNullException If type is null
		@throws PermissionDeniedException If the logged in user
			doesn't have write permission for this file set
		@return The number of removed files
		@since 3.0
	*/
	private FileSetMember removeAllOfTypeExcept(DataFileType type, File file)
	{
		DbControl dc = getDbControl();
		int typeId = type.getId();
		int fileId = file == null ? 0 : file.getId();
		FileSetMember remain = null;
		Iterator<FileSetMemberData> it = getData().getMembers().iterator();
		// Keep track of removed members so that we can invoke metadata extensions
		List<ValidatingFileSetMember> removed = new ArrayList<ValidatingFileSetMember>();
		while (it.hasNext())
		{
			FileSetMemberData member = it.next();
			if (member.getDataFileType().getId() == typeId)
			{
				// Type is matching -- what about the file?
				if (member.getFile().getId() == fileId)
				{
					// Keep
					remain = dc.getItem(FileSetMember.class, member);
				}
				else
				{
					// Remove
					it.remove();
					if (member.getId() > 0)
					{
						FileSetMember m = dc.getItem(FileSetMember.class, member);
						dc.deleteItem(m);
						removed.add(new ValidatingFileSetMember(m));
					}
				}
			}
		}
		if (removed.size() > 0)
		{
			// Create context and invoke the extensions system
			ExtensionsInvoker<ValidationAction> invoker = getInvoker(dc);
			
			// Reset the metadata
			invoker.render(new ResetMetadataRenderer(removed));
		}
		return remain;
	}

	
	/**
		Remove all members in a file set.
		@throws PermissionDeniedException If the logged in user
			doesn't have write permission for this file set
	*/
	public void removeAllMembers()
	{
		checkPermission(Permission.WRITE);
		
		// Load all members of the file set
		DbControl dc = getDbControl();
		List<ValidatingFileSetMember> members = new ArrayList<ValidatingFileSetMember>(getData().getMembers().size());
		for (FileSetMemberData memberData : getData().getMembers())
		{
			FileSetMember member = dc.getItem(FileSetMember.class, memberData);
			members.add(new ValidatingFileSetMember(member));
		}
		
		// Clear the file set
		getData().getMembers().clear();
		
		// Create context and invoke the extensions system
		ExtensionsInvoker<ValidationAction> invoker = getInvoker(dc);
		
		// Reset the metadata
		invoker.render(new ResetMetadataRenderer(members));
	}
	
	/**
		Get a query that return the {@link DataFileType} for all members
		of this file set.
		@return A query object
	*/
	public ItemQuery<DataFileType> getMemberTypes()
	{
		ItemQuery<DataFileType> query = DataFileType.getQuery();
		query.joinPermanent(Hql.innerJoin("fileSetMembers", Item.FILESETMEMBER.getAlias()));
		query.restrictPermanent(
			Restrictions.eq(
				Hql.property(Item.FILESETMEMBER.getAlias(), "fileSet"), 
				Hql.entity(this)
			)
		);
		query.setDistinct(true);
		return query;
	}
	
	/**
		Checks if this file set contains a file for
		each {@link DataFileType} that is required according to
		what has been specified by the {@link Platform} and
		{@link PlatformVariant}.
		@return TRUE if all files are present, FALSE otherwise. If the
			items has no platform TRUE is always returned
		@see PlatformFileType#isRequired()
	*/
	public boolean hasAllRequiredFiles()
	{
		FileStoreEnabled item = getItem();
		Platform platform = item.getPlatform();
		PlatformVariant variant = item.getVariant();
		if (platform == null) return true;

		// Load all required file types
		ItemQuery<PlatformFileType> query = platform.getFileTypes(variant, variant == null);
		query.restrict(
			Restrictions.eq(
				Hql.property("required"), 
				Expressions.parameter("isRequired", true, Type.BOOLEAN)
			)
		);
		List<PlatformFileType> allRequired = query.list(getDbControl());
		if (allRequired.size() == 0) return true;
		
		// Load ID of all member types
		Set<Integer> memberTypes = new HashSet<Integer>();
		for (FileSetMemberData member : getData().getMembers())
		{
			memberTypes.add(member.getDataFileType().getId());
		}
		
		// Check each required type
		for (PlatformFileType pft : allRequired)
		{
			if (!memberTypes.contains(pft.getData().getDataFileType().getId()))
			{
				return false;
			}
		}
		return true;
	}
	
	/**
		Get a query that is returning all file types that
		have been marked as required by the associated platform but has no member
		in this file set.
		@return A query or null if no platform is associated with this
			item
	*/
	public ItemQuery<PlatformFileType> getMissingDataFileTypes()
	{
		FileStoreEnabled item = getItem();
		Platform platform = item.getPlatform();
		PlatformVariant variant = item.getVariant();
		if (platform == null) return null;

		ItemQuery<PlatformFileType> query = platform.getFileTypes(variant, variant == null);
		// Get all required platform file types
		query.restrict(
			Restrictions.eq(
				Hql.property("required"), 
				Expressions.parameter("isRequired", true, Type.BOOLEAN)
			)
		);

		Set<Integer> memberIds = new HashSet<Integer>();
		for (FileSetMemberData member : getData().getMembers())
		{
			memberIds.add(member.getDataFileType().getId());
		}
		// Ignore all platform file types that we already have
		query.restrictPermanent(
			Restrictions.not(
				Restrictions.in(
					Hql.property("dataFileType.id"), 
					Expressions.parameter("members")
				)
			)
		);
		query.setPermanentParameter("members", memberIds, null);
		return query;
	}
	
	/**
		Validate the files in this file set. This method will invoke extensions
		for the "net.sf.basedb.core.filehandler.validator" extension point and
		let them handle the actual validation.
		
		This method will not throw any errors that have been caused by the validation 
		process. Errors are returned as a list of {@link Throwable}:s and is also 
		reported by setting the {@link FileSetMember#isValid()} and 
		{@link FileSetMember#getErrorMessage()} properties for each file.
		<p>
		Note! {@link PermissionDeniedException}:s are not considered validation errors.
		They are ignored and doesn't change the validation status.
		<p>
		{@link InvalidRelationException} is a less severe type of exception
		and can be used by a validator when the actual file it is validating
		is valid but a dependent file is not. For example, a valid CEL file has
		been attached to a raw bioassay, but it doesn't match the CDF file attached 
		to the array design. If a validator throws this exception the member status
		will be set to invalid, but metadata is not reset.
		
		@param dc The DbControl to use in the validation process
		@return A list of validation errors, if everything is successful the
			list is empty
	*/
	public List<Throwable> validate(DbControl dc)
	{
		checkPermission(Permission.WRITE);

		// List that collects all validation errors
		List<Throwable> errors = new LinkedList<Throwable>();
		
		// Load all members of the file set
		List<ValidatingFileSetMember> members = new ArrayList<ValidatingFileSetMember>(getData().getMembers().size());
		for (FileSetMemberData memberData : getData().getMembers())
		{
			FileSetMember member = dc.getItem(FileSetMember.class, memberData);
			members.add(new ValidatingFileSetMember(member));
		}
	
		// Create context and invoke the extensions system
		ExtensionsInvoker<ValidationAction> invoker = getInvoker(dc);
		
		// Perform the validation
		invoker.render(new ValidationRenderer(members, errors));
		
		// Update flags, etc on member data items
		for (ValidatingFileSetMember member : members)
		{
			if (member.hasBeenAccepted()) 
			{
				FileSetMemberData memberData = member.getMember().getData();
				memberData.setValid(member.isValidSoFar());
				memberData.setErrorMessage(member.getMessage());
			}
		}
		return errors;
	}
	
	private ExtensionsInvoker<ValidationAction> getInvoker(DbControl dc)
	{
		// Create context and invoke the extensions system
		FileStoreEnabled ownerItem = getItem();
		ClientContext context = new ClientContext(dc, ownerItem);
		ExtensionsManager xtManager = Application.getExtensionsManager();
		Registry registry = xtManager.getRegistry();
		Settings settings = xtManager.getSettings();
		ExtensionsInvoker<ValidationAction> invoker = registry.useExtensions(context, settings, "net.sf.basedb.core.filehandler.validator");
		return invoker;
	}
	
	
	private static class QueryRuntimeFilterImpl
		implements QueryRuntimeFilter
	{
		@Override
		public void enableFilters(QueryRuntimeFilterManager manager, EntityQuery query, DbControl dc)
		{
			// Do not add any filters.
		}
	}

}
