/*
	$Id: PlatformFileType.java 6898 2015-05-12 09:27:08Z nicklas $

	Copyright (C) 2007 Nicklas Nordborg

	This file is part of BASE - BioArray Software Environment.
	Available at http://base.thep.lu.se/

	BASE is free software; you can redistribute it and/or
	modify it under the terms of the GNU General Public License
	as published by the Free Software Foundation; either version 3
	of the License, or (at your option) any later version.

	BASE is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with BASE. If not, see <http://www.gnu.org/licenses/>.
*/
package net.sf.basedb.core;

import net.sf.basedb.core.data.PlatformFileTypeData;
import net.sf.basedb.core.query.Hql;
import net.sf.basedb.core.query.Restriction;
import net.sf.basedb.core.query.Restrictions;

/**
	This class represents a specific {@link DataFileType}
	that is part of a platform.

	@author Nicklas
	@version 2.5
	@base.modified $Date: 2015-05-12 11:27:08 +0200 (Tue, 12 May 2015) $
*/
public class PlatformFileType
	extends BasicChildItem
	implements UsableDataFileType
{
	
	/**
		The type of item represented by this class.
		@see Item#PLATFORMFILETYPE
		@see #getType()
	*/
	public static final Item TYPE = Item.PLATFORMFILETYPE;

	/**
		This filter will only return items if the logged in user has
		generic read permission to platforms.
	*/
	private static final QueryRuntimeFilter RUNTIME_FILTER = 
		new QueryRuntimeFilterFactory.ChildFilter(TYPE, Item.PLATFORM);

	/**
		Create a new item linking a file type to a platform/variant.
		@param platform The platform
		@param variant The variant (optional)
		@param fileType The file type
	*/
	static PlatformFileType getNew(DbControl dc, Platform platform, PlatformVariant variant,
		DataFileType fileType)
	{
		PlatformFileType type = dc.newItem(PlatformFileType.class);
		PlatformFileTypeData data = type.getData();
		data.setPlatform(platform.getData());
		data.setVariant(variant == null ? null : variant.getData());
		data.setDataFileType(fileType.getData());
		return type;
	}
	
	/**
		Get a <code>PlatformFileType</code> object when you know the ID.
	
		@param dc The <code>DbControl</code> which will be used for
			permission checking and database access.
		@param id The ID of the item to load
		@return The <code>PlatformFileType</code> item
		@throws ItemNotFoundException If an item with the specified ID is not found
		@throws PermissionDeniedException If the logged in user doesn't have 
			{@link Permission#READ} permission to the item
		@throws BaseException If there is another error
	*/
	public static PlatformFileType getById(DbControl dc, int id)
		throws ItemNotFoundException, PermissionDeniedException, BaseException
	{
		PlatformFileType p = dc.loadItem(PlatformFileType.class, id);
		if (p == null) throw new ItemNotFoundException("PlatformFileType[id="+id+"]");
		return p;
	}
	
	/**
		Get the file types that have been registered for a specific platform
		and variant. If the <code>restrict</code> parameter is TRUE the query will
		be restricted to return fewer items. It works like this:
		
		<ul>
		<li>It has no meaning if both <code>platform</code> and
			<code>variant</code> is null
		<li>If <code>variant</code> is null, the query will return file
			types the platform and all variants if <code>restrict=false</code>. If 
			<code>restrict=true</code> the query will only return items
			registered directly with the platform
		<li>If <code>variant</code> isn't null, the query will return file
			types that have been registered for that variant plus file
			types registered with the platform if <code>restrict=false</code>.
			If <code>restrict=true</code> the query will only return file types
			registered for that particular variant.
		</ul>
		
		Note! If a variant is given but no platform, the platform will
		automatically be set from the variant.
		
		@param platform A platform object or null to return items for
			all platforms
		@param variant A platform variant or null to return items for all
			variants
		@param restrict Indicates if the query should be restricted to inlcude
			fewer items
		@return A query returning <code>PlatformFileType</code> items
	*/
	public static ItemQuery<PlatformFileType> getQuery(Platform platform, 
		PlatformVariant variant, boolean restrict)
	{
		ItemQuery<PlatformFileType> query = 
			new ItemQuery<PlatformFileType>(PlatformFileType.class, RUNTIME_FILTER);
		if (platform == null && variant != null) platform = variant.getPlatform();
		if (platform != null)
		{
			query.restrictPermanent(
				Restrictions.eq(
					Hql.property("platform"),
					Hql.entity(platform)
				)
			);
			
			Restriction commonVariant = 
				Restrictions.eq(Hql.property("variant"), null);
			if (variant == null)
			{
				if (restrict)
				{
					// Only load file types that are common to all variants
					query.restrictPermanent(commonVariant);
				}
			}
			else
			{
				Restriction specificVariant = 
					Restrictions.eq(Hql.property("variant"), Hql.entity(variant));
				if (restrict)
				{
					// Only load file types specific to the variant
					query.restrictPermanent(specificVariant);
				}
				else
				{
					// Load common + specific file types
					query.restrictPermanent(Restrictions.or(specificVariant, commonVariant));
				}
			}
			
		}
		return query;
		
	}
	
	/**
		Creates a new member item from the given data.
		@param data the data.
	*/
	PlatformFileType(PlatformFileTypeData data)
	{
		super(data);
	}
	
	@Override
	PlatformFileTypeData getData()
	{
		return (PlatformFileTypeData)getBasicData();
	}
	/*
		From the Identifiable interface
		-------------------------------------------
	*/
	@Override
	public Item getType()
	{
		return TYPE;
	}
	// -------------------------------------------

	/*
		From the BasicChildItem class
		-------------------------------------------
	*/
	@Override
	Item getParentType()
	{
		return Item.PLATFORM;
	}
	// -------------------------------------------
	
	/**
		Get the platform.
		@return A platform item
	*/
	public Platform getPlatform()
	{
		return getDbControl().getItem(Platform.class, getData().getPlatform());		
	}

	/**
		Get the platform variant.
		@return A platform variant item, or null if this file type is associated with
			all variants of the platform
	*/
	public PlatformVariant getVariant()
	{
		return getDbControl().getItem(PlatformVariant.class, getData().getVariant());		
	}

	/**
		Get the type of the file.
		@return A file set member type item
	*/
	@Override
	public DataFileType getDataFileType()
	{
		return getDbControl().getItem(DataFileType.class, getData().getDataFileType());
	}
	
	/**
		Check if a file of this type is required or not. 
		Note! The requirement is not enforced by the core. It should be used as a
		hint to client applications so they can create a proper GUI. To check
		if a {@link FileSet} contains all files that have been marked as required
		use {@link FileSet#hasAllRequiredFiles()}.
	*/
	@Override
	public boolean isRequired()
	{
		return getData().isRequired();
	}
	/**
		Set the required flag for this file type.
	*/
	public void setRequired(boolean required)
	{
		checkPermission(Permission.WRITE);
		getData().setRequired(required);
	}
	
	/**
		Check if more than one file of this type is allowed in a file set or not.
		Note! The requirement is not enforced by the core. It should be used as a
		hint to client applications so they can create a proper GUI. 
	*/
	@Override
	public boolean getAllowMultiple()
	{
		return getData().getAllowMultiple();
	}
	
	/**
		Allow or disallow multiple files of this file type in a file set.
	*/
	public void setAllowMultiple(boolean multiple)
	{
		checkPermission(Permission.WRITE);
		getData().setAllowMultiple(multiple);
	}

}
