/*
	$Id$

	Copyright (C) 2015 Nicklas Nordborg

	This file is part of BASE - BioArray Software Environment.
	Available at http://base.thep.lu.se/

	BASE is free software; you can redistribute it and/or
	modify it under the terms of the GNU General Public License
	as published by the Free Software Foundation; either version 3
	of the License, or (at your option) any later version.

	BASE is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with BASE. If not, see <http://www.gnu.org/licenses/>.
*/
package net.sf.basedb.core;

import java.security.MessageDigest;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.Collection;
import java.util.Collections;
import java.util.Comparator;
import java.util.List;
import java.util.Map;
import java.util.SortedSet;
import java.util.TreeSet;

import net.sf.basedb.core.data.ItemListSyncFilterData;
import net.sf.basedb.core.data.PropertyFilterData;
import net.sf.basedb.core.data.SharedData;
import net.sf.basedb.core.query.Hql;
import net.sf.basedb.core.query.Restrictions;
import net.sf.basedb.util.MD5;
import net.sf.basedb.util.formatter.Formatter;
import net.sf.basedb.util.formatter.PropertyFilterFormatter;
import net.sf.basedb.util.listable.ListableUtil;

/**
	Holds information about a synchronization filter used
	by an item list.
	@author Nicklas
	@since 3.5
*/
public class SyncFilter
	extends ChildItem
	implements Nameable
{

	/**
		The type of item represented by this class.
		@see Item#SYNCFILTER
		@see #getType()
	*/
	public static final Item TYPE = Item.SYNCFILTER;

	/**
		Get a <code>SyncFilter</code> item when you know the id.
	
		@param dc The <code>DbControl</code> which will be used for
			permission checking and database access.
		@param id The id of the item to load
		@return The <code>SyncFilter</code> item
		@throws ItemNotFoundException If an item with the specified 
			id is not found
		@throws PermissionDeniedException If the logged in user doesn't 
			have {@link Permission#READ} permission to the item
		@throws BaseException If there is another error
	*/
	public static SyncFilter getById(DbControl dc, int id)
		throws ItemNotFoundException, PermissionDeniedException, BaseException
	{
		SyncFilter sf = dc.loadItem(SyncFilter.class, id);
		if (sf == null) throw new ItemNotFoundException("SyncFilter[id="+id+"]");
		return sf;
	}
	
	/**
		Get a query configured to retrieve 
		sync filters for the specified item list.
		
		@param list The item list, null is not allowed
		@return An {@link ItemQuery} object
	*/
	public static ItemQuery<SyncFilter> getQuery(ItemList list)
	{
		if (list == null) throw new InvalidUseOfNullException("list");
		ItemQuery<SyncFilter> query = new ItemQuery<SyncFilter>(SyncFilter.class, null);
		query.restrictPermanent(
			Restrictions.eq(
				Hql.property("itemList"), 
				Hql.entity(list)
			)
		);
		return query;
	}

	/**
		Utility method for formatting and sorting a list of property filters.
		@see PropertyFilterFormatter
	*/
	public static List<String> getFormattedAndSortedFilters(Collection<PropertyFilter> filters, Formatter<PropertyFilter> filterFormatter)
	{
		List<String> formattedFilters = new ArrayList<String>(filters.size());
		for (PropertyFilter f : filters)
		{
			formattedFilters.add(filterFormatter.format(f));
		}
		Collections.sort(formattedFilters);
		return formattedFilters;
	}
	
	private static final List<Item> CHAIN_ORDER = Arrays.asList(new Item[] {
		Item.BIOSOURCE, Item.SAMPLE, Item.EXTRACT, Item.PHYSICALBIOASSAY, Item.DERIVEDBIOASSAY, Item.RAWBIOASSAY
	});
	
	private static final List<SourceItemTransform> TRANSFORM_ORDER = Arrays.asList(new SourceItemTransform[] {
			SourceItemTransform.PARENT_TO_CHILD, SourceItemTransform.NONE, SourceItemTransform.CHILD_TO_PARENT
		});
	
	/**
		Create a comparator that sorts sync filters by:
		* Filter group
		* Source item type: BIOSOURCE, SAMPLE, EXTRACT, PHYSICALBIOASSAY, DERIVEDBIOASSAY, RAWBIOASSAY
		* Type of transformation: PARENT_TO_CHILD, NONE, CHILD_TO_PARENT
		* Name
		* Source item must exists is sorted before source item must not exist
		* Id
	*/
	public static Comparator<SyncFilter> compareByChainOrder()
	{
		return new Comparator<SyncFilter>() 
		{
			@Override
			public int compare(SyncFilter o1, SyncFilter o2) 
			{
				// Filter group
				int result = o1.getFilterGroup().compareTo(o2.getFilterGroup());
				if (result != 0) return result;
				
				// Source item type
				int co1 = CHAIN_ORDER.indexOf(o1.getSourceItemType());
				int co2 = CHAIN_ORDER.indexOf(o2.getSourceItemType());		
				result = co1 - co2;
				if (result != 0) return co1 == -1 ? 1 : result;

				// Transform type
				co1 = TRANSFORM_ORDER.indexOf(o1.getSourceItemTransformation());
				co2 = TRANSFORM_ORDER.indexOf(o2.getSourceItemTransformation());				
				result = co1 - co2;
				if (result != 0) return result;

				// Name
				result = o1.getName().compareTo(o2.getName());
				if (result != 0) return result;

				// Source item exists
				result = (o1.getSourceItemMustExist() ? 0 : 1) - (o2.getSourceItemMustExist() ? 0 : 1);
				if (result != 0) return result;

				// Id
				return o1.getId() - o2.getId();
			}
		};
	}
	
	private Item listMemberType;
	private List<PropertyFilter> filters;
	private int numFilterRows;
	private SortedSet<Integer> matching;
	private String matchingMd5;
	
	SyncFilter(ItemListSyncFilterData syncFilter)
	{
		super(syncFilter);
	}
	
	@Override
	ItemListSyncFilterData getData()
	{
		return (ItemListSyncFilterData)getBasicData();
	}
	
	/*
		From the Identifiable interface
		-------------------------------------------
	*/
	@Override
	public Item getType()
	{
		return TYPE;
	}
	// -------------------------------------------
	/*
		From the ChildItem class
		-------------------------------------------
	*/
	@Override
	SharedData getSharedParent()
	{
		return getData().getItemList();
	}
	// -------------------------------------------

	/*
		From the Nameable interface
		-------------------------------------------
	*/
	@Override
	public String getName()
	{
		return getData().getName();
	}
	@Override
	public void setName(String name)
		throws PermissionDeniedException, InvalidDataException
	{
		checkPermission(Permission.WRITE);
		NameableUtil.setName(getData(), name);
	}
	@Override
	public String getDescription()
	{
		return getData().getDescription();
	}
	@Override
	public void setDescription(String description)
		throws PermissionDeniedException, InvalidDataException
	{
		checkPermission(Permission.WRITE);
		NameableUtil.setDescription(getData(), description);
	}
	// -------------------------------------------

	/**
		Get the item list this filter belongs to.
		@return The <code>SyncFilter</code> item
		@throws PermissionDeniedException If the logged in user doesn't have 
			{@link Permission#READ} permission
		@throws BaseException If there is another error
	*/
	public ItemList getItemList()
		throws PermissionDeniedException, BaseException
	{
		return getDbControl().getItem(ItemList.class, getData().getItemList());
	}
	
	/**
		Copy filters and other settings from the given context
		to this synchronization filter. This changes
		the source item type ({@link #getSourceItemType()}), 
		transform ({@link #getSourceItemTransformation()})
		and replaces all existing filters with the filters from the
		item context.
		@param syncContext The context to copy from, or null to only update the transform
		@param transform The new transformation or null to keep the current transformation
	*/
	public void setSyncContext(ItemContext syncContext, SourceItemTransform transform)
	{
		checkPermission(Permission.WRITE);
		if (transform == null) transform = getSourceItemTransformation();
		Item sourceItemType = syncContext != null ? syncContext.getItemType() : getSourceItemType();
		
		List<Item> allowedSourceItems = ListableUtil.getListableItems(getListMemberType(), transform);
		if (allowedSourceItems.contains(sourceItemType))
		{
			// Copy filter from sync context
			if (syncContext != null) syncContext.copyToSyncFilter(getData());
		}
		else
		{
			if (syncContext == null)
			{
				getData().getPropertyFilters().clear();
			}
			else
			{
				throw new InvalidDataException("Sync context is not found among allowed source items " + allowedSourceItems);
			}
		}
		
		getData().setSourceItemTransform(transform.ordinal());
		initFilters();
	}
	
	/**
		Copy the filter settings in this synchronization filter to
		the given item context. All filter settings in the item
		context are replaced with this filter. 
		The context must be for the same item type as the source item
		type of this filter.
	*/
	public void copyToContext(ItemContext context)
	{
		if (context == null) throw new InvalidUseOfNullException("context");
		if (context.getItemType() != getSourceItemType())
		{
			throw new InvalidDataException("Incompatible item types: " + context.getItemType().name() + " != " + getSourceItemType().name());
		}
		context.copyFromSyncFilter(getData());
	}

	/**
		Get the group this synchronization filter belongs to. The group
		is a string where the first character is one of A-Z followed by a '+'
		or '-' sign. Filters that belong to the same group are evaluated together 
		and then merged with UNION operation (for groups with '+' sign) or 
		REMOVE operation (for groups with '-' sign). Groups are processed
		in alphabetical order (A+, A-, B+, B-, .... Z+, Z-)
		@since 3.18
	*/
	public String getFilterGroup()
	{
		return getData().getFilterGroup();
	}
	
	/**
		Set the group that this synchronization filter belongs to.
		See {@link #getFilterGroup()} for a description of allowed values.
		
		@since 3.18
	*/
	public void setFilterGroup(String filterGroup)
	{
		checkPermission(Permission.WRITE);
		if (filterGroup == null) throw new InvalidUseOfNullException("filterGroup");
		if (!filterGroup.matches("[A-Z][+-]"))
		{
			throw new InvalidDataException("Filter group must match '[A-Z][+-]': " + filterGroup);
		}
		getData().setFilterGroup(filterGroup);
	}
	
	/**
		Get the source item type that this synchronization filter applies to.
		This may or may not be the same item type as the member type of the
		list. If this type is different from the list member type an item
		transformation is required, but a transformation can also be used
		for items of the same type.
	*/
	public Item getSourceItemType()
	{
		return Item.fromValue(getData().getSourceItemType());
	}

	/**
		Get the transformation to use for converting the items matching
		this filter to list members.
	*/
	public SourceItemTransform getSourceItemTransformation()
	{
		return SourceItemTransform.values()[getData().getSourceItemTransform()];
	}
	
	/**
		If set, a source item that matches the filter must exist, otherwise it must 
		not exist. Only applies to filters with a source item transformation that is 
		to a parent/child item.
	*/
	public boolean getSourceItemMustExist()
	{
		return getData().getSourceMustExist();
	}
	
	public void setSourceItemMustExist(boolean sourceMustExist)
	{
		checkPermission(Permission.WRITE);
		getData().setSourceMustExist(sourceMustExist);
	}
	
	/**
		If set, this filter is disabled and is not used when syncing.
	*/
	public boolean isDisabled()
	{
		return getData().isDisabled();
	}
	
	/**
		Disabled/enable this filter.
	*/
	public void setDisabled(boolean disabled)
	{
		checkPermission(Permission.WRITE);
		getData().setDisabled(disabled);
	}
	
	/**
		Get the number of source items that matched this particular filter during the last
		synchronization. See {@link ItemList#getSyncDate()}.
		@return The number of matching source items, or -1 if not known
	*/
	public int getLastSyncMatchingSource()
	{
		return getData().getMatchingSource();
	}
	
	/**
		Get the number of transformed items that matched this particular filter during the last
		synchronization. See {@link ItemList#getSyncDate()}. If this filter
		doesn't use a transform the same number as {@link #getLastSyncMatchingSource()}
		is returned.
		@return The number of matching transformed items, or -1 if not known
	*/
	public int getLastSyncMatchingTransformed()
	{
		return getData().getMatchingTransformed();
	}
	
	/**
		Get the number of items that is currently matching this filter. 
		The method will cause a query to be executed against the database,
		but the result is cached and re-used if called multiple times.
	*/
	public int getCurrentMatchingSource(DbControl dc)
	{
		return getMatching(dc).size();
	}

	/**
		Checks if the items that are currently matching this filter are
		the same as the items that matched it during the last sync. The 
		check is done by comparing the MD5 sum of the concatenated
		ID values.
	*/
	public boolean isInSync(DbControl dc)
	{
		if (!getSyncProperties(getData()).equals(getData().getLastSyncProperties())) return false;
		SortedSet<Integer> tmp = getMatching(dc);
		if (tmp.size() != getLastSyncMatchingSource()) return false;
		if (matchingMd5 == null) matchingMd5 = toMd5(matching);
		return matchingMd5.equals(getData().getMd5());
	}
	
	/**
		Get the number of filter rows in this filter.
	*/
	public int getFilterRows()
	{
		if (filters == null) initFilters();
		return numFilterRows;
	}
	
	/**
		Get the filters in the given filter row.
		@param filterRow The index of the filter row which should be between 0 and number of rows - 1, or use -1 to
			get all property filters
	*/
	public Collection<PropertyFilter> getPropertyFilters(int filterRow)
	{
		if (filters == null) initFilters();
		List<PropertyFilter> tmp = new ArrayList<PropertyFilter>(filters.size());
		for (PropertyFilter f : filters)
		{
			if (filterRow == -1 || (f.getRowIndex() == filterRow)) 
			{
				tmp.add(f);
			}
		}
		return tmp;
	}

	private void initFilters()
	{
		numFilterRows = 0;
		filters = new ArrayList<PropertyFilter>();
		for (Map.Entry<String, PropertyFilterData> entry : getData().getPropertyFilters().entrySet())
		{
			PropertyFilter pf = PropertyFilter.create(entry.getKey(), entry.getValue());
			if (pf.getRowIndex() >= numFilterRows) 
			{
				numFilterRows = pf.getRowIndex()+1;
			}
			filters.add(pf);
		}
	}

	private Item getListMemberType()
	{
		if (listMemberType == null)
		{
			listMemberType = Item.fromValue(getData().getItemList().getMemberType());
		}
		return listMemberType;
	}
	
	private SortedSet<Integer> getMatching(DbControl dc)
	{
		if (matching == null)
		{
			Item sourceItemType = getSourceItemType();
			ItemQuery<? extends BasicItem> query = sourceItemType.getQuery();
			ItemContext syncContext = new ItemContext(getSourceItemType(), getData());
			syncContext.configureQuery(dc, query, true);
			matching = new TreeSet<Integer>(query.idList(dc));
		}
		return matching;
	}
	
	/**
		Calculate the MD5 sum of all all ID values in the collection.
		Each ID is converted to a 4-byte array and added to the
		MD5 by {@link MessageDigest#update(byte[])}
		The list is assumed to be sorted in ID order so that
		if two collections contain the same ID values, they
		should produce the same MD5.
	*/
	public static String toMd5(SortedSet<Integer> listOfIds)
	{
		MessageDigest md5 = MD5.newInstance();
		byte[] bytes = new byte[4];
		for (int id : listOfIds)
		{
			bytes[0] = (byte)(id >> 24);
			bytes[1] = (byte)(id >> 16);
			bytes[2] = (byte)(id >> 8);
			bytes[3] = (byte)(id >> 0);
			md5.update(bytes);
		}
		
		return MD5.getHashString(md5);
	}
	
	/**
		Get a string representing the current syncfilter properties that may affect the end
		result even if the query return the same items. We save this string at sync time
		and can then compare it with current settings to se if the item list seems to be
		up to date.
		@since 3.18
	*/
	static String getSyncProperties(ItemListSyncFilterData sf)
	{
		return sf.getFilterGroup()+sf.getSourceItemTransform()+":"+sf.getSourceItemType()+":"+(sf.getSourceMustExist() ? "1" : "0");
	}
	
	public enum SourceItemTransform
	{
		/**
			No transformation of source items is done. The items matching the filter
			will become members of the item list.
		*/
		NONE,
		
		/**
			Source items are transformed to child items. Items matching the filter
			will have their child items looked up which will become members of the 
			list.
		*/
		PARENT_TO_CHILD,
		
		/**
			Source items are transformed to parent items. Items matching the filter
			will have their parent items looked up which will become members of
			the list.
		*/
		CHILD_TO_PARENT;
	}
}
