/*
	$Id: SystemItems.java 6881 2015-04-21 07:59:42Z nicklas $

	Copyright (C) 2005 Johan Enell, Jari Häkkinen, Nicklas Nordborg
	Copyright (C) 2006 Jari Häkkinen, Nicklas Nordborg

	This file is part of BASE - BioArray Software Environment.
	Available at http://base.thep.lu.se/

	BASE is free software; you can redistribute it and/or
	modify it under the terms of the GNU General Public License
	as published by the Free Software Foundation; either version 3
	of the License, or (at your option) any later version.

	BASE is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with BASE. If not, see <http://www.gnu.org/licenses/>.
*/
package net.sf.basedb.core;

import net.sf.basedb.core.data.SystemData;
import net.sf.basedb.core.data.RoleKeyData;

import java.util.Map;
import java.util.HashMap;
import java.util.List;

/**
	This class is used to map the system id of all {@link
	net.sf.basedb.core.data.SystemData} items to the numerical internal 
	id in the current installation. This information is initialised 
	at startup and is then kept in memory.

	@author Nicklas
	@version 2.0
	@base.modified $Date: 2015-04-21 09:59:42 +0200 (Tue, 21 Apr 2015) $
*/
public class SystemItems
{

	private static boolean isInitialised = false;

	/**
		A map from system id --&gt; numeric id
	*/
	private static Map<String, Integer> systemIds = null;
	
	/**
		An array to hold mappings for {@link net.sf.basedb.core.data.RoleKeyData}:
		RoleKeyData.itemType --&gt; RoleKeyData.id
	*/
	private static int[] roleKeys = null;

	/**
		Initialize the class and load all mappings for item.systemId --&gt; item.id
		and for RoleKeyData.itemType --&gt; RoleKeyData.id
		@throws BaseException If there is an error
	*/
	static synchronized void init()
		throws BaseException
	{
		if (isInitialised) return;
		org.hibernate.Session session = null;
		org.hibernate.Transaction tx = null;
		
		try
		{
			systemIds = new HashMap<String,Integer>();
			roleKeys = new int[Item.MAX_VALUE+1];
			session = HibernateUtil.newSession();
			tx = HibernateUtil.newTransaction(session);
			List<?> result = HibernateUtil.getPredefinedQuery(session, "LOAD_SYSTEM_ITEMS").list();
			/*
				SELECT item.id, item.systemId
				FROM net.sf.basedb.core.data.SystemData item
				WHERE item.systemId NOT IS NULL
			*/
			for (int i = 0; i < result.size(); i++)
			{
				Object[] item = (Object[])result.get(i);
				String systemId = (String)item[1];
				Integer id = (Integer)item[0];
				systemIds.put(systemId.toLowerCase(), id);
			}
			
			// Load RoleKeys
			result = HibernateUtil.getPredefinedQuery(session, "LOAD_ROLE_KEY_IDS").list();
			/*
				SELECT rk.id, rk.itemType
				FROM RoleKeyData rk
			*/
			for (int i = 0; i < result.size(); i++)
			{
				Object[] item = (Object[])result.get(i);
				roleKeys[(Integer)item[1]] = (Integer)item[0];
			}
			isInitialised = true;
		}
		finally
		{
			if (tx != null) HibernateUtil.commit(tx);
			if (session != null) HibernateUtil.close(session);
		}
	}
	
	/**
		Unload all settings.
	*/
	static synchronized void unload()
	{
		isInitialised = false;
		if (systemIds != null) systemIds.clear();
		systemIds = null;
		roleKeys = null;
	}

	/**
		Register a system item. Used by the {@link Install} class while
		creating new items.
	*/
	static void add(SystemData item)
	{
		assert isInitialised : "SystemItems has not been initialised";
		systemIds.put(item.getSystemId().toLowerCase(), item.getId());
	}

	/**
		Get the numeric id for an item when you know the system id.
		@param systemId The system id string
		@return The numeric id of the item, or 0 if it could not be found
	*/
	public static int getId(String systemId)
	{
		assert isInitialised : "SystemItems has not been initialised";
		if (systemId == null) return 0;
		Integer id = systemIds.get(systemId.toLowerCase());
		return id == null ? 0 : id.intValue();
	}
	
	/**
		Register a role key. Used by the {@link Install} class while
		creating new items.
	*/
	static void addRoleKey(RoleKeyData roleKey)
	{
		assert isInitialised : "SystemItems has not been initialised";
		Item itemType = Item.fromValue(roleKey.getItemType());
		roleKeys[itemType.getValue()] = roleKey.getId();
	}
	
	/**
		Get the numeric id for a {@link RoleKey} when you know the {@link Item}
		type of the RoleKey.
		@param itemType The item type code as defined by the {@link Item} class
		@return The numeric id for the RoleKey
	*/
	public static int getRoleKeyId(Item itemType)
	{
		assert isInitialised : "SystemItems has not been initialised";
		assert itemType.getValue() < roleKeys.length : "Bad item type: "+itemType+"; value="+itemType.getValue();
		return roleKeys[itemType.getValue()];
	}

}
