/*
  $Id$

	Copyright (C) 2015 Nicklas Nordborg

	This file is part of BASE - BioArray Software Environment.
	Available at http://base.thep.lu.se/

  BASE is free software; you can redistribute it and/or
  modify it under the terms of the GNU General Public License
  as published by the Free Software Foundation; either version 3
  of the License, or (at your option) any later version.

  BASE is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with BASE. If not, see <http://www.gnu.org/licenses/>.
*/
package net.sf.basedb.core.data;

import java.util.Map;
import java.util.HashMap;

/**
	Synchronization options for item lists. This is similar
	to filters used by {@link ContextData}.  
	
	@author Nicklas
	@since 3.5
	@hibernate.class table="`ItemListSyncFilters`" lazy="true"
 */
public class ItemListSyncFilterData
	extends BasicData
	implements NameableData
{
	
	public ItemListSyncFilterData()
	{}

	/*
		From the NameableData interface
		-------------------------------------------
	*/
	private String name;
	@Override
	public String getName()
	{
		return name;
	}
	@Override
	public void setName(String name)
	{
		this.name = name;
	}
	private String description;
	@Override
	public String getDescription()
	{
		return description;
	}
	@Override
	public void setDescription(String description)
	{
		this.description = description;
	}
	// -------------------------------------------
	
	private ItemListData itemList;
	/**
		Get the item list this sync filter belongs to.
		@hibernate.many-to-one column="`itemlist_id`" not-null="true" outer-join="false" update="false"
	*/
	public ItemListData getItemList()
	{
		return itemList;
	}
	public void setItemList(ItemListData itemList)
	{
		this.itemList = itemList;
	}

	private String filterGroup;
	/**
		Get the group for this filter.
		@since 3.18
		@hibernate.property column="`filter_group`" type="string" length="255" not-null="true" 
	*/
	public String getFilterGroup()
	{
		return filterGroup;
	}
	public void setFilterGroup(String filterGroup)
	{
		this.filterGroup = filterGroup;
	}
	
	private int sourceItemType;
	/**
		Get the type of items that this filter is applies to.
		@hibernate.property column="`source_type`" type="int" not-null="true" 
	*/
	public int getSourceItemType()
	{
		return sourceItemType;
	}
	public void setSourceItemType(int sourceItemType)
	{
		this.sourceItemType = sourceItemType;
	}
	
	private int sourceItemTransform;
	/**
		The transform to use for converting filter matches to list members. 
		0 = none; 1 = parent to child, 2 = child to parent
		@hibernate.property column="`source_transform`" type="int" not-null="true"
	*/
	public int getSourceItemTransform()
	{
		return sourceItemTransform;
	}
	public void setSourceItemTransform(int sourceItemTransform)
	{
		this.sourceItemTransform = sourceItemTransform;
	}
	
	private boolean sourceMustExist;
	/**
		If set, the source item must exist, otherwise it must not exist. This
		option is only used if the transform is to a parent/child item
		@hibernate.property column="`source_must_exist`" type="boolean" not-null="true" 
	*/
	public boolean getSourceMustExist()
	{
		return sourceMustExist;
	}
	public void setSourceMustExist(boolean sourceMustExist)
	{
		this.sourceMustExist = sourceMustExist;
	}
	
	private boolean disabled;
	/**
		If set, the filter is disabled and should not be used when syncing.
		@hibernate.property column="`disabled`" type="boolean" not-null="true" 
	*/
	public boolean isDisabled()
	{
		return disabled;
	}
	public void setDisabled(boolean disabled)
	{
		this.disabled = disabled;
	}

	private int matchingSource;
	/**
		The number of matching source items during that last sync.
		@hibernate.property column="`matching_source`" type="int" not-null="true"
	*/
	public int getMatchingSource()
	{
		return matchingSource;
	}
	public void setMatchingSource(int matchingSource)
	{
		this.matchingSource = matchingSource;
	}

	private int matchingTransformed;
	/**
		The number of matching items after the transform during that last sync.
		@hibernate.property column="`matching_transformed`" type="int" not-null="true"
	*/
	public int getMatchingTransformed()
	{
		return matchingTransformed;
	}
	public void setMatchingTransformed(int matchingTransformed)
	{
		this.matchingTransformed = matchingTransformed;
	}
	
	private String lastSyncProperties;
	/**
		A string representation of properties that was affecting that last filter
		during the last sync. Used to check if the result may affect the final
		item list even if the actual items returned by the query hasn't change.
		Example: changing the filterGroup or sourceMustExists options.
		@since 3.18
		@hibernate.property column="`last_sync_properties`" type="string" length="255" not-null="false"
	*/
	public String getLastSyncProperties()
	{
		return lastSyncProperties;
	}
	public void setLastSyncProperties(String lastSyncProperties)
	{
		this.lastSyncProperties = lastSyncProperties;
	}
	
	
	private String md5;
	/**
		Get the MD5 hash of the concatenated ID values. It is always returned as a string
		with 32 hexadecimal characters.
		@hibernate.property column="`md5`" type="string" length="32" not-null="false"
	*/
	public String getMd5()
	{
		return md5;
	}
	public void setMd5(String md5)
	{
		this.md5 = md5;
	}
	
	private int include;
	/**
		Flags for specifying include options to an entity query.
		@hibernate.property column="`include`" type="int" not-null="true"
	*/
	public int getInclude()
	{
		return include;
	}
	public void setInclude(int include)
	{
		this.include = include;
	}

	private Map<String, PropertyFilterData> propertyFilters;
	/**
		Get the map that manages property filters in this sync filter.

		@hibernate.map table="`ItemListSyncPropertyFilters`" lazy="true"
		@hibernate.collection-key column="`syncfilter_id`"
		@hibernate.collection-index column="`property`" type="string" length="255" 
		@hibernate.collection-composite-element class="net.sf.basedb.core.data.PropertyFilterData" not-null="true"
	*/
	public Map<String, PropertyFilterData> getPropertyFilters()
	{
		if (propertyFilters == null) propertyFilters = new HashMap<String, PropertyFilterData>();
		return propertyFilters;
	}
	void setPropertyFilters(Map<String, PropertyFilterData> propertyFilters)
	{
		this.propertyFilters = propertyFilters;
	}


}
