/**
	$Id: AnnotationLogger.java 7381 2017-05-22 12:35:27Z nicklas $

	Copyright (C) 2009 Nicklas Nordborg

	This file is part of BASE - BioArray Software Environment.
	Available at http://base.thep.lu.se/

	BASE is free software; you can redistribute it and/or
	modify it under the terms of the GNU General Public License
	as published by the Free Software Foundation; either version 3
	of the License, or (at your option) any later version.

	BASE is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with BASE. If not, see <http://www.gnu.org/licenses/>.
*/
package net.sf.basedb.core.log.db;

import java.util.List;

import org.hibernate.Hibernate;
import org.hibernate.query.Query;

import net.sf.basedb.core.Annotation;
import net.sf.basedb.core.AnnotationBatcher.CurrentAnnotationInfo;
import net.sf.basedb.core.AnnotationType;
import net.sf.basedb.core.Item;
import net.sf.basedb.core.LogControl;
import net.sf.basedb.core.PredefinedQuery;
import net.sf.basedb.core.Type;
import net.sf.basedb.core.data.AnnotationData;
import net.sf.basedb.core.data.AnnotationSetData;
import net.sf.basedb.core.data.AnnotationTypeData;
import net.sf.basedb.core.data.ChangeHistoryDetailData;
import net.sf.basedb.core.data.LoggableData;
import net.sf.basedb.core.data.ParameterValueData;
import net.sf.basedb.core.hibernate.TypeWrapper;
import net.sf.basedb.core.log.ChangeType;
import net.sf.basedb.core.log.EntityDetails;
import net.sf.basedb.core.log.EntityLogger;
import net.sf.basedb.core.log.LogManager;
import net.sf.basedb.util.Values;
import net.sf.basedb.util.formatter.Formatter;

/**
	Logger for annotations that logs changes as
	updates to the parent item the annotations belong to.
	
	@author Nicklas
	@version 2.13
	@base.modified $Date: 2017-05-22 14:35:27 +0200 (Mon, 22 May 2017) $
*/
public class AnnotationLogger
	implements EntityLogger
{

	private final DbLogManagerFactory factory;

	public AnnotationLogger(DbLogManagerFactory factory)
	{
		this.factory = factory;
	}
	
	/*
		From the EntityLogger interface
		-------------------------------
	*/
	@SuppressWarnings({ "unchecked", "rawtypes" })
	@Override
	public void logChanges(LogManager logManager, EntityDetails details)
	{
		LogControl logControl = logManager.getLogControl();
		
		Object entity = details.getEntity();
		String annotationType = null;
		Type valueType = null;
		boolean logOldPropertyValues = factory.isLoggingOldPropertyValues();
		Integer parentId = null;
		Integer parentType = null;
		Annotation.Source source = null;
		List<?> newValues = null;
		List<?> oldValues = null;

		if (entity instanceof CurrentAnnotationInfo)
		{
			// Log entry from AnnotationBatcher
			CurrentAnnotationInfo info = (CurrentAnnotationInfo)entity;
			AnnotationType at = info.getAnnotationType();
			annotationType = at.getName();
			valueType = info.getValueType();
			
			parentId = info.getItemId();
			parentType = info.getItemType().getValue();
			
			source = Annotation.Source.PRIMARY;
			if (at.getDisableLogOfValues()) logOldPropertyValues = false;
			if (logOldPropertyValues)
			{
				oldValues = info.getOldValues();
				newValues = info.getNewValues();
			}
		}
		else if (entity instanceof AnnotationData)
		{
			// Regular log entry via Hibernate
			AnnotationData annotation = (AnnotationData)entity;
			AnnotationTypeData at = annotation.getAnnotationType();
			if (!Hibernate.isInitialized(at))
			{
				Query<Object[]> q = logControl.createHqlQuery(
						PredefinedQuery.getQueryString("DBLOG_GET_ANNOTATIONTYPE_INFO"), Object[].class);
				q.setParameter("annotationTypeId", at.getId(), TypeWrapper.H_INTEGER);
				Object[] tmp = q.uniqueResult();
				annotationType = (String)tmp[0];
				valueType = Type.fromValue((Integer)tmp[1]);
				if ((Boolean)tmp[2]) logOldPropertyValues = false;
			}
			else
			{
				annotationType = at.getName();
				valueType = Type.fromValue(at.getValueType());
				if (at.getDisableLogOfValues()) logOldPropertyValues = false;
			}
			
			// Get the type of the parent item
			AnnotationSetData as = annotation.getAnnotationSet();
			if (!Hibernate.isInitialized(as))
			{
				Query<Integer> q = logControl.createHqlQuery(
					PredefinedQuery.getQueryString("DBLOG_GET_ANNOTATION_ITEMTYPE"), Integer.class);
				q.setParameter("annotationSetId", as.getId(), TypeWrapper.H_INTEGER);
				parentType = q.uniqueResult();			
			}
			else
			{
				parentType = as.getItemType();
			}
			
			// May happen in case the annotation set has been deleted in this transaction
			if (parentType == null) return;
			
			Class<?> parentClass = Item.fromValue(parentType).getDataClass();
			// If the parent item isn't implementing the LoggableData interface
			// we don't log the changes
			if (!LoggableData.class.isAssignableFrom(parentClass)) return;

			// Get the id of the parent item
			Query<Integer> q = logControl.createHqlQuery(
				PredefinedQuery.getQueryString("DBLOG_GET_ITEMID_WITH_ANNOTATION", parentClass.getName()), Integer.class);
			q.setParameter("annotationSetId", as.getId(), TypeWrapper.H_INTEGER);
			parentId = q.uniqueResult();
			
			// May happen in case the parent item has been deleted in this transaction
			if (parentId == null) return;
		
			source = Annotation.Source.values()[annotation.getSource()];
			if (logOldPropertyValues)
			{
				ParameterValueData<?> values = annotation.getValues();
				if (values != null)
				{
					newValues = values.getValues();
					oldValues = values.getOldValues();
				}
				else
				{
					logOldPropertyValues = false;
				}
			}
		}
		
		ChangeType changeType = details.getChangeType();
		if (changeType == ChangeType.CREATE) 
		{
			changeType = ChangeType.ADD;
		}
		else if (changeType == ChangeType.DELETE)
		{
			changeType = ChangeType.REMOVE;
		}
		
		ChangeHistoryDetailData change = new ChangeHistoryDetailData();
		change.setChangeType(changeType.getValue());
		change.setItemId(parentId);
		change.setItemType(parentType);

		if (source == Annotation.Source.CLONED)
		{
			change.setChangeInfo("ClonedAnnotation["+annotationType+"]");
		}
		else if (source == Annotation.Source.INHERITED)
		{
			change.setChangeInfo("InheritedAnnotation["+annotationType+"]");
		}
		else
		{
			change.setChangeInfo("Annotation["+annotationType+"]");
		}
		if (logOldPropertyValues)
		{
			Formatter f = null;
			if (valueType == Type.DATE)
			{
				f = EntityDetails.DATE_FORMAT;
			}
			else if (valueType == Type.TIMESTAMP)
			{
				f = EntityDetails.DATETIME_FORMAT;
			}
			
			String sNew = Values.getString(newValues, ", ", true, f);
			String sOld = Values.getString(oldValues, ", ", true, f);
			change.setNewValue(sNew.length() == 0 ? null : sNew);
			change.setOldValue(sOld.length() == 0 ? null : sOld);
		}
		((DbLogManager)logManager).logChangeDetails(change, details);
	}
	// -------------------------------
}
