/*
	$Id: NeqRestriction.java 6898 2015-05-12 09:27:08Z nicklas $

	Copyright (C) 2005 Samuel Andersson, Nicklas Nordborg
	Copyright (C) 2006 Jari Häkkinen, Nicklas Nordborg
	Copyright (C) 2007 Nicklas Nordborg

	This file is part of BASE - BioArray Software Environment.
	Available at http://base.thep.lu.se/

	BASE is free software; you can redistribute it and/or
	modify it under the terms of the GNU General Public License
	as published by the Free Software Foundation; either version 3
	of the License, or (at your option) any later version.

	BASE is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with BASE. If not, see <http://www.gnu.org/licenses/>.
*/
package net.sf.basedb.core.query;

import java.util.Arrays;
import java.util.Collection;
import java.util.Collections;

import net.sf.basedb.core.DbControl;
import net.sf.basedb.core.BaseException;
import net.sf.basedb.util.EqualsHelper;

/**
	Compare if two expressions are inequal: e1 &lt;&gt; e2, or
	NOT e1 IS NULL

	@author Samuel, Nicklas
	@version 2.0
	@see Restrictions#neq(Expression, Expression)
	@base.modified $Date: 2015-05-12 11:27:08 +0200 (Tue, 12 May 2015) $
*/
public class NeqRestriction
	implements Restriction
{
	private final Expression e1;
	private final Expression e2;

	NeqRestriction(Expression e1, Expression e2)
	{
		assert e1 != null || e2 != null: "e1 == e2 == null";
		if (e1 == null)
		{
			// Switch to always make e2 contain the null
			e1 = e2;
			e2 = null;
		}
		this.e1 = e1;
		this.e2 = e2;
	}

	/*
		From the QueryElement interface
		-------------------------------------------
	*/
	@Override
	public String toQl(Query query, DbControl dc)
		throws BaseException
	{
		if (e2 != null)
		{
			return "(" + e1.toQl(query, dc) + " <> " + e2.toQl(query, dc) + ")";
		}
		else
		{
			return "(NOT " + e1.toQl(query, dc) + " IS NULL)";
		}
	}
	/**
		@return The expressions to test
	*/
	@Override
	public Collection<? extends QueryElement> getChildren()
	{
		return e2 == null ? Collections.singleton(e1) : Arrays.asList(e1, e2);
	}
	// -------------------------------------------
	/*
		From the Object class
		-------------------------------------------
	*/
	@Override
	public String toString()
	{
		if (e2 == null)
		{
			return "(" + e1 + " IS NOT NULL)";
		}
		else
		{
			return "(" + e1 + " <> " + e2 + ")";
		}
	}
	/**
		This restriction is equal to another NeqRestriction if their 
		child restrictions are equal (ignoring the order).
	*/
	@Override
	public boolean equals(Object other)
	{
		if (this == other) return true;
		if (other == null || this.getClass() != other.getClass()) return false;
		NeqRestriction o = (NeqRestriction)other;
		return (EqualsHelper.equals(e1, o.e1) && EqualsHelper.equals(e2, o.e2)) || 
			(EqualsHelper.equals(e1, o.e2) && EqualsHelper.equals(e2, o.e1));
	}
	@Override
	public int hashCode()
	{
		// Important! The hashcode must be the same even if e1 and e2 are switched!
		return EqualsHelper.hashCode(e1, e2);
	}
	// -------------------------------------------

}

