package net.sf.basedb.core.query;

import net.sf.basedb.core.DbControl;
import net.sf.basedb.core.ItemContext;
import net.sf.basedb.core.PropertyFilter;
import net.sf.basedb.util.extensions.Action;
import net.sf.basedb.util.extensions.ActionFactory;
import net.sf.basedb.util.extensions.ClientContext;

/**
	Actions for creating restrictions on queries. The actions are
	used by the {@link ItemContext#configureQuery(DbControl, EntityQuery, ClientContext, boolean)}
	method to let extensions handle some filtering in table lists and synchronization filters.
	<p>
	The {@link ActionFactory} is asked to create a {@link QueryFilterAction} once
	for every call to the <code>configureQuery()</code> method. The {@link ClientContext#getCurrentItem()}
	returns the {@link ItemContext} instance and the {@link ItemContext#getQuery()} returns
	the query that was passed to the <code>configureQuery()</code> method.
	
	Action factories should check if there is a filter on a column they want to handle, for example
	by calling {@link ItemContext#hasPropertyFilter(String)} or {@link ItemContext#hasExtensionFilter(String)}.
	<p>
	All created actions are called multiple times:
	
	<ul>
	<li>{@link #getColumnRestriction(PropertyFilter)} is called once for every column with a filter, including
		filters that normally are handled by the default implementation. Extensions should return null if they
		want BASE to use normal handling of the filter. Columns restrictions are combined row-wise with AND.
		
	<li>{@link #getRowRestriction(int)} is called once for every filter row after all column restrictions
		have been handled. If the extension returns a restriction, it is combined with AND with the other
		column restrictions. Restrictions for different rows are combined with OR.
	
	<li>{@link #getQueryRestriction()} is called once after all row and columns restrictions have been handled.
		The query restrictions is combined with AND.
	</ul>
	
	@author nicklas
	@since 3.18
*/
public interface QueryFilterAction 
	extends Action
{
	
	/**
		Generate a restriction for the given filter. This method is called for all
		columns with a filter. The extension should check {@link PropertyFilter#getProperty()}
		or {@link PropertyFilter#isExtensionFilter(String)} if the filter should be handled
		or not. If a restriction is returned it is used in the query instead of the
		default implementation.
	*/
	public Restriction getColumnRestriction(PropertyFilter filter);
	
	/**
		Generate a restriction for the given filter row. This method is mainly intended
		for multi-column filtering that can be combined into a single restriction that is
		more efficient than multiple per-column restrictions. For example, two column
		restrictions that return a list of ID:s can be combined into a single row-restriction
		with only the overlapping ID:s from the two lists.
	*/
	public Restriction getRowRestriction(int rowNo);
	
	/**
		Generate a restriction for the entire query after all column and row filters has
		been processed.
	*/
	public Restriction getQueryRestriction();
	
}
