/**
	$Id: WhenStatement.java 6576 2014-10-22 16:41:55Z nicklas $

	Copyright (C) 2006 Nicklas Nordborg

	This file is part of BASE - BioArray Software Environment.
	Available at http://base.thep.lu.se/

	BASE is free software; you can redistribute it and/or modify it
	under the terms of the GNU General Public License as published by
	the Free Software Foundation; either version 3 of the License, or
	(at your option) any later version.

	BASE is distributed in the hope that it will be useful, but
	WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
	General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with BASE. If not, see <http://www.gnu.org/licenses/>.
*/
package net.sf.basedb.core.query;

import net.sf.basedb.core.InvalidUseOfNullException;
import net.sf.basedb.util.EqualsHelper;

/**
 	Create a WHEN statement to be used in a {@link CaseExpression}.

	@author nicklas
	@version 2.0
	@base.modified $Date: 2014-10-22 18:41:55 +0200 (Wed, 22 Oct 2014) $
	@see Expressions#caseWhen(Expression, WhenStatement[])
*/
public class WhenStatement
{

	private final Restriction condition;
	private final Expression value;
	

	/**
		Create a new <code>WhenStatement</code>

		@param condition The condition that is controlling if this
			value is returned or not, null is not allowed
		@param value The value to return if the condition is TRUE, 
			null is allowed
	*/
	public WhenStatement(Restriction condition, Expression value)
	{
		if (condition == null) throw new InvalidUseOfNullException("condition");
		this.condition = condition;
		this.value = value;
	}
	
	/**
		Get the condition of this statement.
		@return A <code>Restriction</code> object
	*/
	public Restriction getCondition()
	{
		return condition;
	}
	
	/**
		Get the value returned by this statement
		@return An <code>Expression</code> object
	*/
	public Expression getValue()
	{
		return value;
	}
	
	/*
		From the Object class
		-------------------------------------------
	*/
	@Override
	public String toString()
	{
		StringBuilder sb = new StringBuilder();
		sb.append("WHEN ").append(getCondition().toString());
		sb.append(" THEN ").append(getValue() == null ? "null" : getValue().toString());
		return sb.toString();
	}

	/**
		This statement is equal to another WhenStatement if both
		have equal conditions and values.
	*/
	@Override
	public boolean equals(Object other)
	{
		if (this == other) return true;
		if (other == null || this.getClass() != other.getClass()) return false;
		WhenStatement o = (WhenStatement)other;
		return condition.equals(o.condition) && EqualsHelper.equals(value, o.value);
	}

	@Override
	public int hashCode()
	{
		return condition.hashCode() + (value == null ? 0 : value.hashCode());
	}
	// -------------------------------------------
	
	
	
}
