/**
	$Id $

	Copyright (C) 2015 Nicklas Nordborg

	This file is part of BASE - BioArray Software Environment.
	Available at http://base.thep.lu.se/

	BASE is free software; you can redistribute it and/or
	modify it under the terms of the GNU General Public License
	as published by the Free Software Foundation; either version 3
	of the License, or (at your option) any later version.

	BASE is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with BASE. If not, see <http://www.gnu.org/licenses/>.
*/
package net.sf.basedb.util.annotations;

import java.util.ArrayList;
import java.util.HashSet;
import java.util.List;
import java.util.Set;

import net.sf.basedb.core.Annotatable;
import net.sf.basedb.core.Annotation;
import net.sf.basedb.core.AnnotationSet;
import net.sf.basedb.core.DbControl;
import net.sf.basedb.core.Permission;
import net.sf.basedb.core.signal.ThreadSignalHandler;
import net.sf.basedb.core.snapshot.AnnotationSetSnapshot;
import net.sf.basedb.core.snapshot.AnnotationSnapshot;
import net.sf.basedb.core.snapshot.SnapshotManager;
import net.sf.basedb.util.AnnotationUtil;
import net.sf.basedb.util.filter.Filter;

/**
	Manager class for batch inheriting of annotations. The manager is created
	with a list of {@link InheritSpecification}:s, each one holding a rule for
	inheriting a single annotation type.
	
	Use {@link #process(Annotatable)} to apply all rules to the given item.
	
	@author nicklas
	@since 3.5
*/
public class InheritAnnotationsManager 
{

	private final DbControl dc;
	private final SnapshotManager snapshotManager;
	private final AnnotationTypeFilter removeFilter;
	private final AnnotationTypeFilter resyncFilter;
	private final List<InheritSpecification> specifications;
	
	public InheritAnnotationsManager(DbControl dc, SnapshotManager snapshotManager, List<InheritSpecification> specifications)
	{
		this.dc = dc;
		this.snapshotManager = snapshotManager;
		this.removeFilter = new AnnotationTypeFilter(null);
		this.resyncFilter = new AnnotationTypeFilter(Annotation.Source.CLONED);
		this.specifications = new ArrayList<InheritSpecification>();
		for (InheritSpecification spec : specifications)
		{
			if (spec.getRemove())
			{
				removeFilter.add(spec);
			}
			else if (spec.getResync())
			{
				resyncFilter.add(spec);
			}
			else
			{
				this.specifications.add(spec);
			}
		}
	}

	/**
		Process the given item, inheriting and/or removing annotations as specified by
		the rules.
	*/
	public void process(Annotatable item)
	{
		AnnotationSetSnapshot snapshot = null;
		AnnotationSet as = null;
		if (item.isAnnotated())
		{
			as = item.getAnnotationSet();
			snapshot = snapshotManager.getSnapshot(dc, as.getId());
			
			// Process annotations to remove
			// Only works with directly inherited annotations
			for (AnnotationSnapshot inherited : snapshotManager.findAnnotations(dc, snapshot, removeFilter, false, true))
			{
				as.removeInheritedAnnotation(inherited.getThisAnnotation(dc));
			}
			
			// Process annotations to re-sync
			for (AnnotationSnapshot inherited : snapshotManager.findAnnotations(dc, snapshot, resyncFilter, false, true))
			{
				inherited.getThisAnnotation(dc).resyncCloned();
			}
			
		}
		
		if (specifications.size() == 0) return;	// Nothing more to do
		
		as = item.getAnnotationSet();
		
		Set<Annotatable> allParents = AnnotationUtil.getAllAnnotatableParentItems(dc, item, new UsableParentFilter());
		for (InheritSpecification spec : specifications)
		{
			// Process each inherit specification in turn
			// Keep track of the first one found so we can
			// handle the case when removeExisting=true and noDuplicates=false
			boolean isFirst = true;
			
			for (Annotatable parent : allParents)
			{
				ThreadSignalHandler.checkInterrupted();
				// Check if subtype of parent item matches subtype in specification
				// If not set, all parents are accepted
				if (spec.isAcceptableParent(parent))
				{
					AnnotationSet pas = parent.getAnnotationSet();
					AnnotationSetSnapshot parentSnapshot = snapshotManager.getSnapshot(dc, pas.getId());
				
					// Find annotation on the parent item (the list can have at most one entry)
					List<AnnotationSnapshot> tmp = snapshotManager.findAnnotations(dc, parentSnapshot, spec, true, false);
					if (tmp.size() > 0)
					{
						// We have an annotation that we should inherit
						AnnotationSnapshot toInherit = tmp.get(0);

						// Check if there are any existing annotation in case we need to act on them
						if (isFirst && (spec.getReplaceExisting() || spec.getNoDuplicates()))
						{
							isFirst = false;
							List<AnnotationSnapshot> existing = snapshotManager.findAnnotations(dc, snapshot, spec, false, true);
							
							if (existing.size() > 0)
							{
								// Replace existing annotations
								if (spec.getReplaceExisting())
								{
									for (AnnotationSnapshot e : existing)
									{
										if (e.getInheritedAnnotationId() != toInherit.getThisAnnotationId())
										{
											// We only remove the annotation if it is not the same we are going to add later
											as.removeInheritedAnnotation(e.getThisAnnotation(dc));
										}
										else
										{
											// Found existing annotation;  re-sync if it is a cloned annotation
											e.getThisAnnotation(dc).resyncCloned();
										}
									}
								}
								// Continue with next specification if not allowing duplicates
								else if (spec.getNoDuplicates())
								{
									break;
								}
							}
						}
		
						as.inheritAnnotation(toInherit.getThisAnnotation(dc), spec.getClone());
						
						// Continue with next specification if not allowing duplicates
						if (spec.getNoDuplicates()) break;
					}
				}
				
			}
		}
		
	}
	
	/**
		Filter used to find parent items that are annotated and for which the current user has
		permission to USE the annotations.
	*/
	static class UsableParentFilter
		implements Filter<Annotatable>
	{
		
		UsableParentFilter()
		{}
		
		@Override
		public boolean evaluate(Annotatable a)
		{
			return a.isAnnotated() && a.hasPermission(Permission.USE);
		}
	}
	
	/**
	 	Filter for finding annotation types to process.
	*/
	static class AnnotationTypeFilter 
		implements Filter<AnnotationSnapshot>
	{

		private final Annotation.Source source;
		private final Set<Integer> annotationTypes;
	
		AnnotationTypeFilter(Annotation.Source source) 
		{
			this.source = source;
			this.annotationTypes = new HashSet<Integer>();
		}

		@Override
		public boolean evaluate(AnnotationSnapshot snapshot) 
		{
			boolean sourceMatch = source == null || source.equals(snapshot.getSource());
			boolean typeMatch = annotationTypes.contains(snapshot.getAnnotationTypeId());
			return sourceMatch && typeMatch;
		}
		
		void add(InheritSpecification spec)
		{
			annotationTypes.add(spec.getAnnotationType().getId());
		}
	}
}
