package net.sf.basedb.util.excel;

import java.util.HashMap;
import java.util.IdentityHashMap;
import java.util.Locale;
import java.util.Map;

import org.apache.poi.ss.usermodel.CellStyle;
import org.apache.poi.ss.usermodel.DataFormat;
import org.apache.poi.ss.usermodel.Font;
import org.apache.poi.ss.usermodel.Workbook;
import org.apache.poi.ss.util.DateFormatConverter;

/**
	Helper class for keeping track of formatting styles that has
	been created for given Excel worksheet. An instance of this
	class is needed when calling {@link ExcelValue#writeTo(org.apache.poi.ss.usermodel.Cell, CellStyleCreator)}.
	
	@author nicklas
	@since 3.15
*/
public class CellStyleCreator 
{

	/**
		Converts a Java date format string to the corresponding Excel
		date format string.
	*/
	public static String convertJavaDateFormatToExcelFormat(String javaFormat)
	{
		if (javaFormat == null) return null;
		String excelFormat = DateFormatConverter.convert(Locale.getDefault(), javaFormat);
		
		// Get rid of Locale prefix since [$-xxxx] and suffix ;@
		String prefix = DateFormatConverter.getPrefixForLocale(Locale.getDefault());
		int clipStart = excelFormat.startsWith(prefix) ? prefix.length() : 0;
		int clipEnd = excelFormat.endsWith(";@") ? 2 : 0;
		if (clipStart > 0 || clipEnd > 0)
		{
			excelFormat = excelFormat.substring(clipStart, excelFormat.length() - clipEnd);
		}
		return excelFormat;
	}

	
	private final Workbook workbook;
	private final DataFormat dataFormat;
	// Base style --> Format string --> Final cell style
	private final Map<CellStyle, Map<String, CellStyle>> styles;
	private CellStyle baseStyle;

	/**
		Creates a new style creator for the given Excel workbook.
	*/
	public CellStyleCreator(Workbook workbook)
	{
		this.workbook = workbook;
		this.dataFormat = workbook.createDataFormat();
		this.styles = new IdentityHashMap<>();
	}
	
	/**
		Create a new cell style. 
	*/
	public CellStyle createStyle()
	{
		return workbook.createCellStyle();
	}
	
	/**
		Create a new font style. 
	*/
	public Font createFont()
	{
		return workbook.createFont();
	}
	
	/**
		Get the current base style for new cells.
	*/
	public CellStyle getBaseStyle()
	{
		return baseStyle;
	}
	
	/**
		Set a base style to use for new cells. When a base style has been set it used
		as a template for all new styles that are created with {@link #getCellStyle(String)}.
		This can be useful for setting different fonts, colors, etc.
	*/
	public void setBaseStyle(CellStyle baseStyle)
	{
		this.baseStyle = baseStyle;
	}
	
	/**
		Get the cell style for formatting numbers without decimals.
		This is equivalent to {@link #getCellStyle(String)} with "0"
		as parameter.
	*/
	public CellStyle getIntCellStyle()
	{
		return getCellStyle("0");
	}
	
	/**
		Get the cell style for the given date format string. If a style
		already exists it is re-used, otherwise a new style is created.
		Note that if the format string is a date format string, it must be
		in Java format. Use {@link #getCellStyle(String)} if the format
		string is in Excel format.
		@param javaFormat The Java format string 
	*/
	public CellStyle getDateCellStyle(String javaFormat)
	{
		String format = convertJavaDateFormatToExcelFormat(javaFormat);
		return getCellStyle(format);
	}
	
	/**
		Get the cell style for the given format string. If a style
		already exists it is re-used, otherwise a new style is created.
		Note that if the format string is a date format string, it must be
		in Excel format. Use {@link #getDateCellStyle(String)} if the format
		string is in Java format.
		
		@param format The format string
	*/
	public CellStyle getCellStyle(String format)
	{
		if (format == null) return baseStyle;
		Map<String, CellStyle> formats = styles.get(baseStyle);
		if (formats == null)
		{
			formats = new HashMap<>();
			styles.put(baseStyle, formats);
		}
		CellStyle style = formats.get(format);
		if (style == null)
		{
			style = workbook.createCellStyle();
			if (baseStyle != null) style.cloneStyleFrom(baseStyle);
			style.setDataFormat(dataFormat.getFormat(format));
			formats.put(format, style);
		}
		return style;
	}
	
}
