package net.sf.basedb.util.excel;

import java.io.ByteArrayInputStream;
import java.io.IOException;
import java.io.InputStream;
import java.nio.charset.Charset;
import java.nio.charset.StandardCharsets;
import java.text.NumberFormat;
import java.util.ArrayList;
import java.util.Date;
import java.util.HashMap;
import java.util.Iterator;
import java.util.List;
import java.util.Map;

import org.apache.poi.poifs.filesystem.FileMagic;
import org.apache.poi.ss.formula.eval.NotImplementedException;
import org.apache.poi.ss.formula.eval.NotImplementedFunctionException;
import org.apache.poi.ss.usermodel.Cell;
import org.apache.poi.ss.usermodel.CellType;
import org.apache.poi.ss.usermodel.DataFormatter;
import org.apache.poi.ss.usermodel.DateUtil;
import org.apache.poi.ss.usermodel.FormulaError;
import org.apache.poi.ss.usermodel.FormulaEvaluator;
import org.apache.poi.ss.usermodel.Row;
import org.apache.poi.ss.usermodel.Sheet;
import org.apache.poi.ss.usermodel.Workbook;
import org.apache.poi.xssf.usermodel.XSSFWorkbook;

import net.sf.basedb.core.File;
import net.sf.basedb.core.Type;
import net.sf.basedb.util.FileUtil;
import net.sf.basedb.util.InputStreamCombiner;
import net.sf.basedb.util.encode.EncoderDecoder;
import net.sf.basedb.util.encode.ToSpaceEncoderDecoder;
import net.sf.basedb.util.formatter.Formatter;

/**
	Utility class for converting Excel worksheets to CSV-formatted InputStreams.
	
	NOTE! If the Excel worksheet contains numbers and/or date values it is recommended
	that a number and/or date formatter is specified. If not, the values will be converted 
	to text using the formatting specification from the Excel workbook which may depende on 
	locale settings and other stuff (=the results may be unpredicatable). So far we only
	support for one formatter of each type. Future use cases may require different formatters
	for different columns, but we will wait with this until it is needed.
	
	@since 3.15
*/
public class XlsxToCsvUtil
{

	/**
		The MIME type for Excel workbook files.
	*/
	public static final String XLSX_MIME_TYPE = "application/vnd.openxmlformats-officedocument.spreadsheetml.sheet";
	
	
	/**
		Check the MIME type and file extension to see if the file could be an Excel file.
		The file contents is not checked.
	*/
	public static boolean seemsLikeAnExcelFile(File file)
	{
		if (file == null) return false;
		return (XLSX_MIME_TYPE.equals(file.getMimeType()) || file.getName().endsWith(".xlsx"));
	}
	
	/**
		Check if the given stream is an Excel file by looking at the first few bytes.
		The stream must support seek(). Call {@link #prepareForCheck(InputStream)}
		before calling this method if it is not sure.
	*/
	public static boolean isAnExcelFile(InputStream in)
	{
		try
		{
			return FileMagic.valueOf(in) == FileMagic.OOXML;
		}
		catch (Exception ex)
		{}
		return false;
	}
	
	/**
		Prepare the input stream for checking with 
		{@link #isAnExcelFile(InputStream)}. If the given stream
		supports seek() it is returned as it is, otherwise it is
		wrapped by a buffered stream.
	*/
	public static InputStream prepareForCheck(InputStream in)
	{
		return FileMagic.prepareToCheckMagic(in);
	}

	private Charset charset;
	private String columnSeparator;
	private String rowSeparator;
	private boolean writeTrailingColumns;
	private boolean evaluateFormulas;
	private EncoderDecoder encoder;
	private Formatter<Date> dateFormat;
	private Formatter<Date> timestampFormat;
	private Formatter<Number> numberFormat;
	
	private Workbook workbook;
	private List<String> sheetNames;
	private Map<String, SheetInfo> sheets;
	
	/**
		Create a new instance using defult settings.
	*/
	public XlsxToCsvUtil() 
	{
		this.columnSeparator = "\t";
		this.rowSeparator = "\n";
		this.writeTrailingColumns = true;
		this.evaluateFormulas = false;
		this.charset = StandardCharsets.UTF_8;
		this.encoder = new ToSpaceEncoderDecoder();
		this.sheets = new HashMap<>();
	}
	
	/**
		Get the current column separator. The default is {tab}.
	*/
	public String getColumnSeparator()
	{
		return columnSeparator;
	}
	/**
		Set the column separator.
	*/
	public void setColumnSeparator(String columnSeparator) 
	{
		this.columnSeparator = columnSeparator;
	}

	/**
		Get the current row separator. The default is {newline}.
	*/
	public String getRowSeparator()
	{
		return rowSeparator;
	}
	/**
		Set the row separator.
	*/
	public void setRowSeparator(String rowSeparator) 
	{
		this.rowSeparator = rowSeparator;
	}

	/**
		If TRUE, trailing empty columns are written to the CSV file so
		that all lines have the same number of columns. If FALSE, 
		trailing emtpy columns are skipped. The default is TRUE.
	*/
	public boolean getWriteTrailingColumns()
	{
		return writeTrailingColumns;
	}
	
	public void setWriteTrailingColumns(boolean writeTrailingColumns)
	{
		this.writeTrailingColumns = writeTrailingColumns;
	}
	
	/**
		If TRUE, formulas are evaluated and the result is written 
		to the CSV file. If FALSE, the formula expression is written.
		The default is FALSE.
	*/
	public boolean getEvaluateFormulas()
	{
		return evaluateFormulas;
	}
	
	public void setEvaluateFormulas(boolean evaluateFormulas)
	{
		this.evaluateFormulas = evaluateFormulas;
	}
	
	/**
		Get the current character set. The default is UTF-8.
	*/
	public Charset getCharset()
	{
		return charset;
	}
	
	/**
		Set the character set to use for the text.
	*/
	public void setCharset(Charset charset)
	{
		this.charset = charset;
	}

	/**
		Get the current encoder. The default is the ToSpaceEncoderDecoder.
	*/
	public EncoderDecoder getEncoder()
	{
		return encoder;
	}
	
	/**
		Set the encoder to use for converting text values so that they
		do not conflict with column or row separators.
	*/
	public void setEncoder(EncoderDecoder encoder)
	{
		this.encoder = encoder;
	}
	
	/**
		Get the date formatter that is used for formatting cells with date-like values.
		If null, the default implementation is used (which may not be predicatable).
	*/
	public Formatter<Date> getDateFormat()
	{
		return dateFormat;
	}
	
	/**
		Set a date formatter to use for formatting cells with date-like values.
		If null, the default implementation is used (which may not be predicatable).
	*/
	public void setDateFormat(Formatter<Date> dateFormat)
	{
		this.dateFormat = dateFormat;
	}
	
	/**
		Get the date formatter that is used for formatting cells with timestamp-like values.
		If null, the default implementation is used (which may not be predicatable).
	*/
	public Formatter<Date> getTimestampFormat()
	{
		return timestampFormat;
	}
	
	/**
		Set a date formatter to use for formatting cells with timestamp-like values.
		If null, the default implementation is used (which may not be predicatable).
	*/
	public void setTimestampFormat(Formatter<Date> timestampFormat)
	{
		this.timestampFormat = timestampFormat;
	}

	
	/**
		Get the formatter that is used for formatting cells with numeric values.
		If null, the default implementation is used (which may not be predicatable).
	*/
	public Formatter<Number> getNumberFormat()
	{
		return numberFormat;
	}
	
	/**
		Set a formatter that is used for formatting cells with numeric values.
		If null, the default implementation is used (which may not be predicatable).
	*/
	public void setNumberFormat(Formatter<Number> numberFormat)
	{
		this.numberFormat = numberFormat;
	}
	
	/**
		Read an Excel workbook from the stream.
	*/
	public void readWorkbook(InputStream xlsx)
		throws IOException
	{
		this.workbook = new XSSFWorkbook(xlsx);
		this.sheetNames = new ArrayList<>(workbook.getNumberOfSheets());
		Iterator<Sheet> it = workbook.sheetIterator();
		while (it.hasNext()) sheetNames.add(it.next().getSheetName());
	}
	
	/**
		Get the current workbook.
	*/
	public Workbook getWorkbook()
	{
		return workbook;
	}
	
	/**
		Get the number of sheets in the workbook.
	*/
	public int getNumSheets()
	{
		return workbook.getNumberOfSheets();
	}
	
	/**
		Get the names of the sheets in the workbook.
	*/
	public List<String> getSheetNames()
	{
		return sheetNames;
	}
	
	/**
		Convert the given named sheet to a CSV InputStream.
	*/
	public SheetInfo getSheetAsCsv(String name)
	{
		SheetInfo si = sheets.get(name);
		if (si == null)
		{
			Sheet sheet = workbook.getSheet(name);
			if (sheet == null) throw new IllegalArgumentException("No sheet with name: " + name);
			si = new SheetInfo(sheet);
			sheets.put(name, si);
		}
		return si;
	}

	/**
		Convert the given sheet to an CSV InputStream. The first sheet has index = 0.
	*/
	public SheetInfo getSheetAsCsv(int index)
	{
		Sheet sheet = workbook.getSheetAt(index);
		SheetInfo si = sheets.get(sheet.getSheetName());
		if (si == null) 
		{
			si = new SheetInfo(sheet);
			sheets.put(si.getName(), si);
		}
		return si;
	}

	/**
		Convert all sheets in the workbook to a single CSV-like stream. Each sheet 
		is converted as if {@link SheetInfo#parseToCsv()} has been called. To
		separate the sheets each section is started with [Name-of-sheet] and 
		then followed by the sheet data.
	*/
	public InputStream getWorkbookAsCsv()
	{
		return new WorkbookStream();
	}
	
	
	/**
		Closes the current workbook. It is safe to do this before
		reading the CSV streams to the end.
	*/
	public void close()
	{
		FileUtil.close(workbook);
	}

	/**
		Represents a single sheet in a workbook. Use it to get access to a CSV Input stream 
		that is generated from the table data. It can also be used to get access to the
		actual underlying Cell element for reading data directly.
	*/
	public class SheetInfo
	{
		private final Sheet sheet;
		private final DataFormatter dataFormat;
		private final FormulaEvaluator formulaEval;

		private final int firstRow;
		private final int lastRow;
		
		private final int firstColumn;
		private final int lastColumn;
		
		private byte[] csvBytes;
		
		SheetInfo(Sheet sheet)
		{
			this.sheet = sheet;
			this.dataFormat = new DataFormatter();
			this.formulaEval = evaluateFormulas ? workbook.getCreationHelper().createFormulaEvaluator() : null;
			
			this.firstRow = sheet.getFirstRowNum();
			int maxRow = sheet.getLastRowNum(); // Note! index of last row that contains data!
			int lastRowWithData = firstRow;
			
			// To make sure that columns align between rows, we must find the span of column indexes
			int minCol = Integer.MAX_VALUE;
			int maxCol = -1;
			for (int rowNo = firstRow; rowNo <= maxRow; rowNo++)
			{
				Row row = sheet.getRow(rowNo);
				if (row == null || row.getPhysicalNumberOfCells() == 0) continue; // Empty row -- skip it
				minCol = Math.min(minCol, row.getFirstCellNum());
				int maxRowCol = row.getLastCellNum(); // Note! getLastCellNum() returns index of last cell+1 that contains data!
				
				// Sometimes there are a lot of empty or otherwise unused cells to the right on each row...
				// We try to skip null, BLANK and ERROR cells and FORMULA if we don't evaluate formulas
				CellType cellType = null;
				do
				{
					maxRowCol--;
					if (maxRowCol < 0) break;
					Cell c = row.getCell(maxRowCol);
					cellType = c == null ? CellType.BLANK : c.getCellType();
					if (cellType == CellType.FORMULA && evaluateFormulas) 
					{
						cellType = c.getCachedFormulaResultType();
					}
				} while (cellType == CellType.BLANK || cellType == CellType.ERROR || cellType == CellType.FORMULA);

				// Ok, found data on this row
				if (maxRowCol >= 0) lastRowWithData = rowNo;
				maxCol = Math.max(maxCol, maxRowCol);
			}
			this.lastRow = lastRowWithData;
			this.firstColumn = minCol;
			this.lastColumn = maxCol;
		}
		
		/**
			Get the name of the sheet.
		*/
		public String getName()
		{
			return sheet.getSheetName();
		}
		
		/**
			Get the row number for the first row that contains data.
		*/
		public int getFirstRow()
		{
			return firstRow;
		}
		
		/**
			Get the row number for the last row that contains data.
		*/
		public int getLastRow()
		{
			return lastRow;
		}
		
		/**
			Get the column number for the first column that contains data.
		*/
		public int getFirstColumn()
		{
			return firstColumn;
		}
		
		/**
			Get the column number for the last column that contains data.
		*/
		public int getLastColumn()
		{
			return lastColumn;
		}
		
		/**
			Get the cell at the specified row and column.
			@return A Cell instance or null
		*/
		public Cell getCell(int row, int col)
		{
			Row r = sheet.getRow(row);
			return r == null ? null : r.getCell(col);
		}

		/**
			Get the the cell value as a string. Numeric cells are 
			are checked if the seem to have a date-like format syntax
			(with or without time). If so, the value is formatted with
			the date or timestamp formatter, otherwise with the number
			formatter. If the required formatter has not been set, the
			value is formatted with the built-in formatter from Excel.
		*/
		public String getCellValueAsString(Cell cell)
		{
			if (cell == null) return null;
	
			String value = null;
			CellType cellType = cell.getCellType();
			CellTypeWithInfo info = null;
			if (cellType == CellType.FORMULA && evaluateFormulas)
			{
				info = evaluateFormula(cell);
				cellType = info.cellType;
			}
			if (cellType == CellType.NUMERIC)
			{
				boolean isDate = DateUtil.isCellDateFormatted(cell);
				if (isDate)
				{
					String formatString = cell.getCellStyle().getDataFormatString();
					boolean hasTimePart = hasTimePart(formatString);
					boolean hasDatePart = hasDatePart(formatString);					
					Date d = cell.getDateCellValue();
					if (hasDatePart && hasTimePart)
					{
						if (timestampFormat != null) value = timestampFormat.format(d);
					}
					else if (hasDatePart)
					{
						if (dateFormat != null) value = dateFormat.format(d);
					}
					else
					{
						// Only time part -- let it be handled by the DataFormat
					}
				}
				else
				{
					if (numberFormat != null)
					{
						value = numberFormat.format(cell.getNumericCellValue());
					}
				}
			}
			if (value == null)
			{
				value = dataFormat.formatCellValue(cell, formulaEval);
			}
			if (info != null && info.info != null) value += info.info;
			return value;
		}

		/**
			Get the cell value as a number. If the cell is a numeric cell, the value is returned
			directly, otherwise the given parser is used to parse the string data from cell,
			via {@link Type#parseString(String, NumberFormat, boolean)}
		*/
		public Double getCellValueAsNumber(Cell cell, NumberFormat parser, boolean nullIfException)
		{
			if (cell == null) return null;
			
			Double value = null;
			CellType cellType = cell.getCellType();
			if (cellType == CellType.FORMULA && evaluateFormulas)
			{
				cellType = evaluateFormula(cell).cellType;
			}
			if (cellType == CellType.NUMERIC)
			{
				value = cell.getNumericCellValue();
			}
			else
			{
				String s = dataFormat.formatCellValue(cell, formulaEval);
				value = (Double)Type.DOUBLE.parseString(s, parser, nullIfException);
			}
			return value;
		}
		
		/**
			Get the cell value as a date. If the cell is a numeric cell, the value is returned
			directly as if it was a date (the style format string is NOT checked),
			otherwise the given parser is used to parse the string data from cell, 
			via {@link Type#parseString(String, Formatter, boolean)}
		*/
		public Date getCellValueAsDate(Cell cell, Formatter<Date> parser, boolean nullIfException)
		{
			if (cell == null) return null;
			
			Date value = null;
			CellType cellType = cell.getCellType();
			if (cellType == CellType.FORMULA && evaluateFormulas)
			{
				cellType = evaluateFormula(cell).cellType;
			}
			if (cellType == CellType.NUMERIC)
			{
				value = cell.getDateCellValue();
			}
			else
			{
				String s = dataFormat.formatCellValue(cell, formulaEval);
				value = (Date)Type.DATE.parseString(s, parser, nullIfException);
			}
			return value;
		}
		
		private CellTypeWithInfo evaluateFormula(Cell cell)
		{
			CellTypeWithInfo ct = new CellTypeWithInfo();
			try
			{
				ct.cellType = formulaEval.evaluateFormulaCell(cell);
			}
			catch (NotImplementedException ex)
			{
				cell.setCellErrorValue(FormulaError.NAME.getCode());
				ct.cellType = CellType.ERROR;
				if (ex.getCause() instanceof NotImplementedFunctionException)
				{
					NotImplementedFunctionException nof = (NotImplementedFunctionException)ex.getCause();
					ct.info = nof.getFunctionName();
				}
			}
			catch (RuntimeException ex)
			{
				cell.setCellErrorValue(FormulaError.VALUE.getCode());
				ct.cellType = CellType.ERROR;
			}
			return ct;
		}
		
		/**
			Convert the sheet to a CSV and get an InputStream for reading the data.
			Note that the CSV will only contain data for the range of cells between
			firstRow, lastRow, firstColumn and lastColumn.
		*/
		public InputStream parseToCsv()
		{
			if (csvBytes == null) parseToBytes();
			return new ByteArrayInputStream(csvBytes);
		}
		
		/**
			Get the size of the CSV in bytes.
			@since 3.17.1
		*/
		public int getSize()
		{
			if (csvBytes == null) parseToBytes();
			return csvBytes.length;
		}
		
		private void parseToBytes()
		{
			StringBuilder out = new StringBuilder();
			for (int rowNo = firstRow; rowNo <= lastRow; rowNo++)
			{
				Row row = sheet.getRow(rowNo);
				if (row == null || row.getPhysicalNumberOfCells() == 0) 
				{
					if (writeTrailingColumns)
					{
						for (int colNo = firstColumn; colNo < lastColumn; colNo++)
						{
							out.append(columnSeparator);
						}
					}
				}
				else
				{
					int lastRowCol = writeTrailingColumns ? lastColumn : row.getLastCellNum()-1;  // Note! getLastCellNum() returns index of last cell+1 that contains data!
					for (int colNo = firstColumn; colNo <= lastRowCol; colNo++)
					{
						if (colNo > firstColumn) out.append(columnSeparator);
						Cell cell = row.getCell(colNo);
						String value = getCellValueAsString(cell);
						out.append(value == null ? "" : encoder.encode(value));
					}
				}
				out.append(rowSeparator);
			}
			csvBytes = out.toString().getBytes(charset);
		}
	}
	
	private boolean hasTimePart(String formatString)
	{
		if (formatString == null) return false;
		return formatString.contains("S") || formatString.contains("H");
	}
	
	private boolean hasDatePart(String formatString)
	{
		if (formatString == null) return false;
		return formatString.contains("Y") || formatString.contains("D");
	}

	/**
		Input stream implementation that combines streams from all sheets in the current workbook.
	*/
	class WorkbookStream
		extends InputStream
	{
		private SheetInfo currentSheet;
		private InputStream currentStream;
		private int nextIndex;
		
		WorkbookStream()
		{
			this.nextIndex = 0;
			nextSheet();
		}

		/**
			Switch to the next worksheet.
		*/
		private void nextSheet()
		{
			FileUtil.close(currentStream);
			if (nextIndex < getNumSheets())
			{
				currentSheet = getSheetAsCsv(nextIndex);
				currentStream = getPrefixedStream(currentSheet);
				nextIndex++;
			}
			else
			{
				currentSheet = null;
				currentStream = null;
			}
		}
		
		/**
			Create a combined stream by adding a "prefix": [sheet-name]
			and then the regular CSV stream for the given worksheet.
		*/
		private InputStream getPrefixedStream(SheetInfo sheet)
		{
			String prefix = "["+encoder.encode(sheet.getName())+"]"+rowSeparator;
			return new InputStreamCombiner(
				new ByteArrayInputStream(prefix.getBytes(charset)), 
				sheet.parseToCsv());
		}

		
		@Override
		public int available() 
			throws IOException
		{
			return currentStream != null ? currentStream.available() : 0;
		}

		@Override
		public void close() 
			throws IOException 
		{
			if (currentStream != null) currentStream.close();
			currentStream = null;
			currentSheet = null;
		}

		@Override
		public int read() 
			throws IOException
		{
			int result = -1;
			while (result == -1 && currentStream != null)
			{
				result = currentStream.read();
				if (result == -1) nextSheet();
			}
			return result;
		}

		@Override
		public int read(byte[] b, int off, int len)
			throws IOException 
		{
			int result = -1;
			while (result == -1 && currentStream != null)
			{
				result = currentStream.read(b, off, len);
				if (result == -1) nextSheet();
			}
			return result;
		}
		
	}
	
	// Used to return some error information if formula evaluation fails
	static class CellTypeWithInfo
	{
		CellType cellType;
		String info;
	}
	
}
