/**
	$Id: TableWriter.java 7675 2019-03-27 08:34:35Z nicklas $

	Copyright (C) 2009 Nicklas Nordborg

	This file is part of BASE - BioArray Software Environment.
	Available at http://base.thep.lu.se/

	BASE is free software; you can redistribute it and/or
	modify it under the terms of the GNU General Public License
	as published by the Free Software Foundation; either version 3
	of the License, or (at your option) any later version.

	BASE is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with BASE. If not, see <http://www.gnu.org/licenses/>.
*/
package net.sf.basedb.util.export;

import java.io.PrintWriter;
import java.io.Writer;
import java.util.Collections;
import java.util.HashMap;
import java.util.Map;

import net.sf.basedb.util.encode.EncoderDecoder;
import net.sf.basedb.util.encode.IdentityEncoderDecoder;
import net.sf.basedb.util.formatter.Formatter;

/**
	Wraps a writer stream and provides methods for easier writing
	of tabular data. The writer can be configured with a 
	{@link #setDataSeparator(String)} (default is 'tab'),
	{@link #setNullValue(String)} (default is an empty string)
	and an {@link #setEncoder(EncoderDecoder)} (no default). 
	Use {@link #tablePrintData(Object...)} to print a singe line 
	with data. Other {@link PrintWriter} methods are not affected 
	by those settings.

	@author Nicklas
	@version 2.12
	@base.modified $Date: 2019-03-27 09:34:35 +0100 (Wed, 27 Mar 2019) $
*/
public class TableWriter
	extends PrintWriter
{
	
	private String dataSeparator = "\t";
	private String nullValue = "";
	private EncoderDecoder encoder;
	private final Map<Integer, Formatter<?>> formatters;

	/**
		Create a new table writer that is writing it's output to the
		given writer.
		@param out The writer to write the data to
	*/
	public TableWriter(Writer out)
	{
		super(out);
		this.formatters = new HashMap<>();
	}

	/**
		The data separator string is written between
		each data column in the output. The default value is 'tab'.
		@return The current value of the data separator string
	*/
	public String getDataSeparator()
	{
		return dataSeparator;
	}
	
	/**
		Change the data separator string. Note! If this value is modified
		then a different encoder is probably also needed. See
		{@link #setEncoder(EncoderDecoder)}.
		@param dataSeparator The new data separator
	*/
	public void setDataSeparator(String dataSeparator)
	{
		this.dataSeparator = dataSeparator;
	}

	/**
		The null value string is written to the output instead of
		null data values. The default value is an empty string.
		@return The current value of the null value string
	*/
	public String getNullValue()
	{
		return nullValue;
	}
	
	/**
		Change the null value string.
		@param nullValue The new null value string
	*/
	public void setNullValue(String nullValue)
	{
		this.nullValue = nullValue;
	}

	/**
		Get the current default encoder for this writer.
		@return The current encoder or null
		@since 2.15
	*/
	public EncoderDecoder getEncoder()
	{
		return encoder;
	}
	
	/**
		Change the encoder.
		@param encoder The new encoder or null to not encode
		@since 2.15
	*/
	public void setEncoder(EncoderDecoder encoder)
	{
		this.encoder = encoder;
	}

	/**
		Encode the given string with the default encoder.
		@param s The string to encode
		@return The encoded string, or the argument if no default encoder
			has been specified
		@since 2.15
	*/
	public String tableEncode(String s)
	{
		EncoderDecoder e = getEncoder();
		return e == null ? s : e.encode(s);
	}

	/**
		Set a data formatter for a given column. From now on, the
		given formatter will be used instead of the toString() method 
		to convert data values in that column. It is important that
		the data is of the correct type for the formatter.
		@since 3.15
	*/
	public void setColumnFormatter(int index, Formatter<?> formatter)
	{
		this.formatters.put(index, formatter);
	}
	
	/**
		@see #setColumnFormatter(int, Formatter)
		@since 3.15
	*/
	public Formatter<?> getColumnFormatter(int index)
	{
		return formatters.get(index);
	}

	
	/**
		Print a header line to the parent writer. The only difference between
		this method and the {@link #tablePrintData(Object...)} method is that
		no column formatters are used (since they operate on data and the headers
		are typically strings).
		@since 3.15
	*/
	public void tablePrintHeaders(Object... headers)
	{
		tablePrintData(getEncoder(), null, headers);
	}
	
	/**
		Print a data line to the parent writer using the
		default encoder and column formatters.

		@param data The values to write
	*/
	public void tablePrintData(Object... data)
	{
		tablePrintData(getEncoder(), formatters, data);
	}
	
	/**
		Print a data line to the parent writer. Each value is separated by the
		data separator string ('tab' by default). Null values are replaced 
		with the null value string (empty string by default). Non-null objects
		are converted to strings values by calling either their toString() method or
		using the formatter assigned for each column. If an encoder has been specified, 
		the value is finally encoded with {@link EncoderDecoder#encode(String)}.
	
		@param data The values to write
		@param encoder The encoder to use, or null to not use any encoder
		@since 2.15
	*/
	public void tablePrintData(EncoderDecoder encoder, Object... data)
	{
		tablePrintData(encoder, formatters, data);
	}
	
	/**
		Print a data line to the parent writer. Each value is separated by the
		data separator string ('tab' by default). Null values are replaced 
		with the null value string (empty string by default). Non-null objects
		are converted to strings values by calling either their toString() method or
		using the formatter assigned for each column. If an encoder has been specified, 
		the value is finally encoded with {@link EncoderDecoder#encode(String)}.
	
		@param data The values to write
		@param formatters Column formatters to use (null to always use the toString() method)
		@param encoder The encoder to use, or null to not use any encoder
		@since 2.15
	*/
	@SuppressWarnings({ "rawtypes", "unchecked" })
	public void tablePrintData(EncoderDecoder encoder, Map<Integer, Formatter<?>> formatters, Object... data)
	{
		String ds = getDataSeparator();
		String nv = getNullValue();
		if (encoder == null) encoder = new IdentityEncoderDecoder();
		if (formatters == null) formatters = Collections.emptyMap();
		for (int col = 0; col < data.length; col++)
		{
			Object d = data[col];
			if (col > 0) print(ds);
			Formatter fmt = formatters.get(col);
			print(d == null ? nv : encoder.encode(fmt == null ? d.toString() : fmt.format(d)));
		}
		print("\n");
	}
	
}
