/*
	$Id: RawFunction.java 6875 2015-04-20 09:08:18Z nicklas $

	Copyright (C) 2005 Nicklas Nordborg
	Copyright (C) 2006 Jari Häkkinen, Nicklas Nordborg

	This file is part of BASE - BioArray Software Environment.
	Available at http://base.thep.lu.se/

	BASE is free software; you can redistribute it and/or
	modify it under the terms of the GNU General Public License
	as published by the Free Software Foundation; either version 3
	of the License, or (at your option) any later version.

	BASE is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with BASE. If not, see <http://www.gnu.org/licenses/>.
*/
package net.sf.basedb.util.jep;

import net.sf.basedb.core.BaseException;
import net.sf.basedb.core.DynamicQuery;
import net.sf.basedb.core.PermissionDeniedException;
import net.sf.basedb.core.RawDataType;
import net.sf.basedb.core.DbControl;
import net.sf.basedb.core.Metadata;
import net.sf.basedb.core.data.RawData;
import net.sf.basedb.core.query.Dynamic;
import net.sf.basedb.core.query.Expression;
import net.sf.basedb.util.BioAssaySetUtil;

import java.util.Stack;

import org.nfunk.jep.Node;
import org.nfunk.jep.ParseException;


/**
	A JEP function class that adds a <code>raw(string)</code> function to a 
	JEP expression parser. The function will look up the value of the raw data 
	property with the given name. For example: 
	<code>raw('ch1FgMean') - raw('ch1BgMean')</code>
	<p>
	To be able to use this function it must be registered with the JEP
	parser and, before the expression is evaluated, a raw data object must be set.
	For example we can evaluate an expression for every raw data object
	in a raw bioassay:
	
	<pre class="code">
DbControl dc = ...
RawBioAssay assay = ...
String expression = "raw('ch1FgMean') - raw('ch1BgMean')";
RawFunction raw = new RawFunction(assay.getRawDataType());
JEP jep = JepUtil.newJep(expression, raw);
DataResultIterator&lt;RawData&gt; result = 
   assay.getRawData().iterate(dc);
while (result.hasNext())
{
   raw.setRawData(result.next());
   double value = jep.getValue();
   // Do something with the value
}
result.close();
</pre>

	@author Nicklas
	@version 2.0
	@base.modified $Date: 2015-04-20 11:08:18 +0200 (Mon, 20 Apr 2015) $
	@see Jep
	@see BioAssaySetUtil#createJepExpression(DbControl, String, DynamicQuery)
*/
public class RawFunction
	implements JepExpressionFunction
{

	private final RawDataType rawDataType;
	private final Metadata<RawData> metaData;
	private final DbControl dc;
	private final boolean allowUse;
	private int numParameters;
	private RawData rawData;

	/**
		Create a new instance of this function. The new instance cannot be used
		to dynamically evaluate expressions. It should only be used for converting
		JEP formulas to {@link Expression}:s.
		@see Jep#formulaToExpression(String, JepFunction[])
	*/
	public RawFunction(boolean allowUse)
	{
		this.dc = null;
		this.metaData = null;
		this.rawDataType = null;
		this.allowUse = allowUse;
	}
	
	/**
		Create a new instance of this function working with raw data of
		the specified raw data type.
		@param rawDataType The type of raw data this function uses
	*/
	public RawFunction(DbControl dc, RawDataType rawDataType, boolean allowUse)
	{
		this.dc = dc;
		this.rawDataType = rawDataType;
		this.metaData = Metadata.getInstance(RawData.class, rawDataType.getEntityName());
		this.allowUse = allowUse;
	}
	
	/*
		From the JepFunction interface
		-------------------------------------------
	*/
	/**
		@return The string "raw"
	*/
	@Override
	public String getFunctionName()
	{
		return "raw";
	}
	// -------------------------------------------
	/*
		From the JepExpressionFunction interface
		-------------------------------------------
	*/
	/**
		Use the {@link Dynamic#rawData(String)} method to create an
		expression referencing a raw data property.
	*/
	@Override
	public Expression toExpression(Node node)
	{
		if (!allowUse)
		{
			throw new PermissionDeniedException("Not allowed to use 'raw' function with merged data");
		}
		int numChildren = node.jjtGetNumChildren();
		if (numChildren != 1)
		{
			throw new BaseException("Invalid number of expressions for 'raw' function: " + numChildren);
		}
		return Dynamic.rawData(Jep.nodeToString(node.jjtGetChild(0)));
	}
	// -------------------------------------------
	/*
		From the PostfixMathCommandI interface
		-------------------------------------------
	*/
	/**
		@return Always 1
	*/
	@Override
	public int getNumberOfParameters()
	{
		return 1;
	}
	@Override
	public void setCurNumberOfParameters(int n)
	{
		this.numParameters = n;
	}
	@Override
	public boolean checkNumberOfParameters(int n)
	{
		return n == 1;
	}
	@Override
	@SuppressWarnings({"unchecked", "rawtypes"})
	public void run(Stack stack)
		throws ParseException
	{
		if (stack == null || stack.empty()) 
		{
			throw new ParseException("Stack is empty");
		}
		Object propertyName = stack.pop();
		if (propertyName instanceof String)
		{
			stack.push(raw((String)propertyName));
		}
		else
		{
			throw new ParseException("Invalid parameter type: " + propertyName + "; expected string");
		}
	}
	// -------------------------------------------

	/**
		Get the raw data type this function uses.
	*/
	public RawDataType getRawDataType()
	{
		return rawDataType;
	}
	/**
		Set a new raw data object that will be used the next time the
		JEP expression is evaluated.
		@param rawData The raw data object to use
	*/
	public void setRawData(RawData rawData)
	{
		this.rawData = rawData;
	}
	/**
		Get the current raw data object used when evaluating the JEP expression.
	*/
	public RawData getRawData()
	{
		return rawData;
	}
	
	/**
		Get the value of the specified property of the current raw data object.
		This method uses Hibernate metadata to find the value.
		@param propertyName The name of the raw data property. Null is not allowed.
	 	@return The value as a java.lang.Object 
	 	@throws ParseException If required arguments is null 
	*/
	public Object raw(String propertyName)
		throws ParseException
	{
		if (rawData == null || propertyName == null)
		{
			throw new ParseException("No raw data object has been specified for function raw('"+propertyName+"')");
		}
		return metaData.getPropertyPath(propertyName, false).getValue(dc, rawData);
	}
}
