/**
	$Id$

	Copyright (C) 2015 Nicklas Nordborg

	This file is part of BASE - BioArray Software Environment.
	Available at http://base.thep.lu.se/

	BASE is free software; you can redistribute it and/or
	modify it under the terms of the GNU General Public License
	as published by the Free Software Foundation; either version 3
	of the License, or (at your option) any later version.

	BASE is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with BASE. If not, see <http://www.gnu.org/licenses/>.
*/
package net.sf.basedb.util.listable;

import java.util.HashSet;
import java.util.Set;

import net.sf.basedb.core.DbControl;
import net.sf.basedb.core.Item;
import net.sf.basedb.core.ItemQuery;
import net.sf.basedb.core.Sample;
import net.sf.basedb.core.query.Expressions;
import net.sf.basedb.core.query.Hql;
import net.sf.basedb.core.query.Restrictions;

/**
	Source item transformer implementation that transform sample items to their
	parent sample items. From the given set of sample ID:s the transformer will
	recursively load as many parent samples as possible. The source items
	may optionally be included in the target set.

	@author Nicklas
	@since 3.5
*/
public class SampleToParentSampleTransformer 
	extends AbstractSourceItemTransformer
{

	private final boolean includeSourcesInTarget;
	private boolean pushOnly;
	
	/**
		Create a new sample to parent sample transformer.
		@param includeSourcesInTarget TRUE to include the source items in the target
	*/
	public SampleToParentSampleTransformer(boolean includeSourcesInTarget) 
	{
		super(Item.SAMPLE, Item.SAMPLE);
		this.includeSourcesInTarget = includeSourcesInTarget;
	}

	/**
		Create a new sample to parent sample transformer that only load parents
		to samples that has a subtype with "push annotations" set. This functionality
		should only be used in the first transformation step.
		@see PushToParentsTransformer
		@since 3.16
	*/
	public SampleToParentSampleTransformer(boolean includeSourcesInTarget, boolean childrensThatPushOnly)
	{
		this(includeSourcesInTarget);
		this.pushOnly = childrensThatPushOnly;
	}

	@Override
	public Set<Integer> transform(TransformContext context, Set<Integer> source) 
	{
		DbControl dc = context.getDbControl();
		
		ItemQuery<Sample> query = Sample.getQuery();
		query.setIncludes(context.getInclude());
		query.join(Hql.innerJoin("childCreationEvents", "ce"));
		query.join(Hql.innerJoin("ce", "event", "evt"));
		query.join(Hql.innerJoin("evt", "bioMaterial", "bm"));
		query.restrict(
			Restrictions.in(
				Hql.alias("bm"), 
				Expressions.parameter("children")
			)
		);
		if (pushOnly)
		{
			query.join(Hql.innerJoin("bm", "itemSubtype", "st"));
			query.restrict(Restrictions.eq(Hql.property("st", "pushAnnotations"), Expressions.bool(true)));
		}
		
		// Keep track of all seen children and parents
		Set<Integer> allParents = new HashSet<Integer>();
		Set<Integer> allChildren = new HashSet<Integer>();
		if (includeSourcesInTarget) allParents.addAll(source);

		// Children that we have not yet checked
		Set<Integer> children = source;
		while (children.size() > 0)
		{
			Set<Integer> parents = safeIdList(dc, query, "children", children);
			
			// Store new parents and "used" children
			allParents.addAll(parents);
			allChildren.addAll(children);
			
			// Remove children that we have already seen...
			// ...the remaining parents become our new children
			parents.removeAll(allChildren);
			children = parents;
		}
		
		return allParents;
	}


}
