/**
	$Id$

	Copyright (C) 2015 Nicklas Nordborg

	This file is part of BASE - BioArray Software Environment.
	Available at http://base.thep.lu.se/

	BASE is free software; you can redistribute it and/or
	modify it under the terms of the GNU General Public License
	as published by the Free Software Foundation; either version 3
	of the License, or (at your option) any later version.

	BASE is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with BASE. If not, see <http://www.gnu.org/licenses/>.
*/
package net.sf.basedb.util.listable;

import java.util.ArrayList;
import java.util.List;

import net.sf.basedb.core.Item;
import net.sf.basedb.core.SyncFilter.SourceItemTransform;

/**
	Transformer factory implementation for transforming to extracts. The current
	implementation support transforming from:
	
	<ul>
	<li>{@link SourceItemTransform#PARENT_TO_CHILD}: BIOSOURCE, SAMPLE, EXTRACT
	<li>{@link SourceItemTransform#CHILD_TO_PARENT}: EXTRACT, PHYSICALBIOASSAY, DERIVEDBIOASSAY, RAWBIOASSAY
	</ul>

	@author Nicklas
	@since 3.5
*/
public class ToExtractSourceItemTransformerFactory 
	extends AbstractSourceItemTransformerFactory
{

	private static Item[] PARENT_TO_CHILD = 
		{ Item.BIOSOURCE, Item.SAMPLE, Item.EXTRACT };
	
	private static Item[] CHILD_TO_PARENT =
		{ Item.EXTRACT, Item.PHYSICALBIOASSAY, Item.DERIVEDBIOASSAY, Item.RAWBIOASSAY };
	
	
	public ToExtractSourceItemTransformerFactory() 
	{
		super(Item.EXTRACT, PARENT_TO_CHILD, CHILD_TO_PARENT);
	}
	
	
	@Override
	public SourceItemTransformer create(final Item sourceItemType, final SourceItemTransform transform) 
	{
		List<SourceItemTransformer> chain = new ArrayList<SourceItemTransformer>();
		Item stepBySource = null;
		
		if (transform == SourceItemTransform.PARENT_TO_CHILD)
		{
			// Transforming a list of parent items to child extracts
			// We do step-wise loading from BIOSOURCE -> SAMPLE -> EXTRACT
			if (sourceItemType == Item.BIOSOURCE)
			{
				// Load child samples of the biosources
				chain.add(new BioSourceToSampleTransformer());
				stepBySource = Item.SAMPLE;
			}
			if (sourceItemType == Item.SAMPLE || stepBySource == Item.SAMPLE)
			{
				// Expand source selection to parents if the children push annotations upwards
				if (sourceItemType == Item.SAMPLE) chain.add(new PushToParentsTransformer(sourceItemType));
				// Load child samples and include the source samples
				chain.add(new SampleToChildSampleTransformer(true));
				// Transform to child extracts
				chain.add(new SampleToExtractTransformer());
				stepBySource = Item.EXTRACT;
			}
			if (sourceItemType == Item.EXTRACT || stepBySource == Item.EXTRACT)
			{
				// Expand source selection to parents if the children push annotations upwards
				if (sourceItemType == Item.EXTRACT) chain.add(new PushToParentsTransformer(sourceItemType));
				// Load child extract and maybe include source extracts
				chain.add(new ExtractToChildExtractTransformer(sourceItemType != Item.EXTRACT));
			}
		}
		else if (transform == SourceItemTransform.CHILD_TO_PARENT)
		{
			childToParentChain(chain, sourceItemType, Item.EXTRACT);
		}

		return chain.size() == 0 ? null : new SourceItemTransformerChain(chain);
	}

	
	/**
		Utility method for creating the transformer chain from child items
		up to extracts. Supported source item types are: RAWBIOASSAY, 
		DERIVEDBIOASSAY, PHYSICALBIOASSAY and EXTRACT
	*/
	static void childToParentChain(List<SourceItemTransformer> chain, Item sourceItemType, Item targetItemType)
	{
		Item stepBySource = null;
		// Transforming a list of child items to parent extracts
		// We do step-wise loading from RAWBIOASSAY -> DERIVEDBIOASSAY -> PHYSICALBIOASSAY -> EXTRACT
		CollectExtracts collectedExtracts = null;
		if (sourceItemType == Item.RAWBIOASSAY)
		{
			// Collect extracts from the source raw bioassays
			collectedExtracts = new CollectExtracts();
			chain.add(collectedExtracts.collectFromRawBioAssays());
			// Load parent derived bioassays
			chain.add(new RawBioAssayToDerivedBioAssayTransformer());
			stepBySource = Item.DERIVEDBIOASSAY;
		}
		if (sourceItemType == Item.DERIVEDBIOASSAY || stepBySource == Item.DERIVEDBIOASSAY)
		{
			// Collect extracts from all derived bioassays in the parent chain
			if (collectedExtracts == null) collectedExtracts = new CollectExtracts();
			chain.add(collectedExtracts.collectFromDerivedBioAssays(true));
			
			// Load physical bioassays
			chain.add(new DerivedBioAssayToPhysicalBioAssayTransformer());
			stepBySource = Item.PHYSICALBIOASSAY;
		}
		if (sourceItemType == Item.PHYSICALBIOASSAY || stepBySource == Item.PHYSICALBIOASSAY)
		{
			// Load all extracts used on the physical bioassays
			chain.add(new PhysicalBioAssayToExtractTransformer());
			stepBySource = Item.EXTRACT;
		}
		if (sourceItemType == Item.EXTRACT || stepBySource == Item.EXTRACT)
		{
			if (collectedExtracts != null)
			{
				// Collected parents and childre to downstreams extracts (linked from bioassays)
				chain.add(collectedExtracts.collectChildrenOfCollected(true));
				chain.add(collectedExtracts.collectParentsOfCollected(true));
			}
			// Load parent extracts, maybe including the source extracts
			chain.add(new ExtractToParentExtractTransformer(sourceItemType != targetItemType, collectedExtracts));
		}
		
	}
}
