/**
	$Id: OverviewContext.java 6875 2015-04-20 09:08:18Z nicklas $

	Copyright (C) 2008 Nicklas Nordborg

	This file is part of BASE - BioArray Software Environment.
	Available at http://base.thep.lu.se/

	BASE is free software; you can redistribute it and/or
	modify it under the terms of the GNU General Public License
	as published by the Free Software Foundation; either version 3
	of the License, or (at your option) any later version.

	BASE is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with BASE. If not, see <http://www.gnu.org/licenses/>.
*/
package net.sf.basedb.util.overview;

import net.sf.basedb.core.BasicItem;
import net.sf.basedb.core.Include;
import net.sf.basedb.core.ItemQuery;
import net.sf.basedb.core.Project;
import net.sf.basedb.core.query.Order;
import net.sf.basedb.core.snapshot.SnapshotManager;
import net.sf.basedb.util.overview.cache.NodeCache;
import net.sf.basedb.util.overview.loader.NodeLoader;
import net.sf.basedb.util.overview.loader.NodeLoaderFactory;
import net.sf.basedb.util.overview.validator.NodeValidator;
import net.sf.basedb.util.overview.validator.NodeValidatorFactory;

/**
	Interface that should be implemented by overview generating
	classes. This interface defines service methods that overview
	generating class to {@link NodeLoader}:s and {@link NodeValidator}:s
	so that they can do their job.
	
	@author Nicklas
	@version 2.10
	@base.modified $Date: 2015-04-20 11:08:18 +0200 (Mon, 20 Apr 2015) $
*/
public interface OverviewContext
{

	/**
		Get the node loader factory that is used to create node loaders.
		@return A node loader factory
	*/
	@SuppressWarnings("rawtypes")
	public NodeLoaderFactory getNodeLoaderFactory();
	
	/**
		Get the node validator factory that is used to create node validators.
		@return A node loader factory
	*/
	@SuppressWarnings("rawtypes")
	public NodeValidatorFactory getNodeValidatorFactory();

	/**
		Get the snapshot manager that is used to load annotation
		values. This method may re-use an existing snapshot manager
		as long as a transaction is active. Once the transaction has
		been committed (or rollbacked) the snapshot manager should
		no longer be used since it may hold inconsistent data. Note
		that it is always safe to create a new snapshot manager.
		@since 2.14
	*/
	public SnapshotManager getSnapshotManager();
	
	/**
		Get the project the overview is using as the "current" project.
		This is mostly needed by validators to check if items are using
		the project default protocols, software, etc.
		@return A project, or null
	*/
	public Project getProject();

	/**
		Get the validation options that are currently in effect.
		@since 2.13
	*/
	public ValidationOptions getValidationOptions();
	
	/**
		Store an auxilliary object in the overview cache. This is a service
		provided by the overview implementation that can be used by node
		loaders and/or validators to store information that may be needed
		again in the future and is	expensive to retreive (eg. from the database). 
		This is for example used to store annotation types that have been marked 
		as "Required by MIAME", parents to annotatable items and more.
		<p>
		Note! The overview impementation is not forced to implement a
		cache, or it may optionally decide to disable the cache. Node 
		loaders and validators should not depend on the cache for 
		correct functionality.
		
		@param cacheKey The key to store the value under
		@param value The value to store in the cache, or null to
			remove the entry
		@see #getCachedObject(Object)
	*/
	public void setCachedObject(Object cacheKey, Object value);

	/**
		Retreive an object stored in the overview cache.
		@param cacheKey The key to use for cache lookup
		@return The object, or null if no object has been cached with the
			given key
		@see #setCachedObject(Object, Object)
	*/
	public Object getCachedObject(Object cacheKey);


	/**
		Register the failure of a validation rule.
		<p>
		Note! The overview implementation is not forced to register the
		failure, or it may optioanlly only register certain types of
		failures.
		
		@param validator The validation rule that fails
		@param node The node that was validated
		@param message An optional message, if null the message from
			{@link Validator#getFailureSummary()} is used
		@param fixes Optional fixes that may be used to correct the problem
	*/
	public Failure createFailure(Validator validator, Node node, String message, Fix... fixes);

	/**
		Same as {@link #initQuery(ItemQuery, Order...)} but easier to use when
		sorting the items in ascending order by a single column.
		@param query The query that should be initialised
		@param sortby The property name to sort by (may be null)
		@return The same query object
	*/
	public <I extends BasicItem> ItemQuery<I> initQuery(ItemQuery<I> query, String sortby);

	/**
		Initialise a query with "global options" and optionally a sort
		order. 
		<p>
		Note! The "global options" can differ between overview implementation,
		but typically include settings some restrictions base on {@link Include}
		options, or permissions on the query.
		
		@param query The query that should be initialised
		@param sortby An array with Order objects (may be null)
		@return The same query object
		@see #initQuery(ItemQuery, Order...)
	*/
	public <I extends BasicItem> ItemQuery<I> initQuery(ItemQuery<I> query, Order... sortby);
	
	/**
		Get the current node cache. Node loaders may use the cache to store
		nodes for already seen items. Then, when the same item is seen again
		the existing node (including all children) is simply cloned. This can
		make the performance a lot better.
		
		@return A node cache, or null if the current overview doesn't want
			to use a cache
	*/
	public NodeCache<Object> getNodeCache();
	
}
