/*
	$Id: InitDB.java 7899 2021-01-20 12:18:30Z nicklas $

	Copyright (C) 2005 Samuel Andersson, Jari Häkkinen, Nicklas Nordborg
	Copyright (C) 2006 Jari Häkkinen, Nicklas Nordborg, Martin Svensson

	This file is part of BASE - BioArray Software Environment.
	Available at http://base.thep.lu.se/

	BASE is free software; you can redistribute it and/or
	modify it under the terms of the GNU General Public License
	as published by the Free Software Foundation; either version 3
	of the License, or (at your option) any later version.

	BASE is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with BASE. If not, see <http://www.gnu.org/licenses/>.
*/
package net.sf.basedb.install;

import java.sql.Driver;
import java.sql.DriverManager;
import java.util.Arrays;
import java.util.List;
import java.util.Properties;

import net.sf.basedb.core.Application;
import net.sf.basedb.core.Config;
import net.sf.basedb.core.HibernateUtil;
import net.sf.basedb.core.Install;
import net.sf.basedb.core.Migration;
import net.sf.basedb.core.ProgressReporter;
import net.sf.basedb.core.Update;
import net.sf.basedb.core.Version;
import net.sf.basedb.core.Application.StartupOptions;
import net.sf.basedb.core.hibernate.SchemaGenerator;
import net.sf.basedb.util.ConsoleProgressReporter;
import net.sf.basedb.util.ChainedProgressReporter;

/**
	This class initialises the BASE database. It is assumed that
	an empty database with no tables or data already exists, and
	that the <code>base.config</code> file has been properly
	filled in with connection information.

	@author Nicklas
	@version 2.0
	@base.modified $Date$
*/
public class InitDB
{
	public static void main(String[] args)
	{
		String cmd = args[0];
		
		try
		{
			if ("install".equals(cmd) || "update".equals(cmd))
			{
				if (args.length < 2 || args.length > 3)
				{
					showUsage(cmd);
					return;
				}
				String rootPassword = args.length == 3 ? args[2] : args[1];
				String rootLogin = args.length == 3 ? args[1] : "root";
				
				SchemaGenerator.Mode mode = "update".equals(cmd) ? 
						SchemaGenerator.Mode.UPDATE : SchemaGenerator.Mode.INSTALL;
				ChainedProgressReporter progress = new ChainedProgressReporter(new ConsoleProgressReporter());
				progress.setRange(0, 30);
				showDbInfo(progress);
				int schemaVersion = Install.createTables(mode, progress, rootLogin, rootPassword);
				progress.setRange(30, 50);
				if (mode == SchemaGenerator.Mode.UPDATE)
				{
					Update.adjustExistingItems(progress, schemaVersion, rootLogin, rootPassword);
				}
				progress.setRange(50, 75);
				Install.initDatabase(mode, progress, rootLogin, rootPassword);
				progress.setRange(75, 90);
				if (mode == SchemaGenerator.Mode.UPDATE)
				{
					Update.updateDatabase(progress, rootLogin, rootPassword);
				}
				progress.setRange(95, 100);
				progress.display(0, "Installing web application...");
				Webclient.install(rootLogin, rootPassword, "", "");
				progress.display(100, "Web application installed successfully.\n");
			}
			else if ("info".equals(cmd))
			{
				showDbInfo(null);
				Application.start(new StartupOptions()
						.disableExtensions(true)
						.disableInternalJobQueue(true)
						);
				HibernateUtil.dbIndexes(true, false, false, false);
				Application.stop();
			}
			else if ("updateindexes".equals(cmd))
			{
				boolean verbose = args.length >= 2 ? args[1].equals("-v") : false;
				showDbInfo(null);
				Application.start(new StartupOptions()
						.disableExtensions(true)
						.disableInternalJobQueue(true)
						);
				HibernateUtil.dbIndexes(verbose, false, false, true);
				Application.stop();
			}
			else if ("dropindexes".equals(cmd))
			{
				boolean verbose = args.length >= 2 ? args[1].equals("-v") : false;
				showDbInfo(null);
				Application.start(new StartupOptions()
						.disableExtensions(true)
						.disableInternalJobQueue(true)
						);
				HibernateUtil.dbIndexes(verbose, false, true, false);
				Application.stop();
			}
			else if ("dynamicdb".equals(cmd))
			{
				List<String> opt = Arrays.asList(args);
				showDbInfo(null);
				boolean verbose = opt.contains("-v");
				boolean drop = opt.contains("-dropindexes");
				Application.start(new StartupOptions()
						.disableExtensions(true)
						.disableInternalJobQueue(true)
						);
				HibernateUtil.dynamicDbIndexes(verbose, false, drop);
				Application.stop();
			}
			else if ("migrate".equals(cmd))
			{
				// args[1] = 'import' or 'export'
				// args[2] = path to output dir
				if (args.length != 3)
				{
					showUsage(cmd);
					return;
				}
				boolean export = "export".equals(args[1]);
				
				showDbInfo(null);
				String path = args[2];
				if (export)
				{
					ProgressReporter progress = new ConsoleProgressReporter(false);
					progress.display(0, "Starting export...\n");
					ChainedProgressReporter chained = new ChainedProgressReporter(progress);
					chained.setRange(1, 99);
					Migration.exportAll(path, chained);
					progress.display(100, "Export completed successfully!\n");
				}
				else
				{
					ProgressReporter progress = new ConsoleProgressReporter();
					ChainedProgressReporter chained = new ChainedProgressReporter(progress);
					progress.display(0, "Starting import...");
					// Create empty database
					chained.setRange(1, 5);
					Install.createTables(SchemaGenerator.Mode.INSTALL, chained, null, null);

					// Import all data
					chained.setRange(6, 99);
					Migration.importAll(path, chained);

					progress.display(100, "Import completed successfully!\n");
				}
				Application.stop();
			}
			else 
			{
				System.out.println("Unknown command: " + cmd);
				System.out.println("Usage: InitDB cmd <options>");
				System.out.println("    cmd: install | update | info | updateindexes | dropindexes | dynamicdb");
				System.out.println("options: ");
				System.out.println("           -v : Verbose output (if cmd = updateindexes, dropindexs or dynamicdb)");
				System.out.println(" -dropindexes : Drop indexes when cmd = dynamicdb");
			}
		}
		catch (Throwable t)
		{
			System.out.println("\n");
			t.printStackTrace();
		}
	}

	private static void showUsage(String cmd)
	{
		if ("install".equals(cmd) || "update".equals(cmd))
		{
			String script = "install".equals(cmd) ? "initdb.sh" : "updatedb.sh";
			System.out.println("Usage: "+script+" [root login] <root password>");
			System.out.println("       root login   : The login to use for the root user account");
			System.out.println("       root password: The password to give to the root user account");
		}
		else if ("migrate".equals(cmd))
		{
			System.out.println("Usage: migrate.sh <cmd> <path>");
			System.out.println("       cmd    : export | import");
			System.out.println("       path   : The path to a directory where data files should be saved/read");
		}
	}

	static void showDbInfo(ProgressReporter progress)
	{
		final Properties properties = System.getProperties();
		final String dbUrl = Config.getString("db.url");
		String driverClass = "unknown class";
		String driverVersion = "unknown version";
		try
		{
			final Driver driver = DriverManager.getDriver(dbUrl);
			driverClass = driver.getClass().getName();
			driverVersion = "version " + driver.getMajorVersion() + "." + driver.getMinorVersion();
		}
		catch (Throwable t)
		{}
		
		String message = 
			"--System information-----------------------------\n" + 
			"BASE      : " + Version.getVersion() + "\n" + 
			"Database  : " + dbUrl + "\n" + 
			"Dialect   : " + Config.getString("db.dialect") + "\n" + 
			"JDBC      : " + driverClass + "; " + driverVersion + "\n" +
			"Hibernate : " + org.hibernate.Version.getVersionString() + "\n"+
			"Java      : " + properties.getProperty("java.runtime.name") + "; " +
				properties.getProperty("java.runtime.version") + "; " +
				properties.getProperty("java.vendor") + "\n" + 
			"OS        : " + properties.getProperty("os.name") +"; " +
				properties.getProperty("os.arch") + "; " +
				properties.getProperty("os.version") +"\n" +
			"-------------------------------------------------";
		if (progress != null)
		{
			progress.display(0, message);
		}
		else 
		{
			System.out.println(message);
		}
	}
	
}

