/**
	$Id: PropertyIdMethod.java 5590 2011-03-16 11:48:47Z nicklas $

	Copyright (C) 2008 Nicklas Nordborg

	This file is part of BASE - BioArray Software Environment.
	Available at http://base.thep.lu.se/

	BASE is free software; you can redistribute it and/or
	modify it under the terms of the GNU General Public License
	as published by the Free Software Foundation; either version 3
	of the License, or (at your option) any later version.

	BASE is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with BASE. If not, see <http://www.gnu.org/licenses/>.
*/
package net.sf.basedb.plugins.batchimport;

import java.util.ArrayList;
import java.util.Arrays;
import java.util.List;

import net.sf.basedb.core.Annotatable;
import net.sf.basedb.core.Annotation;
import net.sf.basedb.core.AnnotationType;
import net.sf.basedb.core.BasicItem;
import net.sf.basedb.core.DbControl;
import net.sf.basedb.core.Include;
import net.sf.basedb.core.Item;
import net.sf.basedb.core.ItemQuery;
import net.sf.basedb.core.Type;
import net.sf.basedb.core.query.Annotations;
import net.sf.basedb.core.query.Expressions;
import net.sf.basedb.core.query.Hql;
import net.sf.basedb.core.query.Orders;
import net.sf.basedb.core.query.Restrictions;

/**
	Identification method implementation that can use an annotation
	for identifying items.

	@author nicklas
	@since 3.15
*/
public class AnnotationIdMethod
	implements IdMethod
{

	/**
		Get identification methods that are based on annotations. It will
		make a query for all annotation types that are suitable as identifiers:
		
		 * The value type must be one of STRING, INT or LONG
		 * It must not be an enumeration
		 * The 'identifier' flag must be set
		@param itemType Either a specified item or null to return all possible annotation id methods
	*/
	public static List<IdMethod> getIdMethods(DbControl dc, Item itemType)
	{
		ItemQuery<AnnotationType> query = AnnotationType.getQuery(itemType);
		query.include(Include.MINE, Include.IN_PROJECT, Include.SHARED, Include.OTHERS);
		// The annotation type must be STRING, INTEGER or LONG
		query.restrict(Restrictions.in(Hql.property("valueType"), Expressions.parameter("valueTypes", Type.INT)));
		query.setParameter("valueTypes", Arrays.asList(Type.STRING.getValue(), Type.INT.getValue(), Type.LONG.getValue()), Type.INT);
		// Enumerations can't be used
		query.restrict(Restrictions.eq(Hql.property("enumeration"), Expressions.bool(false)));
		// The 'identifier' flag must be set
		query.restrict(Restrictions.eq(Hql.property("identifier"), Expressions.bool(true)));
		query.order(Orders.asc(Hql.property("name")));
		// Get a list of possible annotation types
		List<AnnotationType> annotationTypes = new ArrayList<AnnotationType>(query.list(dc));
	
		List<IdMethod> idMethods = new ArrayList<>();
		for (AnnotationType at : annotationTypes)
		{
			idMethods.add(new AnnotationIdMethod(at));
		}
		return idMethods;
	}

	
	private final AnnotationType annotationType;
	/**
		Create a new id method.
	*/
	public AnnotationIdMethod(AnnotationType annotationType)
	{
		this.annotationType = annotationType;
	}
	
	/*
		From the IdMethod interface
		---------------------------
	*/
	@Override
	public String getMethod()
	{
		return "annotation-type-"+annotationType.getId();
	}
	
	@Override
	public String getTitle()
	{
		return "[A] " + annotationType.getName();
	}
	
	@Override
	public String getColumnMappingParameterName()
	{
		return AbstractItemImporter.idAnnotationColumnMapping.getName();
	}

	@Override
	public boolean isTrulyUnique()
	{
		return false;
	}

	/**
		Adds a restriction to the query:
		<code>property = :identifier</code>
	*/
	@Override
	public <I extends BasicItem> ItemQuery<I> prepareQuery(DbControl dc, ItemQuery<I> query)
	{
		if (annotationType == null)
		{
			throw new NullPointerException("No annotation type has been specified for 'Annotation' identification method.");
		}
		query.joinPermanent(Annotations.innerJoin(annotationType, "idat"));
		query.restrictPermanent(
				Restrictions.eq(
					Hql.alias("idat"),
					Expressions.parameter("identifier")
				)
			);
		return query;
	}

	/**
		Find all items with <code>property = identifier</code>.
	*/
	@Override
	public <I extends BasicItem> List<I> find(DbControl dc, ItemQuery<I> query, String identifier)
	{
		if (annotationType == null)
		{
			throw new NullPointerException("No annotation type has been specified for 'Annotation' identification method.");
		}
		Type idType = annotationType.getValueType();
		query.setParameter("identifier", idType.parseString(identifier), idType);
		return query.list(dc);
	}
	
	/**
		Set the annotation on a newly created item.
	*/
	@Override
	public <I extends BasicItem> boolean setIdentifierOnNewItem(DbControl dc, I item, String identifier)
	{
		if (annotationType == null)
		{
			throw new NullPointerException("No annotation type has been specified for 'Annotation' identification method.");
		}
		Annotatable aItem = (Annotatable)item;
		Type idType = annotationType.getValueType();
		Annotation annotation = aItem.getAnnotationSet().getAnnotation(annotationType);
		boolean changed = annotation.setValueIfDifferent(idType.parseString(identifier), null);
		return changed;
	}
	
	// ---------------------------
	
	/*
		From the Object class
		---------------------
	*/
	@Override
	public String toString()
	{
		return annotationType.getName();
	}
	// --------------------------

}
