/**
	$Id: FallbackIdMethod.java 7610 2019-02-27 13:18:43Z nicklas $

	Copyright (C) 2008 Nicklas Nordborg

	This file is part of BASE - BioArray Software Environment.
	Available at http://base.thep.lu.se/

	BASE is free software; you can redistribute it and/or
	modify it under the terms of the GNU General Public License
	as published by the Free Software Foundation; either version 3
	of the License, or (at your option) any later version.

	BASE is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with BASE. If not, see <http://www.gnu.org/licenses/>.
*/
package net.sf.basedb.plugins.batchimport;

import java.util.Collections;
import java.util.List;

import net.sf.basedb.core.BasicItem;
import net.sf.basedb.core.DbControl;
import net.sf.basedb.core.ItemNotFoundException;
import net.sf.basedb.core.ItemQuery;
import net.sf.basedb.util.Values;

/**
	Identification method implementation that first tries a parent id
	method, and, if no item is found and the identifier is a number,
	tries to load the item by id.

	@author nicklas
	@version 2.8
	@base.modified $Date: 2019-02-27 14:18:43 +0100 (Wed, 27 Feb 2019) $
*/
public class FallbackIdMethod
	implements IdMethod
{
	
	/**
		Use the name for finding items. If no item is found and the identifier
		is numerical, try loading it by ID.
	*/
	public static final FallbackIdMethod NAME_OR_ID = 
		new FallbackIdMethod("nameOrId", "Name or Internal ID", PropertyIdMethod.NAME);
	
	public static final FallbackIdMethod NAME_OR_EXTERNALID_OR_ID =
		new FallbackIdMethod("nameOrExternalIdOrId", "Name, external ID or internal ID", 
				MultiPropertyIdMethod.NAME_OR_EXTERNALID);
		
	public static final FallbackIdMethod NAME_OR_BARCODE_OR_ID =
		new FallbackIdMethod("nameOrBarcodeOrId", "Name, barcode or internal ID", 
				MultiPropertyIdMethod.NAME_OR_BARCODE);
	
	/**
		Use the name or system ID for finding items. If no item is found and 
		the identifier is numerical, try loading it by ID.
		@since 3.0
	*/
	public static final FallbackIdMethod NAME_OR_SYSTEMID_OR_ID =
		new FallbackIdMethod("nameOrSystemIdOrId", "Name, system ID or internal ID", 
				MultiPropertyIdMethod.NAME_OR_SYSTEMID);
	
	/**
		ID method that can be used for units. Use the name or unit symbol for finding items.
		If no item is found and the identifier is numerical, try loading it by ID.
		@since 3.8
	*/
	public static final FallbackIdMethod NAME_OR_UNITSYMBOL_OR_ID =
		new FallbackIdMethod("nameOrUnitSymbolOrId", "Name, unit symbol or internal ID", 
				MultiPropertyIdMethod.NAME_OR_UNITSYMBOL);

	
	private final String method;
	private final String title;
	private final IdMethod parent;
	
	/**
		Create a new id method.

		@param method The method name, which should be unique among all methods
		@param title The title to use for display purposes
		@param parent The parent id method, which will be tried first
	*/
	public FallbackIdMethod(String method, String title, IdMethod parent)
	{
		this.method = method;
		this.title = title;
		this.parent = parent;
	}
	
	/*
		From the IdMethod interface
		---------------------------
	*/
	@Override
	public String getMethod()
	{
		return method;
	}
	
	@Override
	public String getTitle()
	{
		return title;
	}

	@Override
	public boolean isTrulyUnique()
	{
		return parent.isTrulyUnique();
	}
	
	@Override
	public String getColumnMappingParameterName()
	{
		return parent.getColumnMappingParameterName();
	}
	
	/**
		Adds a restriction to the query:
		<code>property = :identifier</code>
	*/
	@Override
	public <I extends BasicItem> ItemQuery<I> prepareQuery(DbControl dc, ItemQuery<I> query)
	{
		return parent.prepareQuery(dc, query);
	}

	/**
		First, find items using the parent id method. If no items is found with this, 
		try to convert the identifier to a number and use <code>getById()</code> to
		load the item.
	*/
	@Override
	public <I extends BasicItem> List<I> find(DbControl dc, ItemQuery<I> query, String identifier)
	{
		List<I> items = parent.find(dc, query, identifier);
		if (items.size() == 0)
		{
			int id = Values.getInt(identifier);
			if (id != 0)
			{
				try
				{
					I item = query.getItemType().getById(dc, id);
					items = Collections.singletonList(item);
				}
				catch (ItemNotFoundException ex)
				{}
			}
		}
		return items;
	}
	// ---------------------------
	
	/*
		From the Object class
		---------------------
	*/
	@Override
	public String toString()
	{
		return title;
	}
	// --------------------------

}
