/*
	$Id: TestPlate.java 7665 2019-03-20 13:31:03Z nicklas $

	Copyright (C) 2005 Johan Enell, Nicklas Nordborg
	Copyright (C) 2006 Johan Enell, Jari Häkkinen, Nicklas Nordborg

	This file is part of BASE - BioArray Software Environment.
	Available at http://base.thep.lu.se/

	BASE is free software; you can redistribute it and/or
	modify it under the terms of the GNU General Public License
	as published by the Free Software Foundation; either version 3
	of the License, or (at your option) any later version.

	BASE is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with BASE. If not, see <http://www.gnu.org/licenses/>.
*/
import net.sf.basedb.core.*;
import net.sf.basedb.core.data.ReporterData;
import net.sf.basedb.util.Coordinate;
import net.sf.basedb.util.FileUtil;
import net.sf.basedb.util.parser.FlatFileParser;

import java.util.Date;
import java.util.HashMap;
import java.util.Iterator;
import java.util.Set;
import java.util.regex.Pattern;

public class TestPlate
{

	static boolean ok = true;
	public static void main(String[] args)
	{
		TestUtil.checkArgs(args);
		TestUtil.begin();
		ok = test_all();
		TestUtil.stop();
	}

	static boolean test_all()
	{
		write("++Testing plates");
		write_header();
		// Standard tests: create, load, list
		int plateGeometryId = TestPlateGeometry.test_create(8, 12, false);
		int plateTypeId = TestPlateType.test_create(plateGeometryId, false);
		int id = test_create(plateTypeId, true);
		int id2 = test_create(plateTypeId, false);
		test_load(id);
		test_list(-1);

		// Test events
		int protocolTypeId = TestItemSubtype.test_create(Item.PROTOCOL, "Plate protocol type");
		int protocolId = TestProtocol.test_create(protocolTypeId, null, false);
		int eventTypeId = TestPlateType.test_create_event_type(plateTypeId, protocolTypeId);
		int eventId = test_create_event(id, eventTypeId, protocolId);
		test_list_events(id, 1);
		
		if (TestUtil.waitBeforeDelete()) TestUtil.waitForEnter();
		test_delete_event(eventId);
		TestPlateType.test_delete_event_type(eventTypeId);
		TestProtocol.test_delete(protocolId);
		TestItemSubtype.test_delete(protocolTypeId);

		// Standard test: Delete
		test_delete(id);
		test_delete(id2);
		TestPlateType.test_delete(plateTypeId);
		TestPlateGeometry.test_delete(plateGeometryId);
		write("++Testing plates "+(ok ? "OK" : "Failed")+"\n");
		return ok;
	}

	static int test_create(int plateTypeId, boolean setAll)
	{
		if (plateTypeId == 0 || !TestUtil.hasPermission(Permission.CREATE, Item.PLATE)) return 0;
		int id = 0;
		DbControl dc = null;
		try
		{
			dc = TestUtil.getDbControl();
			PlateType pt = PlateType.getById(dc, plateTypeId);
			Plate p = Plate.getNew(dc, pt);
			if (setAll)
			{
				p.setName("Test plate");
				p.setDescription("Added at "+new Date());
				p.setBarcode("00557788");
				p.setDestroyed(false);

				PlateGeometry geometry = pt.getPlateGeometry();
				int columns = geometry.getColumns();
				int rows = geometry.getRows();
				DataQuery<ReporterData> query = Reporter.getQuery();
				DataResultIterator<ReporterData> reporters = query.iterate(dc);
				for (int row = 0; row < rows; ++row)
				{
					for (int column = 0; column < columns; ++column)
					{
						if (reporters.hasNext())
						{
							p.getWell(row, column).setReporter(reporters.next());
						}
					}
				}
				
				// Make first and last well empty
				p.getWell(0, 0).setReporter(null);
				p.getWell(rows-1, columns-1).setReporter(null);
			}
			
			
			dc.saveItem(p);
			dc.commit();
			id = p.getId();
			dc = TestUtil.getDbControl();
			dc.reattachItem(p, false);
			write_item(0, p);
			write("--Create plate OK");
		}
		catch (Throwable ex)
		{
			write("--Create plate FAILED");
			ex.printStackTrace();
			ok = false;
		}
		finally
		{
			if (dc != null) dc.close();
		}
		return id;
	}

	static void test_load(int id)
	{
		if (id == 0) return;
		DbControl dc = null;
		try
		{
			dc = TestUtil.getDbControl();
			Plate p = Plate.getById(dc, id);
			write_item(0, p);
			write("--Load plate OK");
		}
		catch (Throwable ex)
		{
			write("--Load plate FAILED");
			ex.printStackTrace();
			ok = false;
		}
		finally
		{
			if (dc != null) dc.close();
		}
	}

	static int[] test_import_from_file(String filename, String headerRegexp, int plateTypeId, int plateIdCol, int rowCol, int columnCol, int reporterCol)
	{
		if (!TestUtil.hasPermission(Permission.CREATE, Item.PLATE)) return null;
		write("--Start import plate from file (" + filename + ")");
		DbControl dc = null;
		long time;
		int numWells = 0;
		HashMap<String, Plate> plates = new HashMap<String, Plate>();
		try
		{
			dc = TestUtil.getDbControl();

			FlatFileParser ffp = new FlatFileParser();
			ffp.setDataHeaderRegexp(Pattern.compile(headerRegexp));
			ffp.setDataSplitterRegexp(Pattern.compile("\\t"));
			ffp.setInputStream(FileUtil.getInputStream(new java.io.File(filename)), "ISO-8859-1");
			ffp.parseHeaders();
			PlateType pt = PlateType.getById(dc, plateTypeId);
			time = System.currentTimeMillis();
			while (ffp.hasMoreData())
			{
				FlatFileParser.Data data = ffp.nextData();
				String plateId = data.getString(plateIdCol);
				int row = Coordinate.alphaToNumeric(data.getString(rowCol))-1;
				int column = Coordinate.alphaToNumeric(data.getString(columnCol))-1;
				String reporterId = data.getString(reporterCol);

				Plate p = plates.get(plateId);
				if (p == null)
				{
					p = Plate.getNew(dc, pt);
					p.setName(filename+"_"+plateId);
					dc.saveItem(p);
					plates.put(plateId, p);
				}
				Well w = p.getWell(row, column);
				w.setReporter(Reporter.getByExternalId(dc, reporterId));
				numWells++;
			}
			dc.commit();
			time = System.currentTimeMillis()-time;
			write("--Import plate from file OK ("+numWells+" wells, time="+time+" ms)");
		}
		catch (Throwable ex)
		{
			write("--Import reporters from file FAILED ("+numWells+" wells)");
			ex.printStackTrace();
			ok = false;
		}
		finally
		{
			if (dc != null) dc.close();
		}
		int[] ids = new int[plates.size()];
		if (plates.size() > 0)
		{
			Iterator<Plate> it = plates.values().iterator();
			for (int i = 0; it.hasNext(); i++)
			{
				Plate p = it.next();
				ids[i] = p.getId();
			}
		}
		return ids;
	}

	static void test_list(int expectedResults)
	{
		DbControl dc = null;
		try
		{
			dc = TestUtil.getDbControl();
			ItemResultList<Plate> l = Plate.getQuery().list(dc);
			for (int i = 0; i<l.size(); i++)
			{
				write_item(i, l.get(i));
			}
			if (expectedResults >= 0 && expectedResults != l.size())
			{
				throw new BaseException("Expected "+expectedResults+" results, not "+l.size());
			}
			write("--List plates OK ("+l.size()+")");
		}
		catch (Throwable ex)
		{
			write("--List plates FAILED");
			ex.printStackTrace();
			ok = false;
		}
		finally
		{
			if (dc != null) dc.close();
		}
	}


	static void test_delete(int id)
	{
		if (id == 0) return;
		DbControl dc = null;
		try
		{
			dc = TestUtil.getDbControl();
			Plate p = Plate.getById(dc, id);
			dc.deleteItem(p);
			Set<ItemProxy> using = p.getUsingItems();
			if (using.size() > 0) 
			{
			   throw new BaseException(p + " is used by " + using);
			}
			dc.commit();
			write("--Delete plate OK");
		}
		catch (Throwable ex)
		{
			write("--Delete plate FAILED");
			ex.printStackTrace();
			ok = false;
		}
		finally
		{
			if (dc != null) dc.close();
		}
	}

	public static void test_delete_all()
	{
		DbControl dc = null;
		try
		{
			dc = TestUtil.getDbControl();
			for(Plate p : Plate.getQuery().list(dc))
			{
				dc.deleteItem(p);
			}
			dc.commit();
			write("--Delete all plates OK");
		}
		catch (Throwable ex)
		{
			write("--Delete all plates FAILED");
			ex.printStackTrace();
			ok = false;
		}
	}
	
	static void write_header()
	{
		if (!TestUtil.getSilent())
		{
			System.out.println("   \tID \tName      \tDescription\tPlate type\tBarcode\tDestroyed");
			System.out.println("-- \t-- \t--------- \t-----------\t----------\t-------\t---------");
		}
	}
	static void write_item(int i, Plate p)
		throws BaseException
	{
		if (!TestUtil.getSilent()) System.out.println(i+":\t"+p.getId()+"\t"+p.getName()+"\t"+p.getDescription()+
			"\t"+p.getPlateType()+"\t"+p.getBarcode()+"\t"+p.isDestroyed());
	}
	static void write_item(int i, PlateEvent pe)
		throws BaseException
	{
		if (!TestUtil.getSilent()) System.out.println(i+":\t"+pe.getId()+"\t"+
			"\t"+pe.getPlate()+"\t"+pe.getProtocol()+"\t"+pe.getComment()+"\t"+pe.getEntryDate()+"\t"+pe.getEventDate());
	}
	static void write(String message)
	{
		System.out.println(message);
	}

	static int test_create_event(int plateId, int eventTypeId, int protocolId)
	{
		if (plateId == 0 || eventTypeId == 0) return 0;
		int id = 0;
		DbControl dc = null;
		try
		{
			dc = TestUtil.getDbControl();
			Plate p = Plate.getById(dc, plateId);
			PlateEventType pet = PlateEventType.getById(dc, eventTypeId);
			PlateEvent pe = p.getEvent(pet);
			pe.setComment("Test plate event");
			pe.setEventDate(new Date());
			if (protocolId != 0) pe.setProtocol(Protocol.getById(dc, protocolId));
			dc.saveItem(pe);
			dc.commit();
			id = pe.getId();
			dc = TestUtil.getDbControl();
			dc.reattachItem(pe, false);
			write_item(0, pe);
			write("--Create event OK");
		}
		catch (Exception ex)
		{
			write("--Create event FAILED");
			ex.printStackTrace();
			ok = false;
		}
		finally
		{
			if (dc != null) dc.close();
		}
		return id;
	}
	
	static void test_list_events(int plateId, int expectedResults)
	{
		if (plateId == 0) return;
		DbControl dc = null;
		try
		{
			dc = TestUtil.getDbControl();
			Plate p = Plate.getById(dc, plateId);
			
			ItemResultList<PlateEvent> l = p.getEvents().list(dc);
			for (int i = 0; i<l.size(); i++)
			{
				write_item(i, l.get(i));
			}
			if (expectedResults >= 0 && expectedResults != l.size())
			{
				throw new BaseException("Expected "+expectedResults+" results, not "+l.size());
			}
			write("--List events OK ("+l.size()+")");
		}
		catch (Exception ex)
		{
			write("--List events FAILED");
			ex.printStackTrace();
			ok = false;
		}
		finally
		{
			if (dc != null) dc.close();
		}
	}

	static void test_delete_event(int eventId)
	{
		if (eventId == 0) return;
		DbControl dc = null;
		try
		{
			dc = TestUtil.getDbControl();
			PlateEvent pe = PlateEvent.getById(dc, eventId);
			dc.deleteItem(pe);
			dc.commit();
			write("--Delete event OK");
		}
		catch (Exception ex)
		{
			write("--Delete event FAILED");
			ex.printStackTrace();
			ok = false;
		}
		finally
		{
			if (dc != null) dc.close();
		}
	}

}
