/*
	$Id: TestUtil.java 7899 2021-01-20 12:18:30Z nicklas $

	Copyright (C) 2005, 2006 Jari Häkkinen, Nicklas Nordborg

	This file is part of BASE - BioArray Software Environment.
	Available at http://base.thep.lu.se/

	BASE is free software; you can redistribute it and/or
	modify it under the terms of the GNU General Public License
	as published by the Free Software Foundation; either version 3
	of the License, or (at your option) any later version.

	BASE is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with BASE. If not, see <http://www.gnu.org/licenses/>.
*/
import net.sf.basedb.core.Application;
import net.sf.basedb.core.Config;
import net.sf.basedb.core.SessionControl;
import net.sf.basedb.core.DbControl;
import net.sf.basedb.core.BaseException;
import net.sf.basedb.core.Item;
import net.sf.basedb.core.Permission;
import net.sf.basedb.core.Version;
import net.sf.basedb.core.Application.StartupOptions;
import net.sf.basedb.core.authentication.LoginRequest;

import java.io.File;
import java.io.FileFilter;
import java.net.InetAddress;
import java.sql.Connection;
import java.sql.DatabaseMetaData;
import java.sql.Driver;
import java.sql.DriverManager;
import java.util.Properties;

public class TestUtil
{

	static 
	{
		// Require that asserts are enabled for test programs
		boolean assertsEnabled = false;
		assert (assertsEnabled = true) == true;
		if (!assertsEnabled) throw new AssertionError("Asserts must be enabled.");
	}

	private static boolean silent = false;
	private static boolean waitBeforeDelete = false;
	private static String login;
	private static String password;
	private static String client;

	public static void checkArgs(String[] args)
	{
		login = "root";
		password = "root";
		client = "net.sf.basedb.clients.test";
		
		for (int i = 0; i < args.length; ++i)
		{
			String arg = args[i];
			if (arg.startsWith("-")) 
			{
				parseOption(arg);
			}
			else
			{
				if (args.length != i+2)
				{
					showUsage();
					System.exit(0);
				}
				login = arg;
				password = args[i+1];
				i = args.length;
			}
		}
	}

	private static void parseOption(String option)
	{
		if ("-s".equals(option))
		{
			silent = true;
		}
		else if ("-w".equals(option))
		{
			waitBeforeDelete = true;
		}
		else
		{
			showUsage();
			System.exit(0);
		}
	}

	private static void showUsage()
	{
		System.out.println("Usage: run.sh class [-s] [-w] [user password]");
		System.out.println("       class: One of the Test* class files");
		System.out.println("       -s: silent mode");
		System.out.println("       -w: wait for keypress before items are deleted");
		System.out.println("       user: The login for a user account");
		System.out.println("       password: The password for the user account");
	}

	public static boolean getSilent()
	{
		return silent;
	}
	
	public static boolean waitBeforeDelete()
	{
		return waitBeforeDelete;
	}

	private static SessionControl sc = null;

	public static DbControl getDbControl()
		throws BaseException
	{
		DbControl dc = sc.newDbControl();
		return dc;
	}
	
	public static boolean hasPermission(Permission permission, Item itemType)
	{
		return sc.hasPermission(permission, itemType);
	}

	public static String getLogin()
	{
		return login;
	}
	public static String getPassword()
	{
		return password;
	}
	
	public static String getClient()
	{
		return client;
	}

	public static void begin()
	{
		try
		{
			Config.setProperty("plugins.dir", System.getProperty("user.dir"));
			showDbInfo();
			Application.start(new StartupOptions()
					.disableExtensions(true)
					.disableInternalJobQueue(true)
					);
			Application.getStaticCache().cleanUp(new FileFilter()
			{
				@Override
				public boolean accept(File pathname)
				{
					return true;
				}
			});
			sc = Application.newSessionControl(client, getLocalIp(), null);
			login();
		}
		catch (Throwable ex)
		{
			System.out.println("FAILURE INITIALIZING TEST");
			ex.printStackTrace();
			System.exit(0);
		}
	}
	
	public static void logout()
		throws BaseException
	{
		sc.logout();
	}
	
	public static void login()
		throws BaseException
	{
		login(login, password);
	}
	
	public static void login(String login, String password)
		throws BaseException
	{
		LoginRequest loginRequest = new LoginRequest(login, password);
		loginRequest.setComment("Running test program");
		sc.login(loginRequest);
	}

	public static void stop()
	{
		try
		{
			sc.logout();
			Application.stop();
		}
		catch (Throwable ex)
		{
			ex.printStackTrace();
		}
	}

	public static SessionControl getSessionControl()
	{
		return sc;
	}
	
	public static String getLocalIp()
	{
		try
		{
			return InetAddress.getLocalHost().toString();
		}
		catch (Throwable ex)
		{
			return "unknown";
		}
	}
	
	public static void waitForEnter()
	{
		try
		{
			System.out.print("Press ENTER to continue...");
			System.in.read();
			System.in.skip(System.in.available());
		}
		catch (Exception ex)
		{}
	}
	
	public static void listThreads()
	{
		Thread[] threads = new Thread[Thread.activeCount()];
      int numThreads = Thread.enumerate(threads);
      for (int i = 0; i < numThreads; i++) 
      {
      	System.out.println(i + ": " + threads[i].getName()); 
      }
	}
	
	public static boolean isWindows()
	{
		return System.getProperty("os.name").toLowerCase().contains("windows");
	}
	public static boolean isLinux()
	{
		return System.getProperty("os.name").toLowerCase().contains("linux");
	}
	public static boolean isMac()
	{
		return System.getProperty("os.name").toLowerCase().contains("mac");
	}
	
	private static void showDbInfo()
	{
		final Properties properties = System.getProperties();
		final String dbUrl = Config.getString("db.url");
		String driverClass = "unknown class";
		String driverVersion = "unknown version";
		String database = "Unknown database";
		
		try
		{
			final Driver driver = DriverManager.getDriver(dbUrl);
			driverClass = driver.getClass().getName();
			driverVersion = "version " + driver.getMajorVersion() + "." + driver.getMinorVersion();
			Properties p = new Properties();
			p.setProperty("user", Config.getString("db.username"));
			p.setProperty("password", Config.getString("db.password"));
			Connection c = driver.connect(dbUrl, p);
			DatabaseMetaData data = c.getMetaData();
			database = data.getDatabaseProductName() + " " + data.getDatabaseProductVersion();
			c.close();
		}
		catch (Throwable t)
		{}
		
		String message = 
			"--System information-----------------------------\n" + 
			"BASE      : " + Version.getVersion() + "\n" + 
			"Database  : " + database + "\n" + 
			"Dialect   : " + Config.getString("db.dialect") + "\n" + 
			"Hibernate : " + org.hibernate.Version.getVersionString() + "\n"+
			"JDBC      : " + driverClass + "; " + driverVersion + "\n" +
			"URL       : " + dbUrl + "\n" +
			"Java      : " + properties.getProperty("java.runtime.name") + "; " +
				properties.getProperty("java.runtime.version") + "; " +
				properties.getProperty("java.vendor") + "\n" + 
			"OS        : " + properties.getProperty("os.name") +"; " +
				properties.getProperty("os.arch") + "; " +
				properties.getProperty("os.version") +"\n" +
			"-------------------------------------------------";
		System.out.println(message);
	}
	
}

