/* $Id: table.js 7894 2020-12-08 08:55:29Z nicklas $
	------------------------------------------------------------------
	Copyright (C) 2005 Nicklas Nordborg
	Copyright (C) 2006 Jari Häkkinen, Nicklas Nordborg, Martin Svensson, Gregory Vincic
	Copyright (C) 2007 Johan Enell, Nicklas Nordborg

	This file is part of BASE - BioArray Software Environment.
	Available at http://base.thep.lu.se/

	BASE is free software; you can redistribute it and/or
	modify it under the terms of the GNU General Public License
	as published by the Free Software Foundation; either version 3
	of the License, or (at your option) any later version.

	BASE is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with BASE. If not, see <http://www.gnu.org/licenses/>.
	------------------------------------------------------------------

	JavaScript functions for the Table taglib.

	@author Nicklas
	@version 2.0
*/
'use strict';

var Table = function()
{
	// Hold parsed column definitions -- initialized lazily when Table.getColumnDefs is called
	var columnDefs = [];

	var table = {};
	var internal = {};
	
	/**
		Initialize all control elements in the table, such as
		sorting, filtering, pageing, etc.
	*/
	internal.initTable = function(element, autoInit)
	{
		if (autoInit != 'table') return;
		var tableId = element.id;
		var tableFrm = document.forms[tableId];
		var itemType = Data.get(element, 'item-type');
		var attributes = {'table-id': tableId, 'item-type': itemType};
		
		// Set flag to enable 'please wait...' overlay if re-loading takes time
		Data.set(tableFrm, 'enable-please-wait', 1000); // Display a message after 1s
		
		// Add 'click' handler to page navigator
		// A single handler is added to the main div and relies 
		// on event bubbling from the actual targets
		var navigator = element.getElementsByClassName('table-navigator');
		for (var i = 0; i < navigator.length; i++)
		{
			Events.addEventHandler(navigator[i], 'click', Table.navigatorOnClick);
		}
		
		// Limit the 'rowsperpage' field to numbers and submit the form if ENTER is pressed
		var rowsPerPage = tableFrm.rowsperpage;
		if (rowsPerPage)
		{
			Events.addEventHandler(rowsPerPage, 'keypress', Events.integerOnly);
			Events.doOnEnter(rowsPerPage, Forms.submit);
		}
		
		// Add event handler for the 'view/preset' selection list
		// and icons for clearing the filter and show hidden columns with a filter
		if (tableFrm.presetselector)
		{
			Events.addEventHandler(tableFrm.presetselector, 'change', table.presetOnChange);
			Events.addEventHandler(tableId+'.clearfilter', 'click', table.clearFilterOnClick, attributes);
			Events.addEventHandler(tableId+'.showcolumns', 'click', table.showColumnsOnClick, attributes);
		}
		
		// Add 'click' handlers to column headers for sorting
		var sort = element.getElementsByClassName('table-sort');
		for (var i = 0; i < sort.length; i++)
		{
			Events.addEventHandler(sort[i], 'click', table.sortOnClick, attributes);
		}
		
		// Add 'drag-drop' handlers for moving columns
		var cols = element.getElementsByClassName('table-col-draggable');
		for (var i = 0; i < cols.length; i++)
		{
			var dropTarget = cols[i];
			Data.set(dropTarget, 'column-index', i);
			Events.addEventHandler(dropTarget, 'dragstart', table.beginColumnDrag, attributes);
			Events.addEventHandler(dropTarget, 'dragover', table.checkColumnDropTarget);
			Events.addEventHandler(dropTarget, 'dragleave', table.leaveColumnDropTarget);
			Events.addEventHandler(dropTarget, 'drop', table.columnDropped);
			Events.addEventHandler(dropTarget, 'dragend', table.endColumnDrag);
			
			// Right-clicking should bring up a menu with selection of columns
			Events.addEventHandler(dropTarget, 'mouseup', table.showColumnSelectionMenu);
			Events.addEventHandler(dropTarget, 'contextmenu', table.showColumnSelectionMenu);
		}
		
		// Add 'hide column' handler to column headers
		var cols = element.getElementsByClassName('table-col-hide');
		for (var i = 0; i < cols.length; i++)
		{
			Events.addEventHandler(cols[i], 'click', table.hideColumnOnClick, attributes);
		}
		
		// Submit form when clicked or ENTER is pressed 
		var submit = element.getElementsByClassName('table-submit');
		for (var i = 0; i < submit.length; i++)
		{
			var field = submit[i];
			var type = field.type;
			if (type == 'radio' || type == 'checkbox')
			{
				// Radio and checkboxes react to 'click'
				Events.addEventHandler(field, 'click', Forms.submit);
			}
			else if (type == 'text')
			{
				// Text fields react to ENTER key
				Events.doOnEnter(field, Forms.submit);
			}
			else if (type == 'select-one')
			{
				// Selection list react to 'onchange'
				Events.addEventHandler(field, 'change', Forms.submit);
			}
		}
		
		// Add 'click' handler to the 'Toggle all' icon
		var tableCheck = element.getElementsByClassName('table-check');
		for (var i = 0; i < tableCheck.length; i++)
		{
			Buttons.addClickHandler(tableCheck[i], table.checkUncheckOnClick, attributes);
			if (i == 0) element.tableCheck = tableCheck[i];
		}
		
		// Add 'click' handler to the 'Add filter row' and 'Remove filter rows' icons
		var filterRowActions = element.getElementsByClassName('table-filter-row-action');
		for (var i = 0; i < filterRowActions.length; i++)
		{
			Buttons.addClickHandler(filterRowActions[i], table.filterRowActionOnClick, attributes);
		}
		
		// Add 'click' handlers to all main table items
		// This must be different from the regular item click handler
		// since we must also support returning selected items from a
		// popup dialog (the regular click handler only supports navigating
		// to the 'view' or 'edit' page.
		var items = element.getElementsByClassName('table-item');
		for (var i = 0; i < items.length; i++)
		{
			var item = items[i];
			Buttons.addClickHandler(item, table.mainItemOnClick, attributes);
		}
		
		// Add 'click' handler to all 'Marked for removal' trashcan icons
		// This will delete the clicked item permanently
		var items = element.getElementsByClassName('table-delete-item');
		for (var i = 0; i < items.length; i++)
		{
			var item = items[i];
			Buttons.addClickHandler(item, Buttons.deleteItemPermanently, attributes);
		}

		// Add 'click' handler to all 'Shared item' icons
		// This will open the Share popup dialog 
		var items = element.getElementsByClassName('table-share-item');
		for (var i = 0; i < items.length; i++)
		{
			var item = items[i];
			Buttons.addClickHandler(item, Buttons.shareItem, attributes);
		}
	}
	Doc.addElementInitializer(internal.initTable);
	
	/**
		Event handler for showing a column selection menu 
		when right-clicking a column header.
	*/
	table.showColumnSelectionMenu = function(event)
	{
		var frm = document.forms['reggie'];
		// Context menu on 'right' mouse button
		// Can't just check the button since two events are sent ('mouseup' and 'contextmenu')
		var showContext = event.type == 'contextmenu' && event.button == 2;
		
		if (showContext)
		{
			event.preventDefault(); // Prevents the default right-click menu from appearing
			
			var tableId = Data.get(event.currentTarget, 'table-id');
			var afterColumnId = Data.get(event.currentTarget, 'column-id');
			
			var ctxMenu = Doc.element(tableId+'-select-visible-columns');
			var filterDiv = Doc.element(tableId+'-select-visible-columns-filter');
			var noMatchDiv = Doc.element(tableId+'-select-visible-columns-nomatch');

			if (!ctxMenu)
			{
				// Create the <div> with the menu items
				// One per table (in the rare case that there is more than one per page)
				ctxMenu = document.createElement('div');
				ctxMenu.className = 'menu vertical bg-filled-100 table-select-columns';
				ctxMenu.id = tableId+'-select-visible-columns';
				ctxMenu.tabIndex = 0;
				document.body.appendChild(ctxMenu);
				
				// Submit changes when ENTER key is pressed
				ctxMenu.form = document.forms[tableId];
				Events.doOnEnter(ctxMenu, Forms.submit)
	
				var colDefs = table.getColumnDefs(tableId);
				var activateFilter = colDefs.length > 30;
				if (activateFilter)
				{
					// Add a filter <div>
					filterDiv = document.createElement('div');
					filterDiv.className = 'menuitem column-filter bottomborder';
					filterDiv.id = tableId+'-select-visible-columns-filter';
					ctxMenu.appendChild(filterDiv);
					// Need to catch both events to handle {backspace} correctly in IE
					Events.addEventHandler(ctxMenu, 'keypress', table.filterColumnSelectionMenu);
					Events.addEventHandler(ctxMenu, 'keydown', table.filterColumnSelectionMenu);

					// Add 'no matching columns' <div>
					noMatchDiv = document.createElement('div');
					noMatchDiv.className = 'menuitem';
					noMatchDiv.innerHTML = 'No matching columns';
					noMatchDiv.id = tableId+'-select-visible-columns-nomatch';
				}
				
				var allCols = document.createElement('div');
				allCols.className = 'table-all-columns';
				ctxMenu.appendChild(allCols);
				
				var hasAnnotationSeparator = false;
				for (var i = 0; i < colDefs.length; i++)
				{
					var col = colDefs[i];
					if (!col.alwaysShow)
					{
						if (col.isAnnotation != hasAnnotationSeparator)
						{
							// Insert separator when switching 
							// between regular and annotation columns
							hasAnnotationSeparator = col.isAnnotation;
							var sep = document.createElement('div');
							sep.className = 'menuseparator';
							allCols.appendChild(sep);
						}
						
						var html = '<span class="padding">';
						html += '&nbsp;</span>'+Strings.encodeTags(col.title);
						
						var mnuItem = document.createElement('div');
						mnuItem.className = 'menuitem interactable table-col' + (col.visible?' visible-column':'');
						mnuItem.title = col.title;
						mnuItem.filterText = col.title.toLowerCase();
						mnuItem.innerHTML = html;
						Data.set(mnuItem, 'column-id', col.id);
						Data.set(mnuItem, 'table-id', tableId);
						Data.set(mnuItem, 'after-column-id', afterColumnId);
						Events.addEventHandler(mnuItem, 'click', table.showOrHideSelectedColumn);
						allCols.appendChild(mnuItem);
					}
				}
				
				if (activateFilter) ctxMenu.appendChild(noMatchDiv);
			}
			else
			{
				// If the menu already has been created we 
				// only need to update the 'after-column-id' attribute
				var menuItems = ctxMenu.getElementsByClassName('table-col');
				for (var i = 0; i < menuItems.length; i++)
				{
					var item = menuItems[i];
					Data.set(item, 'after-column-id', afterColumnId);
					item.innerHTML = '<span class="padding">&nbsp;</span>'+Strings.encodeTags(item.title);
					Doc.show(item);
				}
			}
			
			if (filterDiv) filterDiv.innerHTML = 'Type to filter columns';
			if (noMatchDiv) Doc.hide(noMatchDiv);
			
			// Position the menu near the mouse
			var winPos = App.getWindowPosition();
			var alignToRight = event.clientX + 200 > winPos.width;
			Menu.showTopMenu(ctxMenu, alignToRight ? winPos.width-event.clientX : event.clientX, event.clientY, alignToRight);
			ctxMenu.filterText = '';
			ctxMenu.focus();
		}
	}
	
	// Reacts to keys and either insert or delete a character from the filter field
	table.filterColumnSelectionMenu = function(event)
	{
		var target = event.currentTarget;
		if (event.keyCode == 8)
		{
			// Backspace delete one character
			target.filterText = target.filterText.substring(0, target.filterText.length-1);
			event.preventDefault();
		}
		else if (event.charCode > 0)
		{
			// Add the typed character to the filter
			target.filterText += String.fromCharCode(event.charCode);
		}
		else
		{
			return;
		}
		
		// Update the displayed filter text
		Doc.element(target.id+'-filter').innerHTML = Strings.encodeTags(target.filterText || 'Type to filter columns');
		
		var items = target.getElementsByClassName('table-col');
		var numMatching = 0;
		var filterText = target.filterText.toLowerCase();
		for (var i = 0; i < items.length; i++)
		{
			var item = items[i];
			var startIndex = item.filterText.indexOf(filterText);
			if (startIndex >= 0)
			{
				numMatching++;
				Doc.show(item);
				var endIndex = startIndex+filterText.length;
				var html = '<span class="padding">&nbsp;</span>';
				if (endIndex > startIndex)
				{
					html += Strings.encodeTags(item.title.substring(0, startIndex));
					html += '<b>'+Strings.encodeTags(item.title.substring(startIndex, endIndex))+'</b>';
					html += Strings.encodeTags(item.title.substring(endIndex));
				}
				else
				{
					html += Strings.encodeTags(item.title);
				}
				item.innerHTML = html;
			}
			else
			{
				Doc.hide(item);
			}
		}
		Doc.showHide(target.id+'-nomatch', numMatching==0);
	}
	
	/**
		Event handler invoked when selecting a column from the 
		context menu. If the selected column is visible it is hidden,
		otherwise it is inserted after the column that was clicked
		to bring up the context menu.
	*/
	table.showOrHideSelectedColumn = function(event)
	{
		var tableId = Data.get(event.currentTarget, 'table-id');
		var afterColumnId = Data.get(event.currentTarget, 'after-column-id');
		var columnId = Data.get(event.currentTarget, 'column-id');
		
		var visibleCols = table.getVisibleColumnIdsAsArray(tableId);
		var index = visibleCols.indexOf(columnId);
		if (index >= 0)
		{
			// Hide this column
			visibleCols.splice(index, 1);
			Doc.removeClass(event.currentTarget, 'visible-column');
		}
		else
		{
			// Insert after the column that was clicked
			var afterIndex = visibleCols.indexOf(afterColumnId);
			visibleCols.splice(afterIndex+1, 0, columnId);
			Doc.addClass(event.currentTarget, 'visible-column');
		}
		
		if (event.altKey || event.ctrlKey || event.shiftKey)
		{
			// Keep the context menu open and make it possible to select more columns
			var frm = document.forms[tableId];
			frm.columns.value = visibleCols.join(',');
			event.stopPropagation();
		}
		else
		{
			table.setColumns(tableId, visibleCols.join(','));
		}
	}
	
	/**
		Click handler that changes the sort order of the table. The sort property
		and sort direction should be stored in the data-sort-property and data-sort-direction
		properties of the target element.
	*/
	table.sortOnClick = function(event)
	{
		var column = event.currentTarget;
		var sortProperty = Data.get(column, 'sort-property');
		var sortDirection = Data.get(column, 'sort-direction');
		var tableId = Data.get(column, 'table-id');
		var frm = document.forms[tableId];
		var mainDirection = frm.direction.value;
		if (frm.sortby.value != sortProperty || mainDirection != sortDirection)
		{
			if (event.altKey || event.ctrlKey || event.shiftKey) 
			{
				// Append the current sort property to the list of already sorted columns
				// If the sort property is already used we keep it, but use the new sort direction
				var sortArray = frm.sortby.value.split(',');
				var index = sortArray.indexOf(sortProperty);
				// Check for '+' and '-' prefix if the sort order is different from main direction
				if (index == -1) index = sortArray.indexOf('+'+sortProperty);
				if (index == -1) index = sortArray.indexOf('-'+sortProperty);
				// Currently not sorted, insert at end
				if (index == -1) index = sortArray.length;
				// Add the sort property as the last property
				if (sortDirection != mainDirection)
				{
					sortProperty = (sortDirection == 'ASC' ? '+' : '-') + sortProperty;
					sortDirection = mainDirection;
				}
				sortArray[index] = sortProperty;
				sortProperty = sortArray.join(',');
			}
			frm.sortby.value = sortProperty;
			frm.direction.value = sortDirection;
			Forms.submit(frm);
		}
	}
	
	/**
		Start dragging a table column. Store table-id, column-id 
		and column-index on the table element.
	*/
	table.beginColumnDrag = function(event)
	{
		var dragSrc = {};
		dragSrc.tableId = Data.get(event.currentTarget, 'table-id');
		dragSrc.columnId = Data.get(event.currentTarget, 'column-id');
		dragSrc.columnIndex = Data.int(event.currentTarget, 'column-index');
		
		// Store the drag data on the <table> div
		Doc.element(dragSrc.tableId).dragSrc = dragSrc;
		event.dataTransfer.effectAllowed = 'move';
		// In FF, we also need to call setData() or the drag event will not start
		// but this results in an exception in IE!!!
		try
		{
			event.dataTransfer.setData('application/json', JSON.stringify(dragSrc));
		}
		catch (e)
		{}
	}
	
	/**
		Clean up after dragging.
	 */
	table.endColumnDrag = function(event)
	{
		var tableId = Data.get(event.currentTarget, 'table-id');
		Doc.element(tableId).dragSrc = null;
	}
	
	/**
		Checks if dropping is allowed at this target.
		We only allow dragging within the same table,
		and the target column must be different from the 
		source column.
	*/
	table.checkColumnDropTarget = function(event)
	{
		var tableId = Data.get(event.currentTarget, 'table-id');
		var dragSrc = Doc.element(tableId).dragSrc;
		
		var dropLocation = table.checkTheDropTarget(dragSrc, event.currentTarget);
		if (dropLocation)
		{
			// Enable drop target!
			event.preventDefault();
			event.dataTransfer.dropEffect = 'move';
			// Left or right depending on the order or columns
			Doc.addClass(event.currentTarget, dropLocation < 0 ? 'table-drop-left' : 'table-drop-right');
		}
	}

	/**
		Checks if dropping is allowed at this target.
		We only allow dragging within the same table,
		and the target column must be different from the 
		source column.
		returns: 0 = no drop, -1 drop to left, +1 drop to right
	*/
	table.checkTheDropTarget = function(dragSrc, dragTarget)
	{
		if (!dragSrc) return false;
		var tableId = Data.get(dragTarget, 'table-id');
		var columnIndex = Data.int(dragTarget, 'column-index');
		return tableId == dragSrc.tableId ? columnIndex - dragSrc.columnIndex : 0;
	}
	
	/**
		Remove drop indicator classes when leaving.
	*/
	table.leaveColumnDropTarget = function(event)
	{
		Doc.removeClass(event.currentTarget, 'table-drop-right');
		Doc.removeClass(event.currentTarget, 'table-drop-left');
	}

	/**
		Get the visibile columns as an array with column id values.
		The special case is that 'all' is expanded into actual
		column ids.
	*/
	table.getVisibleColumnIdsAsArray = function(tableId)
	{
		var columns = table.getColumns(tableId);
		var colArray;
		if (columns == 'all')
		{
			// Special case, load the existing columns
			var colDefs = table.getColumnDefs(tableId);
			colArray = [];
			for (var colNo = 0; colNo < colDefs.length; colNo++)
			{
				var col = colDefs[colNo];
				if (!col.alwaysHide) colArray[colArray.length] = col.id;
			}
		}
		else
		{
			colArray = columns.split(',');
		}
		return colArray;
	}
	
	/**
		Another column was dropped here!
	*/
	table.columnDropped = function(event)
	{
		Doc.removeClass(event.currentTarget, 'table-drop-right');
		Doc.removeClass(event.currentTarget, 'table-drop-left');
		
		// Check that we can drop here
		var tableId = Data.get(event.currentTarget, 'table-id');
		var dragSrc = Doc.element(tableId).dragSrc;
		if (!table.checkTheDropTarget(dragSrc, event.currentTarget)) return;
		
		// Get the existing columns as an array
		var colArray = table.getVisibleColumnIdsAsArray(tableId);
		
		var indexSrc = colArray.indexOf(dragSrc.columnId);
		var indexDest = colArray.indexOf(Data.get(event.currentTarget, 'column-id'));
		// Remove the dragged column from the array
		if (indexSrc >= 0) colArray.splice(indexSrc, 1);
		// Insert the dragged column before or after the column it was dropped on
		if (indexDest >= 0) colArray.splice(indexDest, 0, dragSrc.columnId);
		
		// Update and submit the table
		table.setColumns(tableId, colArray.join(','));
	}
	
	/**
		Event handler for hiding a single column.
	*/
	table.hideColumnOnClick = function(event)
	{
		var columnId = Data.get(event.currentTarget, 'column-id');
		var tableId = Data.get(event.currentTarget, 'table-id');
		
		var colArray = table.getVisibleColumnIdsAsArray(tableId);
		var colIndex = colArray.indexOf(columnId);
		
		if (colIndex >= 0)
		{
			colArray.splice(colIndex, 1);
			table.setColumns(tableId, colArray.join(','));
		}
	}

	/**
		Event handler for the page navigator. The actual target should 
		have a table-page attribute that specify the page to move to.
	*/
	table.navigatorOnClick = function(event)
	{
		var target = event.target;
		var page = Data.get(target, 'table-page');
		if (page == null) return;
		
		var tableId = Data.get(event.currentTarget, 'table-id');
		table.setPage(tableId, page);
	}
	
	/**
		Event handler for showing all hidden columns that have a filter.
		The hidden columns should be stored as a comma-separated list
		in 'data-hidden-columns' attribute and the element must have the
		table id stored in 'data-table-id' attribute.
	*/
	table.showColumnsOnClick = function(event)
	{
		var target = event.currentTarget;
		var hiddenColumns = Data.get(target, 'hidden-columns');
		var tableId = Data.get(target, 'table-id');
		table.showColumns(tableId, hiddenColumns);
	}
	
	/**
		Event handler for removing all filters from the table.
		The element must have the table id stored in 'data-table-id' 
		attribute.
	*/
	table.clearFilterOnClick = function(event)
	{
		var target = event.currentTarget;
		var tableId = Data.get(target, 'table-id');
		table.clearFilter(tableId);
	}
	
	/**
		Event handler for the 'view/presets' selection list.
	*/
	table.presetOnChange = function(event)
	{
		var presetSelector = event.currentTarget;
		var selected = presetSelector[presetSelector.selectedIndex];
		var cmd = selected.value;
		var tableId = presetSelector.form.name;
		
		// Reset the list to no selection
		presetSelector.selectedIndex = 0;		
		
		// Perform whatever command as indicated by the selected value
		if (cmd == 'save-as')
		{
			table.saveCurrentContext(tableId);
		}
		else if (cmd == 'manage')
		{
			table.manageContexts(tableId);
		}
		else if (cmd == 'load-preset')
		{
			var presetId = Data.get(selected, 'preset-id');
			table.loadContext(tableId, presetId);
		}
		else if (cmd == 'manage-columns')
		{
			table.configureColumns(tableId);
		}
		else if (cmd == 'clear-filter')
		{
			table.clearFilter(tableId);
		}
		else if (cmd == 'option')
		{
			var key = Data.get(selected, 'option-key');
			var value = Data.get(selected, 'option-value');
			table.setOption(tableId, key, value);
		}
	}

	/**
		Event handler for the 'Check/uncheck all' icon that select
		or deselect all checkboxes on the current table page.
		The element must have the table id stored in 'data-table-id' 
		attribute. The 'data-regexp' may be used to specify a regular
		expression to match checkbox names, otherwise all numeric 
		checkboxes are included.
	*/
	table.checkUncheckOnClick = function(event)
	{
		var target = event.currentTarget;
		var tableId = Data.get(target, 'table-id');
		var regexp = Data.get(target, 'regexp');
		var specialKey = event.altKey || event.ctrlKey || event.shiftKey;
		if (specialKey)
		{
			Forms.checkUncheck(document.forms[tableId], regexp);
		}
		else
		{
			Forms.toggleCheckboxes(document.forms[tableId], regexp);
		}

	}
	
	/**
		Event handler for the 'Add filter row' and 'Remove filter row' icons
		that adds or remove filter rows to/from the table.
		The element must have the table id stored in 'data-table-id' 
		attribute. If 'remove-row' is set, a the given row is removed,
		otherwise a new row is added.
	*/
	table.filterRowActionOnClick = function(event)
	{
		var target = event.currentTarget;
		var tableId = Data.get(target, 'table-id');
		var removeRow = Data.get(target, 'remove-row');
		if (removeRow != null)
		{
			table.removeFilterRow(tableId, removeRow);
		}
		else
		{
			table.addFilterRow(tableId);
		}
	}
	
	/**
		Event handler for clicking on the main item on the table. In normal
		list mode, the event handler simply forwards to the Item.itemOnClick
		method. In selection mode, the corresponding checkbox/radio button
		is selected and returned to the caller and the window is closed.
	*/
	table.mainItemOnClick = function(event)
	{
		var target = event.currentTarget;
		var tableId = Data.get(target, 'table-id');
		var itemId = Data.int(target, 'item-id');
		
		var frm = document.forms[tableId];
		var mode = frm.mode ? frm.mode.value : 'default';
		if (mode == 'selectone')
		{
			var index = Forms.checkRadio(frm.item_id, itemId);
			table.returnSelectedItems(tableId);
			App.closeWindow();
		}
		else if (mode == 'selectmultiple' || mode == 'selectmultiplenobuttons')
		{
			var checkbox = table.toggleCheckBox(tableId, itemId);
			table.returnSelectedItems(tableId);
			App.closeWindow();
		}
		else if (mode == 'selectfilter')
		{
			table.toggleCheckBox(tableId, itemId);
		}
		else
		{
			Items.itemOnClick(event);
		}
	}
	
	/**
		Get all columns defined for the specified table.
		@param tableDiv The ID or main table element
	*/
	table.getColumnDefs = function(tableDiv)
	{
		tableDiv = Doc.element(tableDiv);
		var columns = columnDefs[tableDiv.id];
		if (!columns)
		{
			columns = Data.json(tableDiv.id+'.table-data', 'column-defs');
			// Store each definitition using the 'id' as a key
			for (var i = 0; i < columns.length; i++)
			{
				columns['id'+columns[i].id] = columns[i];
			}
			
			columnDefs[tableDiv.id] = columns;
		}
		return columns;
	}

	/**
		Change the page of the table.
		@param tableDiv The ID or main table element
		@param page The new page number
	*/
	table.setPage = function(tableDiv, page)
	{
		tableDiv = Doc.element(tableDiv);
		var frm = document.forms[tableDiv.id];
		if (frm.page.value != page)
		{
			frm.page.value = page;
			Forms.submit(frm);
		}
	}
	
	/**
		Change the visible columns of the table.
		@param tableDiv The ID or main table element
		@param columns Comma-separated list of column ID:s of the visible columns, or the value 'all'
	*/
	table.setColumns = function(tableDiv, columns)
	{
		tableDiv = Doc.element(tableDiv);
		var frm = document.forms[tableDiv.id];
		if (frm.columns.value != columns)
		{
			frm.columns.value = columns;
			Forms.submit(frm);
		}
	}

	/**
		Show columns in the table, preserving the currently visible columns.
		@param tableDiv The ID or main table element
		@param columns Comma-separated list of column ID:s of the visible columns, or the value 'all'
	*/
	table.showColumns = function(tableDiv, columns)
	{
		tableDiv = Doc.element(tableDiv);
		var frm = document.forms[tableDiv.id];
		if (frm.columns.value != columns)
		{
			frm.columns.value += ',' + columns;
			Forms.submit(frm);
		}
	}
	

	/**
		Get the visible columns of the table.
		@param tableDiv The ID or main table element
	*/
	table.getColumns = function(tableDiv)
	{
		tableDiv = Doc.element(tableDiv);
		var frm = document.forms[tableDiv.id];
		return frm.columns.value;
	}
	
	
	/**
		Clear the all filters 
		@param tableDiv The ID or main table element
	*/
	table.clearFilter = function(tableDiv)
	{
		tableDiv = Doc.element(tableDiv);
		var frm = document.forms[tableDiv.id];
		Forms.addHidden(frm, 'filter:clearAll', '1');
		Forms.submit(frm);
	}
	
	/**
		Add a new filter row to the table
		@param tableDiv The ID or main table element
	*/
	table.addFilterRow = function(tableDiv)
	{
		tableDiv = Doc.element(tableDiv);
		var frm = document.forms[tableDiv.id];
		Forms.addHidden(frm, 'addFilterRows', '1');
		Forms.submit(frm);
	}
	
	/**
		Remove a filter row from the table
		@param tableDiv The ID or main table element
		@param filterIndex The index of the filter row, starting at 0
	*/
	table.removeFilterRow = function(tableDiv, filterIndex)
	{
		tableDiv = Doc.element(tableDiv);
		var frm = document.forms[tableDiv.id];
		Forms.addHidden(frm, 'removeFilterRow', filterIndex);
		Forms.submit(frm);
	}
	
	/**
		Open a popup window that allows a user to configure which columns
		that should be visible and their order.
		@param table The ID or main table element
		@param settingName The name of the setting that is used to store the
			column visibility and order, if not specified the default is 'columns'
	*/
	table.configureColumns = function(tableDiv, settingName)
	{
		tableDiv = Doc.element(tableDiv);
		var itemType = Data.get(tableDiv, 'item-type');
		var subContext = Data.get(tableDiv, 'subcontext', '');
		var enableInheritedAnnotations = Data.get(tableDiv, 'inherited-annotations');
		var enableRelatedItemColumns = Data.get(tableDiv, 'relateditem-columns');
		var disableLinkedItemColumns = Data.get(tableDiv, 'no-linkeditem-columns');
		
		var url = App.getRoot()+'common/columns/configure.jsp?ID='+App.getSessionId();
		url += '&table_id='+tableDiv.id+'&item_type='+itemType+'&subcontext='+subContext;
		if (enableInheritedAnnotations) url += '&enableInheritedAnnotations=1';
		if (enableRelatedItemColumns) url += '&enableRelatedItemColumns=1';
		if (!disableLinkedItemColumns) url += '&enableLinkedItemColumns=1';

		if (settingName) url += '&settingName='+settingName;
		Dialogs.openPopup(url, 'ConfigureColumns', 750, 450);
	}

	/**
		Open a popup dialog that allows the user to save the current 
		table settings to a named preset.
		@param table The ID or main table element
	*/
	table.saveCurrentContext = function(tableDiv)
	{
		tableDiv = Doc.element(tableDiv);
		var itemType = Data.get(tableDiv, 'item-type');
		var subContext = Data.get(tableDiv, 'subcontext', '');
		var url = App.getRoot()+'common/context/saveas.jsp?ID='+App.getSessionId();
		url += '&item_type='+itemType+'&subcontext='+subContext;
		Dialogs.openPopup(url, 'SaveCurrentContext', 450, 300);
	}
	
	/**
		Open a popup dialog that allows the user to manage (eg. load or delete)
		all table setting presets that have been saved for the given table.
		@param table The ID or main table element
	*/
	table.manageContexts = function(tableDiv)
	{
		tableDiv = Doc.element(tableDiv);
		var itemType = Data.get(tableDiv, 'item-type');
		var subContext = Data.get(tableDiv, 'subcontext', '');
		var url = App.getRoot()+'common/context/manage.jsp?ID='+App.getSessionId();
		url += '&item_type='+itemType+'&subcontext='+subContext;
		Dialogs.openPopup(url, 'ManageContexts', 450, 300);
	}
	
	/**
		Load the table settings preset with the given id.
		@param table The ID or main table element
	*/
	table.loadContext = function(tableDiv, contextId)
	{
		tableDiv = Doc.element(tableDiv);
		var frm = document.forms[tableDiv.id];
		Forms.addHidden(frm, 'context', contextId);
		frm.cmd.value = 'LoadContext';
		Forms.submit(frm);
	}
	
	/**
		Change a table option and reload the table.
		
		@param table The ID or main table element
		@param name The name of the option
		@param value The value of the option
	*/
	table.setOption = function(tableDiv, name, value)
	{
		tableDiv = Doc.element(tableDiv);
		var frm = document.forms[tableDiv.id];
		frm[name].value = value;
		Forms.submit(frm);
	}

	
	/**
		Open a popup window with the given width and height
		and submit the table form (POST) to that popup.
		@param tableDiv The ID or main table element
		@param cmd The command value to use when sending the form
			(the old command is restored afterwards)
		@param width The width in pixels of the popup (may be scaled)
		@param height The height in pixels of the popup (may be scaled)
	*/
	table.submitToPopup = function(tableDiv, cmd, width, height)
	{
		tableDiv = Doc.element(tableDiv);
		var frm = document.forms[tableDiv.id];
		var oldCmd = frm.cmd.value;
		var oldTarget = frm.target;
		frm.cmd.value = cmd;
		frm.target = tableDiv.id+cmd;
		Dialogs.openPopup('', frm.target, width, height);
		frm.submit();
		frm.target = oldTarget;
		frm.cmd.value = oldCmd;
	}
	
	/**
		Submit the table using the given cmd if there is 
		at least one item selected.
		@param tableDiv The ID or main table element
		@param cmd The command value to use when sending the form
		@param regexp A regular expression to use for matching checkboxes in the table,
			if not specified all checkboxes with a numeric name are used
	 */
	table.submitSelected = function(tableDiv, cmd, regexp)
	{
		tableDiv = Doc.element(tableDiv);
		var frm = document.forms[tableDiv.id];
		if (!table.checkIfSelected(tableDiv, regexp))
		{
			return;
		}
		frm.cmd.value = cmd;
		Forms.submit(frm);
	}
	
	/**
		Submit a request to delete (or put in trashcan) all selected items
		in the table.
		@param tableDiv The ID or main table element
		@param regexp A regular expression to use for matching checkboxes in the table,
			if not specified all checkboxes with a numeric name are used
		@param confirmFirst If set, ask for confirmation before deleting the items
	*/
	table.deleteItems = function(tableDiv, regexp, confirmFirst)
	{
		tableDiv = Doc.element(tableDiv);
		var frm = document.forms[tableDiv.id];
		var numSelected = table.checkIfSelected(tableDiv, regexp);
		if (numSelected == 0) return;
		if (confirmFirst)
		{
			if (!confirm('You are about to delete '+numSelected+' items. This can\'t be undone. Continue?'))
			{
				return;
			}
		}
		frm.cmd.value = 'DeleteItems';
		Forms.submit(frm);
	}
	
	/**
		Submit a request to undelete (remove from trashcan) all selected items
		in the table.
		@param tableDiv The ID or main table element
		@param regexp A regular expression to use for matching checkboxes in the table,
			if not specified all checkboxes with a numeric name are used
	*/
	table.restoreItems = function(tableDiv, regexp)
	{
		tableDiv = Doc.element(tableDiv);
		var frm = document.forms[tableDiv.id];
		if (!table.checkIfSelected(tableDiv, regexp))
		{
			return;
		}
		frm.cmd.value = 'RestoreItems';
		Forms.submit(frm);
	}

	
	/**
		Open a popup dialog that allows a user to select a new owner for
		all items that are selected in the given table. 
		@param tableDiv The ID or main table element
		@param regexp An optional regular expression matching the names
			of checkboxes that can be selected, if not specified it matches
			checkboxes with numeric names (eg. the name is the id of the item
			in the table)
	 */
	table.setOwnerOfItems = function(tableDiv, regexp)
	{
		tableDiv = Doc.element(tableDiv);
		if (!table.checkIfSelected(tableDiv, regexp))
		{
			return;
		}
		table.submitToPopup(tableDiv, 'SetOwnerOfItems', 450, 300);
	} 

	/**
		Open a popup dialog that allows a user to specify sharing permissions
		for all items that are selected in the given table. 
		@param tableDiv The ID or main table element
		@param regexp An optional regular expression matching the names
			of checkboxes that can be selected, if not specified it matches
			checkboxes with numeric names (eg. the name is the id of the item
			in the table)
	 */
	table.shareItems = function(tableDiv, regexp)
	{
		tableDiv = Doc.element(tableDiv);
		if (!table.checkIfSelected(tableDiv, regexp))
		{
			return;
		}
		table.submitToPopup(tableDiv, 'ShareItems', 600, 400);
	}

	/**
		Open a popup dialog that allows a user to bathc inherit annotations
		to all items that are selected in the given table. 
		@param tableDiv The ID or main table element
		@param regexp An optional regular expression matching the names
			of checkboxes that can be selected, if not specified it matches
			checkboxes with numeric names (eg. the name is the id of the item
			in the table)
	 */
	table.inheritAnnotations = function(tableDiv, regexp)
	{
		tableDiv = Doc.element(tableDiv);
		if (!table.checkIfSelected(tableDiv, regexp))
		{
			return;
		}
		table.submitToPopup(tableDiv, 'InheritAnnotations', 750, 500);
	}

	
	/**
		Open a popup window for selecting a plug-in that can run in the 
		current list context
		@param tableDiv The ID or main table element
		@param pluginType One of: IMPORT, EXPORT, ANALYZE or OTHER (default)
		@param cmd Override the default command for running the plugin
	*/
	table.runPlugin = function(tableDiv, pluginType, cmd)
	{
		if (!cmd)
		{
			if (pluginType == 'EXPORT')
			{
				cmd = 'ExportItems';
			}
			else if (pluginType == 'IMPORT')
			{
				cmd = 'ImportItems';
			}
			else if (pluginType == 'ANALYZE')
			{
				cmd = 'RunListAnalysisPlugin';
			}
			else
			{
				cmd = 'RunListPlugin';
			}
		}
		table.submitToPopup(tableDiv, cmd, 750, 500);
	}

	/**
		Toggle the selected status of a checkbox with 
		@param tableDiv The ID or main table element
		
		@returns The toggled  element or null if no element was found
	*/
	table.toggleCheckBox = function(tableDiv, itemId)
	{
		tableDiv = Doc.element(tableDiv);
		var frm = document.forms[tableDiv.id];

		for (var i=0; i < frm.elements.length; i++)
		{
			var element = frm.elements[i];
			if (element.type == 'checkbox' && element.name == itemId)
			{
				element.checked = !element.checked;
				Events.sendChangeEvent(element);
				return element;
			}
		}
		return null;
	}

	/**
		Get the ID of all selected items in a table.
		
		@param tableDiv The ID or main table element
		@param regexp An optional regular expression matching the names
			of checkboxes that can be selected, if not specified it matches
			checkboxes with numeric names (eg. the name is the id of the item
			in the table)
		@return An array object with the ID:s
	*/
	table.getSelected = function(tableDiv, regexp, messageIfNoneSelected)
	{
		tableDiv = Doc.element(tableDiv);
		if (!regexp) regexp = /\d+/;

		var frm = document.forms[tableDiv.id];
		var items = [];
		if (frm.item_id)
		{
			// Radio buttons
			var element = Forms.getCheckedRadio(frm.item_id);
			if (element != null)
			{
				items[items.length] = element.value;
			}
		}
		else
		{
			// Checkboxes
			for (var i=0; i < frm.elements.length; i++)
			{
				var element = frm.elements[i];
				if (element.type == 'checkbox' && element.name.match(regexp) && element.checked)
				{
					items[items.length] = element.value;
				}
			}
		}
		
		if (items.length == 0 && messageIfNoneSelected)
		{
			Forms.showNotification(tableDiv.tableCheck, messageIfNoneSelected, null, 'pointer-left');
		}
		
		return items;
	}

	
	/**
		Return the selected items to the parent window using either a callback
		method (older way) or an event notification (new way).
		
		@param tableDiv The ID or main table element
		@param regexp An optional regular expression matching the names
			of checkboxes that can be selected, if not specified it matches
			checkboxes with numeric names (eg. the name is the id of the item
			in the table)

		@return The number of selected items
	*/
	table.returnSelectedItems = function(tableDiv, regexp)
	{
		tableDiv = Doc.element(tableDiv);
		if (!regexp) regexp = /\d+/;
		
		var frm = document.forms[tableDiv.id];
		var itemType = Data.get(tableDiv, 'item-type');
		
		// Should we use 'callback method' or 'event notification'?
		var callback = frm.callback ? frm.callback.value : null;
		var notifyTarget = callback ? window.top.opener.document.getElementById(callback) : null;
		var callbackMethod = callback ? window.top.opener[callback] : null;
		
		if (!notifyTarget && !callbackMethod) 
		{
			notifyTarget = tableDiv;
		}
		
		var selected = [];
		if (frm.item_id)
		{
			// Radio buttons allow single selection
			var element = Forms.getCheckedRadio(frm.item_id);
			if (element != null)
			{
				var id = element.value.match(/^\d+$/) ? parseInt(element.value) : element.value;
				selected[selected.length] = {'id': id, 'name': element.title, 'itemType': itemType};
			}
		}
		else
		{
			// Check boxes allow multiple selections
			for (var i=0; i < frm.elements.length; i++)
			{
				var element = frm.elements[i];
				if (element.type == 'checkbox' && element.name.match(regexp) && element.checked)
				{
					var id = element.value.match(/^\d+$/) ? parseInt(element.value) : element.value;
					selected[selected.length] = {'id': id, 'name': element.title, 'itemType': itemType};
				}
			}
		}
		
		var remaining = selected.length-1;
		if (notifyTarget)
		{
			Events.sendCustomEvent(notifyTarget, 'base-selected-start', { 'numSelected': selected.length });
		}
		for (var i = 0; i < selected.length; i++)
		{
			var s = selected[i];
			s.remaining = remaining--;
			if (notifyTarget)
			{
				// Send event to the target in the opener window
				Events.sendCustomEvent(notifyTarget, 'base-selected', s);
			}
			else if (callbackMethod)
			{
				// Call the callback method in the opener window
				callbackMethod.call(null, s.id, s.name);
			}
		}

		return selected.length;
	}

	table.checkIfSelected = function(tableDiv, regexp, messageIfNoneSelected)
	{
		tableDiv = Doc.element(tableDiv);
		var frm = document.forms[tableDiv.id];
		var numSelected = Forms.numChecked(frm, regexp);
		if (numSelected == 0)
		{
			if (!messageIfNoneSelected) 
			{
				messageIfNoneSelected = 'Please select at least one item in the list';
			}
			Forms.showNotification(tableDiv.tableCheck, messageIfNoneSelected, null, 'pointer-left');
		}
		return numSelected;
	}

	return table;
}();


